#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2016, 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
This function:

* Parses the global tag listing string passed in as an argument
* Updates the local properties, which are stored in env.properties
* Checks global topology rules (see :func:`check_rules`)

Properties set by this for later Python steps:

- ProcessTagging_DIR
- globalTagListing

Properties set by this for compatibility to other scripts:

- DB_{HOSTNAME,VENDOR}
- hosts, ihshost
- NUMBER_OF_CUSTOM_NODES
- NODE{#}_{HostName,AppList,NodeName}
- JTSRULHost
- databaseList

Usage::

    process_global_tags.py globalTagListing CMenable LDAPPROP DBVendor [CMActivationKey]?

(see :func:`main`)
'''
import os
import sys

# find qdlib from parent paths and add the first one to the Python search path
# pylint: disable=line-too-long
sys.path.append(next(p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and (p + '/qdlib') not in sys.path))
# pylint: enable=line-too-long

from qdlib import qd, qd_topo, qd_props


# pylint: disable=too-many-branches,too-many-statements
def check_rules(topo, cm_enable, cm_activation_key):
    '''
    Check topology rules and exit with error if they are not met:

    Rules
     1. only one IHS or Liberty must exist (currently Liberty is only for V1)
     2. only one database must exist
     3. JTS must exist
     4. only one instance of any application can exist
     5. if cm_enable is TRUE
          * ensure LDX and GC are present
          * ensure activation key is set

    :param topo: object containing information about the topology
    :param cm_enable: TRUE enable Configuration Management, FALSE disable Configuration Management
    :param cm_activation_key: activation key used if cm_enable is TRUE
    '''

    qd.echo_info("--- check system rules ---")
    rules_invalidated = False

    # 1. only one IHS or Liberty must exists (not optional yet!)
    qd.echo_info("check only 1 IHS or Liberty allowed rule")
    ihs_count = len(topo.get_hosts_by_tags(qd_topo.TAG_IHS))
    liberty_count = len(topo.get_hosts_by_tags(qd_topo.TAG_LIBERTY))
    if ihs_count + liberty_count == 1:
        qd.echo_info("--> rule validated")
    else:
        if ihs_count + liberty_count == 0:
            qd.echo_error("System has no IHS or Liberty present --> rule invalidated")
        else:
            qd.echo_error("System has too many [%d] IHS and Liberty present --> rule invalidated"
                          % (ihs_count + liberty_count))
        rules_invalidated = True

    # 2. only one database must exist
    qd.echo_info("check only 1 DATABASE allowed rule")
    db_count = len(topo.get_hosts_by_tags(*qd_topo.ALL_DB_TAGS))
    if db_count == 1:
        qd.echo_info("--> rule validated")
    else:
        if db_count == 0:
            qd.echo_error("System has no database present --> rule invalidated")
        else:
            qd.echo_error("System has too many [%d] DATABASE server present --> rule invalidated"
                          % db_count)
        rules_invalidated = True

    # 3. JTS must exist
    qd.echo_info("check JTS present rule")
    if topo.get_hosts_by_tags(qd_topo.TAG_JTS):
        qd.echo_info("--> rule validated")
    else:
        qd.echo_error("System has JTS application present --> rule invalidated")
        rules_invalidated = True

    # 4. only one instance of any application can exist
    qd.echo_info("Check no duplicate APPS rule")
    dup_apps_counts = {}
    for app_tag in qd_topo.ALL_APP_TAGS:
        app_hosts = topo.get_hosts_by_tags(app_tag)
        if len(app_hosts) > 1:
            dup_apps_counts[app_tag] = len(app_hosts)
    if not dup_apps_counts.keys():
        qd.echo_info("--> rule validated")
    else:
        qd.echo_error("Duplicate APPS exist --> rule invalidated")
        rules_invalidated = True
        for app, count in dup_apps_counts.items():
            qd.echo_info('[%s] count [%d]' % (app, count))

    # 5. if cm_enable is TRUE
    #      ensure LDX and GC are present
    #      ensure activation key is set
    if cm_enable:
        qd.echo_info("CM enabled, check LDX and GC applications are present rule")
        ldx_present = qd_topo.TAG_LDX in topo.get_all_apps_list()
        gc_present = qd_topo.TAG_GC in topo.get_all_apps_list()
        if ldx_present and gc_present:
            qd.echo_info("--> rule validated")
        else:
            qd.echo_error("LDX present [%s] or GC present [%s] "
                          "--> rule invalidated" % (ldx_present, gc_present))
            rules_invalidated = True
        qd.echo_info("CM enabled, check cm_activation_key rule")
        if cm_activation_key:
            qd.echo_info("--> rule validated")
        else:
            qd.echo_error("cm_activation_key missing [%s] "
                          "--> rule invalidated" % (cm_activation_key))
            rules_invalidated = True

    # confirm all is valid or exit
    if not rules_invalidated:
        qd.echo_info("system rules ALL VALID")
    else:
        qd.echo_error("ONE OR MORE RULES INVALID")
        qd.exit_(1)


# pylint: disable=too-many-arguments
def update_server_props(topo, env_props, cm_enable, cm_activation_key, ldap_prop, db_vendor):
    # pylint: disable=unused-argument
    '''
    update properties contained in {env_props}

    :param topo: object containing information about the topology
    :param env_props: environment properties
    :param cm_enable: TRUE enable Configuration Management, FALSE disable Configuration Management
        (currently not used in this method)
    :param cm_activation_key: activation key used if cm_enable is TRUE
        (currently not used in this method)
    :param ldap_prop: reference to the LDAP server being used e.g. CustomLDAP
        (currently not used in this method)
    :param db_vendor: db2 or oracle
    '''

    # ihs
    ihs_host = topo.get_ihs_host()
    if ihs_host:
        env_props.set_value('ihshost', ihs_host.hostname)

    # db
    env_props.set_value('DB_HOSTNAME', topo.get_db_hostname())
    env_props.set_value('DB_VENDOR', db_vendor)

    # jts (Note: if an ihs server exists this may be overwritten in the final setup step)
    jts_host = topo.get_jts_host()
    env_props.set_value("JTSURLHost", jts_host.hostname)

    # apps
    app_hosts = topo.get_app_hosts()
    for was_node_index, host in enumerate(app_hosts):
        was_node_index += 1
        env_props.set_value('NODE%d_HostName' % was_node_index, host.hostname)
        env_props.set_value('NODE%d_AppList' % was_node_index, host.get_local_apps())
        if host.was_node_name:
            env_props.set_value('NODE%d_NodeName' % was_node_index, host.was_node_name)

    # -- hostnames --
    env_props.set_value('hosts', [h.hostname for h in app_hosts])
    # rs does not have a table
    # -- database tables --
    table_list = topo.get_table_list()
    env_props.set_value('databaseList', table_list)
    # -- apps --
    # appList and allAppsList both existed in the original shell scripts,
    # - looks like allAppsList could be obsolete
    env_props.set_value('appList', topo.get_all_apps_list())
    env_props.set_value('allAppsList', topo.get_all_apps_list())
    # -- custom / dwa nodes --
    env_props.set_value('NUMBER_OF_CUSTOM_NODES', str(len(app_hosts)))
    # DWA is obsolete, remove during future refactoring
    env_props.set_value('NUMBER_OF_DWA_NODES', str(0))


def main(argv):
    '''
    Main method; see module documentation.

    :param globalTagListing: list of elements separated by [;] representing each server
    :param CMenable: TRUE enable Configuration Management, FALSE disable Configuration Management
    :param LDAPPROP: reference to the LDAP server being used e.g. CustomLDAP
    :param DBVendor: database vendor, e.g., db2, oracle
    :param CMActivationKey: (optional) activation key used if cm_enable is TRUE

    *each element in globalTagListing is structured as
    [hostname]:[csv list of tags]:[(optionally) was node name]*
    '''

    # process input
    # pylint: disable=line-too-long
    # The CMActivationKey is optional; not used in test servers, just in pipeline.
    arg_parsed = qd.process_inputs(argv,
                                   'globalTagListing,CMenable[True/False],LDAPPROP,DBVendor,[CMActivationKey]', 1, False)
    # pylint: enable=line-too-long

    # pylint error unbalanced-tuple-unpacking on this line because no guarantee that arg_parsed,
    # i.e., qd.processInputs() always returns 5 items.  Probably should refactor.  For now, ignore.
    # pylint: disable=unbalanced-tuple-unpacking
    global_tag_listing, cm_enable, ldap_prop, db_vendor, cm_activation_key = arg_parsed

    # Normally, let process_inputs to output the info about each argument.  But with
    # optional arguments, the output gets more difficult.  Do it here, now that the
    # values are set.
    qd.echo_info('%s: %s' % ("globalTagListing", global_tag_listing))
    qd.echo_info('%s: %s' % ("CMenable[True/False]", cm_enable))
    qd.echo_info('%s: %s' % ("LDAPPROP", ldap_prop))
    qd.echo_info('%s: %s' % ("DBVendor", db_vendor))
    qd.echo_info('%s: %s' % ("[CMActivationKey]", cm_activation_key))

    # convert to boolean for later use
    cm_enable = qd.parse_boolean(cm_enable, 'cm_enable', False)

    # convert to upper case for later use
    db_vendor = db_vendor.lower()

    # check prop file
    env_props = qd_props.get_default_env_props()
    if qd.is_linux():
        prop_file_permissions = qd.check_output_strip(['stat', '-c', '%a', env_props.filename])
        qd.echo_info("capture file [%s] permissions [%s]" % (env_props.filename,
                                                             prop_file_permissions))

    # get localhost
    local_host = env_props.get_values('localHostname', localHostname=qd.get_host_name())
    qd.echo_info('local_host [%s]' % local_host)

    # Main
    # - check topology globally
    topo = qd_topo.QDTopo(global_tag_listing)
    assert topo.get_host(local_host), 'local_host [%s] must be contained in ' \
        'the global_tag_listing [%s]' % (local_host, global_tag_listing)
    check_rules(topo, cm_enable, cm_activation_key)
    # - generate .sh-style properties
    update_server_props(topo, env_props, cm_enable, cm_activation_key, ldap_prop, db_vendor)
    # - add mandatory properties for Python-based scripts
    env_props.set_value('ProcessTagging_DIR', os.getcwdu())
    env_props.set_value('globalTagListing', global_tag_listing)
    # -- cm_enable --
    env_props.set_value('EnableCM', 'TRUE' if cm_enable else 'FALSE')
    env_props.set_value('CMActivationKey', cm_activation_key)
    env_props.set_value('LDAPPROP', ldap_prop)
    env_props.set_value('DB_VENDOR', db_vendor)
    # -- no more property file update left, so put original permissions back --
    if qd.is_linux():
        qd.echo_info("ensure file [%s] permissions [%s]" % (env_props.filename,
                                                            prop_file_permissions))
        qd.check_call(['chmod', prop_file_permissions, env_props.filename])
        qd.check_call(['ls', '-la', env_props.filename])


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main(sys.argv)
    exit(0)
