#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2016, 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
This script checks:

- Minimum Python version
- Minimum Windows / Linux OS version
'''

import os
import platform
import sys

# find qdlib from parent paths and add the first one to the Python search path
# pylint: disable=line-too-long
sys.path.append(next(p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and (p + '/qdlib') not in sys.path))
# pylint: enable=line-too-long

from qdlib import qd


# CONSTANTS for check_os_version()
LINUX_MIN_VERSION = 6.6  #:
''' Minimum linux OS version '''
WINDOWS_SUPPORTED_RELEASES = ['2012Server']  #:
''' List of supported windows releases '''

# CONSTANTS for check_python_version()
'''
.. warning:: UCD scripts only work with Python 2.x, not 3.x.

'''
PYTHON_VERSION_MIN = "2.7.5".split('.')
PYTHON_MAJOR_VERSION = PYTHON_VERSION_MIN[0]  #:
# - major version is the first part e.g., the 2 from 2.7.5
PYTHON_MINOR_VERSION_MIN = PYTHON_VERSION_MIN[1]  #:
# - minor version is the second part e.g., the 7 from 2.7.5
PYTHON_PATCH_VERSION_MIN = PYTHON_VERSION_MIN[2]  #:
# - patch version is the third part e.g., the 5 from 2.7.5


def check_os_version():
    '''
    OS can only be Linux or Windows

    - Linux : the OS version must be greater than or equal to the minimum version
    - Windows: the OS release must be in the supported list

    :return: true if the operating system is considered valid; false otherwise
    '''
    # NOTE:
    # linux will be tested numerically against constant LINUX_MIN_VERSION
    # windows will be tested by word comparison,
    #         against the list constant WINDOWS_SUPPORTED_RELEASES
    #         9th March 2017: win2012 and win2016Std both are returning '2012Server'

    os_name = platform.system()
    qd.echo_info("Operating system [%s]" % os_name)
    if os_name == "Linux":
        os_dist, os_version, os_id = platform.linux_distribution()
        qd.echo_info("OS distribution is [%s] [%s] [%s]" % (os_dist, os_version, os_id))
        if float(os_version) >= LINUX_MIN_VERSION:
            qd.echo_info("Minimum version [%s], version [%s] is supported"
                         % (LINUX_MIN_VERSION, os_version))
            valid_os = True
        else:
            qd.echo_error("Minimum version [%s], version [%s] is unsupported"
                          % (LINUX_MIN_VERSION, os_version))
            valid_os = False

    elif os_name == "Windows":
        os_release = platform.release()
        if os_release in WINDOWS_SUPPORTED_RELEASES:
            qd.echo_info("Windows release [%s] is in the supported list [%s]"
                         % (os_release, WINDOWS_SUPPORTED_RELEASES))
            valid_os = True
        else:
            qd.echo_error("Windows release [%s] is not in the supported list [%s]"
                          % (os_release, WINDOWS_SUPPORTED_RELEASES))
            valid_os = False
    else:
        # unsupported
        valid_os = False

    # supported or not?
    if valid_os:
        qd.echo_info("Supported operating system")
    else:
        qd.echo_error("Unsupported operating system")

    # return the result of the test
    return valid_os


def check_python_version():
    '''
    The version of Python installed and in the PATH  must be Python 2.x and minor version must be
    greater than or equal to minimum minor version.

    :return: true if the Python version is considered valid; false otherwise
    '''

    qd.echo_info("Minimum required Python version [%s]" % PYTHON_VERSION_MIN)

    # using platform.python_version_tuple instead of sys.version
    # as the returned value will always include the patchlevel (it defaults to 0).
    py_major_version, py_minor_version, py_patch_version = platform.python_version_tuple()
    py_version = py_major_version + '.' + py_minor_version + '.' + py_patch_version
    qd.echo_info("Current Python version [%s]" % py_version)
    # now check the version bit by bit

    # the major version must match to be valid
    if py_major_version != PYTHON_MAJOR_VERSION:
        qd.echo_error("Python version [%s] fails wrong major version test (position 1)"
                      % py_version)
        valid_python = False
    else:
        if int(py_minor_version) < int(PYTHON_MINOR_VERSION_MIN):
            # if minor version is less than then it is not valid
            qd.echo_error("Python version [%s] fails wrong minor version test (position 2)"
                          % py_version)
            valid_python = False
        elif int(py_minor_version) > int(PYTHON_MINOR_VERSION_MIN):
            # if minor version is greater than then it is valid
            qd.echo_info("Python version [%s] meets requirements" % py_version)
            valid_python = True
        else:
            # if minor version is equal then test the patch version
            # if patch version is less than then it is not valid
            # if patch version is greater than or equal then it is valid
            if int(py_patch_version) < int(PYTHON_PATCH_VERSION_MIN):
                qd.echo_error("Python version [%s] fails wrong patch version test (position 3)"
                              % py_version)
                valid_python = False
            else:
                qd.echo_info("Python version [%s] meets requirements" % py_version)
                valid_python = True

    # return the result of the test
    return valid_python


def main():
    '''Main function; see module documentation.'''

    valid_os = check_os_version()
    valid_python = check_python_version()

    if valid_os and valid_python:
        qd.echo_info("SUCCESS: Basic pre req test passed")
    else:
        qd.echo_error("Basic pre req test failed")
        qd.exit_(1)


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main()
    exit(0)
