#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2016, 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
- Use ${localTags} To drive "Link Points" (linux)
- Get credentials files

Sub-scripts for Linux:

- CoreAutomationPrep_packages.sh
- CoreAutomationPrep_checkID.sh
- config_xvfb.sh
'''
import os
import re
import shutil
import socket
import sys

# find qdlib from parent paths and add the first one to the Python search path
# pylint: disable=line-too-long
sys.path.append(next(p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and (p + '/qdlib') not in sys.path))
# pylint: enable=line-too-long

from qdlib import qd, qd_props, media, qd_topo, install_util


# pylint: disable=too-many-locals,too-many-statements,too-many-branches
def check_ifix_version(env_props):
    '''
    Check if iFix is enabled that the iFix versions and base version match these rules:

    1. A CLM iFix must be present to install a DM or DMMS iFix
       e.g., being installed as a pair or triple
    2. iFix must be on the same stream as the build
       e.g., 601 iFix only installed along with a CE601 build
    3. DMY or DMMSY can only be installed alongside CLM iFix00Y
       (that is all that DM test team test)
    4. DM or DMMS iFixes can be installed independently of each other

    :param env_props: environment properties
    '''

    # get the properties to check in the rules
    # env.properties names are
    # [appMediaRoot][iFixEnabled][CLMiFixVersion][DMiFixVersion][DMMSiFixVersion]
    # pylint: disable=line-too-long
    ifix_enabled, clm_ifix_version, dm_ifix_version, dmms_ifix_version = env_props.get_values('iFixEnabled',
                                                                                              'CLMiFixVersion', 'DMiFixVersion', 'DMMSiFixVersion',
                                                                                              CLMiFixVersion=None, DMiFixVersion=None, DMMSiFixVersion=None)
    # pylint: enable=line-too-long
    # only continue if "iFixing is enabled"
    if qd.parse_boolean(ifix_enabled, 'ifix_enabled'):
        qd.echo_info("potentially installing iFix as iFixEnabled [%s]" % ifix_enabled)
    else:
        qd.echo_info("not installing iFix as iFixEnabled [%s], "
                     "so not checking iFix version compatibility" % ifix_enabled)
        return

    # rule1. a CLM iFix must be present to install a DM or DMMS iFix
    # eg being installed as a pair or triple
    if clm_ifix_version == 'NONE':
        if (dm_ifix_version == 'NONE') and (dmms_ifix_version == 'NONE'):
            qd.echo_info("no iFixes are being installed as CLMiFixVersion [%s], DMiFixVersion [%s]"
                         " and DMMSiFixVersion [%s]"
                         % (clm_ifix_version, dm_ifix_version, dmms_ifix_version))
            return
        else:
            qd.echo_error("Can not install DMiFixVersion [%s] or DMMSiFixVersion [%s]"
                          " as CLMiFixVersion [%s]"
                          % (dm_ifix_version, dmms_ifix_version, clm_ifix_version))
            qd.echo_info("CLMiFixVersion must be set as well!")
            qd.exit_(1)

    # now test for valid combinations
    # use these flags at the end to determine the final result
    clm_ifix_settings_valid = True
    dm_ifix_settings_valid = True
    dmms_ifix_settings_valid = True

    # calculate the application release eg 601 from CE601
    # extract the prefix and then the release
    app_media_root = env_props.get_values('appMediaRoot')
    app_media_content, app_release = media.get_media_content_and_app_release(app_media_root)
    if not app_media_content:
        qd.echo_error("Unsupported prefix in [%s] --" % app_media_root)
        qd.exit_(1)
    qd.echo_info("app_media_root [%s], app_media_content [%s], app_release [%s]"
                 % (app_media_root, app_media_content, app_release))

    # rule2. iFix must be on the same stream as the build
    # eg 601 iFix only installed along with a CE601 build
    qd.echo_info("now checking CLM iFix version compatability...")
    # iFix version and base version have to match
    clm_ifix_release, clm_ifix_patch = media.get_ifix_release_and_ifix_patch('CLM',
                                                                             clm_ifix_version,
                                                                             app_release)
    # only continue if [clm_ifix_release] matches [app_release]
    if clm_ifix_release == app_release:
        qd.echo_info("potentially installing CLM iFix as iFix-Release [%s] matches appRelease [%s]"
                     % (clm_ifix_release, app_release))
    else:
        qd.echo_error("Not installing CLM iFix as iFix-Release [%s] does not match appRelease [%s]"
                      % (clm_ifix_release, app_release))
        clm_ifix_settings_valid = False

    # rule4. DM or DMMS iFixes can be installed independently of each other
    if dm_ifix_version == 'NONE':
        qd.echo_info("although enabled no DM iFix is being installed as DMiFixVersion [%s],"
                     " so not checking DM iFix version compatability" % dm_ifix_version)
    else:
        qd.echo_info("now checking DM iFix version compatability")
        # rule2. iFix version and base version have to match
        dm_ifix_release, dm_ifix_patch = media.get_ifix_release_and_ifix_patch('DM',
                                                                               dm_ifix_version,
                                                                               app_release)
        # only continue if [dm_ifix_release] matches [app_release]
        if dm_ifix_release == app_release:
            qd.echo_info("potentially installing DM iFix as iFix-Release [%s] matches "
                         "appRelease [%s]" % (dm_ifix_release, app_release))

            # rule3: DM iFixY can only be installed alongside CLM iFix00Y
            # (that is all that DM test team test)
            if dm_ifix_patch == clm_ifix_patch:
                qd.echo_info("potentially installing DM iFix as DM-Patch [%s] matches "
                             "CLM-Patch [%s]" % (dm_ifix_patch, clm_ifix_patch))
            else:
                qd.echo_error("Not installing DM iFix as DM-Patch [%s]"
                              " does not match CLM-Patch [%s]" % (dm_ifix_patch, clm_ifix_patch))
                dm_ifix_settings_valid = False
        else:
            qd.echo_error("Not installing DM iFix as iFix-Release [%s]"
                          " does not match appRelease [%s]" % (dm_ifix_release, app_release))
            dm_ifix_settings_valid = False

    # rule4. DM or DMMS iFixes can be installed independently of each other
    if dmms_ifix_version == 'NONE':
        qd.echo_info("although enabled no DM iFix is being installed as DMMSiFixVersion [%s],"
                     " so not checking DMMS iFix version compatability" % dmms_ifix_version)
    else:
        qd.echo_info("now checking DMMS iFix version compatability")
        # rule2. iFix version and base version have to match
        # pylint: disable=line-too-long
        dmms_ifix_release, dmms_ifix_patch = media.get_ifix_release_and_ifix_patch('DMMS',
                                                                                   dmms_ifix_version,
                                                                                   app_release)
        # pylint: enable=line-too-long
        # only continue if [dmms_ifix_release] matches [app_release]
        if dmms_ifix_release == app_release:
            qd.echo_info("potentially installing DMMS iFix as iFix-Release [%s] matches "
                         "appRelease [%s]" % (dmms_ifix_release, app_release))

            # rule3: DMMS iFixY can only be installed alongside CLM iFix00Y
            # (that is all that DMMS test team test)
            if dmms_ifix_patch == clm_ifix_patch:
                qd.echo_info("potentially installing DMMS iFix as DMMS-Patch [%s] matches "
                             "CLM-Patch [%s]" % (dmms_ifix_patch, clm_ifix_patch))
            else:
                qd.echo_error("Not installing DMMS iFix as DMMS-Patch [%s]"
                              " does not match CLM-Patch [%s]" % (dmms_ifix_patch, clm_ifix_patch))
                dmms_ifix_settings_valid = False
        else:
            qd.echo_error("Not installing DMMS iFix as iFix-Release [%s]"
                          " does not match appRelease [%s]" % (dmms_ifix_release, app_release))
            dmms_ifix_settings_valid = False

    # finally determine how to exit method
    qd.echo_info("CLMiFixSettingsValid [%s], DMiFixSettingsValid [%s], DMMSiFixSettingsValid [%s]"
                 % (clm_ifix_settings_valid, dm_ifix_settings_valid, dmms_ifix_settings_valid))
    if clm_ifix_settings_valid and dm_ifix_settings_valid and dmms_ifix_settings_valid:
        qd.echo_info("checks finished, potentially installing iFixes")
    else:
        qd.echo_error("One or more checks failed, not installing iFixes")
        qd.exit_(1)


def get_file_stat(target_file):
    '''
    Return file permission

    :param target_file: target file
    :return: file permissions for target file
    '''
    check_result = qd.check_output_strip(['stat', '-c', '%a', target_file])
    return check_result


# usage checkAndCreateLinkFolder tempLinkDir tempLink tempTargetDir
def _check_and_create_link_folder(link_dir, link, target_dir):
    '''
    Check link to target folder; create if it does not exist.

    :param link_dir: directory containing the link(s)
    :param link: link to be checked (or created)
    :param target_dir: target directory (to be linked to)
    '''

    if qd.is_linux():
        if os.path.islink(link):
            if qd.check_output_strip(['readlink', link]) == target_dir:
                qd.echo_info('link [%s] to the folder [%s] already exists' % (link, target_dir))
                check_result = get_file_stat(target_dir)
                if int(check_result) == 777:
                    qd.echo_info("[%s] has correct permission" % target_dir)
                else:
                    qd.echo_error("[%s] permission is set to %s" % (target_dir, check_result))
                    qd.exit_(1)
            else:
                qd.echo_error('Link [%s] exists but not to folder [%s]'
                              % (link, target_dir))
                qd.exit_(1)
        else:
            qd.echo_info("link from [%s] to [%s]" % (link, target_dir))
            if not os.path.exists(target_dir):
                os.makedirs(target_dir)
            qd.check_call(['ln', '-s', target_dir, link])
            qd.chmod_777(target_dir)
            # check link created
            if os.path.islink(link) and qd.check_output_strip(['readlink', link]) == target_dir:
                qd.echo_info('link created')
            else:
                qd.echo_error("Failed to create link")
                qd.exit_(1)

        qd.echo_info('list the folder with the links [%s]' % link_dir)
        qd.call(['ls', '-la', link_dir])
        qd.echo_info('list the target folder [%s]' % target_dir)
        qd.call(['ls', '-la', target_dir])

    elif qd.is_windows():
        # create link_dir\link in Windows and ensure the permission
        prod_dir = qd.path_join(link_dir, link)
        qd.echo_warning('Linking folder in Window is not supported in this version of '
                        'UCD. Just use a normal folder [%s] and the link target [%s] is ignored.'
                        % (prod_dir, target_dir))
        qd.check_and_create_folder(prod_dir)
        qd.chmod_777(prod_dir)

    else:
        assert False, 'Unsupported OS'


def process_link_dirs(env_props, thost):
    '''
    Create the historical links between "non IBMQD" and /IBMQD.  Links are created depending upon
    the TAGS on the server.

    .. warning::
       removed the link from /opt/IBM/xx to /IBMQD/xx
       where xx is {InstallationManager,IMShared,WebSphere,HttpServer,Plugins,JazzTeamServer}

    :param env_props: environment properties
    :param thost: server
    '''

    # get ucd_root_dir
    ucd_root_dir = env_props.get_values(qd_props.PN_QD_ROOT)
    tag_list = thost.tags[:]  # env_props.get_values('localTags')

    qd.echo_info("Ensure that the database links are in place....")
    qd.echo_info("localTagList [%s]" % ','.join(tag_list))

    if qd_topo.TAG_DB in thost.tags and env_props.get_values('DB_VENDOR') == "db2":
        # need a second link [/db2fs] also in "/" folder
        db2fs_folder = qd.path_join(ucd_root_dir, qd_props.DEFAULT_REL_DB2FS_DIR)
        qd.check_and_create_folder(db2fs_folder)
        qd.chmod_777(db2fs_folder)
        # need a folder to hold CLM data warehouse folder
        dw_folder = qd.path_join(ucd_root_dir, 'DWFolder')
        qd.check_and_create_folder(dw_folder)
        qd.chmod_777(dw_folder)

    if qd_topo.TAG_DB in thost.tags and env_props.get_values('DB_VENDOR') == "oracle":
        # link will appear in "/" folder
        db_storage_folder = qd.path_join(ucd_root_dir, qd_props.DEFAULT_REL_ORACLE_STORAGE_DIR)
        qd.check_and_create_folder(db_storage_folder)
        qd.chmod_777(db_storage_folder)
        # need a folder to hold CLM data warehouse folder
        dw_folder = qd.path_join(ucd_root_dir, 'DWFolder')
        qd.check_and_create_folder(dw_folder)
        qd.chmod_777(dw_folder)


def check_rh_packages_installed(thost):
    '''
    Check RH preReq packages, currently uses a shell script [CoreAutomationPrep_packages.sh].
    Only to be used on Linux.

    :param thost: server
    '''

    qd.echo_info("--- check RH preReq packages are installed ---")
    assert qd.is_linux()
    check_pre_req_xvfb_flag = (qd_topo.TAG_RM in thost.tags or qd_topo.TAG_DM in thost.tags)
    check_pre_req_dm_libs_flag = (qd_topo.TAG_DM in thost.tags)
    qd.echo_info('Delegating to CoreAutomationPrep_package.sh for cheking RedHat packages...')
    qd.check_call(['sh', 'CoreAutomationPrep_packages.sh',
                   'TRUE' if check_pre_req_xvfb_flag else 'FALSE',
                   'TRUE' if check_pre_req_dm_libs_flag else 'FALSE'])


def ensure_properties_files(env_props, thost):
    '''
    - Ensure LDAP, database, and user properties file exist.
    - Ensure DB2 username meets the requirements.

    :param env_props: environment properties
    :param thost: server
    '''

    # ensure custom LDAP properties have been set
    # Only check on WAS nodes
    if qd_topo.TAG_WAS in thost.tags:
        if env_props.get_values('LDAPPROP', LDAPPROP=None) == 'CustomLDAP':
            qd.echo_info("check custom LDAP properties have been set")
            # ensure that the property file exists
            ldap_props = qd_props.get_custom_ldap_props(env_props)
            qd.check_file_exists(ldap_props.filename)

    # ensure user credentials have been set
    qd.echo_info("check user credentials have been set")
    user_props = qd_props.get_user_props(env_props)
    qd.check_file_exists(user_props.filename)
    qd.chmod_777(user_props.filename, mode='600')

    # ensure Database creation properties have been set
    # - check on all servers, but only if properties are required
    db_props = qd_props.get_database_props(env_props)
    if db_props is not None:
        qd.check_file_exists(db_props.filename)

        # ensure DB2 user name meets the requirements
        if env_props.get_values('DB_VENDOR') == "db2":
            ibm_qse_inst_user, ibm_qse_inst_group = user_props.get_values('ibmQseInst_User',
                                                                          'ibmQseInst_Group')
            if check_db2_id(ibm_qse_inst_user):
                qd.echo_info('Passed test for id [%s]' % ibm_qse_inst_user)
            else:
                qd.echo_error("Failed test for id [%s]"
                              % ibm_qse_inst_user)
                qd.exit_(1)
            if check_db2_id(ibm_qse_inst_group):
                qd.echo_info('Passed test for group name [%s]' % ibm_qse_inst_group)
            else:
                qd.echo_error("Failed test for group name [%s]"
                              % ibm_qse_inst_group)
                qd.exit_(1)


def ensure_media_properties_set(env_props, thost):
    # Disable pylint unused-argument to keep method consistent
    # pylint: disable=unused-argument
    '''
    - Ensure the media properties file
    - Download IM credentials if necessary

    :param env_props: environment properties
    :param thost: server
    '''

    # ensure media properties have been set
    auto_prep = env_props.get_values('AUTO_PREP')
    media_props = qd_props.get_media_props(env_props, error_if_not_exist=False)
    if os.path.isfile(media_props.filename):
        qd.echo_info("media properties set")
        return

    media_properties_type = env_props.get_values('mediaPropertiesType')
    media_filename = {'pipeline': 'MediaParameters/media.pipeline',
                      'ftp': 'MediaParameters/media.ftp',
                      'local': 'MediaParameters/media.local'}.get(media_properties_type)
    if media_filename:
        qd.echo_info("using %s media properties" % media_properties_type)
        media_file_src = qd.path_join(auto_prep, media_filename)
        qd.check_file_exists(media_file_src)
        shutil.copy(media_file_src, media_props.filename)
    else:
        qd.echo_error('Script [{scriptname}] does not support [%s] media properties'
                      % media_properties_type)
        qd.exit_(1)

    # check_file_exists
    ls_cmd = qd.os_switch('ls -la', 'dir') + " " + qd.dirname(media_props.filename)
    qd.check_call(ls_cmd, shell=True)
    qd.check_file_exists(media_props.filename)

    # Append the media version properties file to the media properties file
    qd.echo_info("Appending mediaVersion.properties to media.properties...")
    media_version_file_src = qd.path_join(auto_prep, "MediaParameters/mediaVersion.properties")
    qd.check_file_exists(media_version_file_src)
    # open the file for reading (r=read, t=text)
    # read contents of media version properties file
    with open(media_version_file_src, 'rt') as media_version_properties:
        data = media_version_properties.read()
    data = '\n' + data
    # append media version properties  to media properties file (a=append, t=text)
    with open(media_props.filename, 'at') as media_properties:
        media_properties.write(data)
    # confirm file exists
    qd.check_call(ls_cmd, shell=True)
    qd.check_file_exists(media_props.filename)


def ensure_umask_set():
    '''
    Check the root user umask setting
    '''
    umask_directory_bit = qd.check_output("umask|cut -c1", shell=True)
    umask_user_bit = qd.check_output("umask|cut -c2", shell=True)
    umask_group_bit = qd.check_output("umask|cut -c3", shell=True)
    umask_other_bit = qd.check_output("umask|cut -c4", shell=True)
    if int(umask_directory_bit) == 0:
        qd.echo_info("umask for directory bit is set correct [%s]" % umask_directory_bit)
    else:
        qd.echo_error("umask directory bit must be 0, current setting is [%s]"
                      % umask_directory_bit)
        qd.exit_(1)
    if int(umask_user_bit) < 2:
        qd.echo_info("umask for user bit is set correct [%s]" % umask_user_bit)
    else:
        qd.echo_error("umask user bit must be 0 or 1, current setting is [%s]"
                      % umask_user_bit)
        qd.exit_(1)
    if int(umask_group_bit) < 3:
        qd.echo_info("umask for group bit is set correct [%s]" % umask_group_bit)
    else:
        qd.echo_error("umask group bit must be 0,1 or 2, current setting is [%s]"
                      % umask_group_bit)
        qd.exit_(1)
    if int(umask_other_bit) < 3:
        qd.echo_info("umask for other bit is set correct [%s]" % umask_other_bit)
    else:
        qd.echo_error("umask other bit must be 0 or 1, current setting is [%s]"
                      % umask_other_bit)
        qd.exit_(1)


def ensure_home_dir_permission():
    '''
    Check /home directory permission is valid
    '''
    permission_list = ['755', '757', '775', '777']
    ensure_dir_permission("/home", permission_list)


def ensure_qdroot_dir_permission(qdroot_dir):
    '''
    Check [QD_ROOT] directory permission (likely to be /IBMQD) is valid.  LINUX ONLY.

    :param qdroot_dir: full path to the [QD_ROOT] directory
    '''
    permission_list = ['775', '777']
    ensure_dir_permission(qdroot_dir, permission_list)


def ensure_dir_permission(temp_dir, permission_list):
    '''
    Check directory permission against a permission list.  LINUX ONLY.

    :param temp_dir: full path to the directory to check
    :param permission_list: list of valid permissions, as defined by chmod eg 755
    :return: nothing returned, but success is logged and failure results in hard exit

    NOTE: if there is no match with the [current permissions] and the [permission_list]
    then the test will error and exit.
    '''
    directory_permission = get_file_stat(temp_dir)

    if directory_permission in permission_list:
        qd.echo_info("[%s] has correct permissions set [%s]"
                     % (temp_dir, directory_permission))
    else:
        qd.echo_error("[%s] has wrong permissions set, "
                      "current setting is [%s], expected one from [%s]"
                      % (temp_dir, directory_permission, permission_list))
        qd.exit_(1)


def download_pipeline_media_credentials(media_props):
    # Disable pylint invalid-name because function name is > 30 chars and want to keep it
    # pylint: disable=invalid-name
    '''
    Download [IM.Cred and password.txt]

    :param media_props: media properties
    '''

    # other credentials that change every 90 days
    qd.echo_info("get pipeline media credentials")

    # get the props
    AUTO_PREP, mediaServer, imRepoCredRoot = media_props.get_values('AUTO_PREP', 'mediaServer',
                                                                    'imRepoCredRoot')

    # create download folder
    keyring_dir = qd.path_join(AUTO_PREP, 'RATIONAL_SVT_IM_KEYRING')
    if not os.path.isdir(keyring_dir):
        os.makedirs(keyring_dir)

    # download
    def download_cred(filename):
        '''
         download file from media server (url)
        '''

        url = mediaServer + '/' + imRepoCredRoot + '/' + filename
        downloaded_file = media.do_wget(url, keyring_dir, None, media_props)
        if downloaded_file:
            qd.echo_info("file now exists, downloaded %s successful" % filename)
            return downloaded_file
        else:
            qd.echo_error("File [%s] does not exist, download failed" % filename)
            qd.echo_info("Installation Aborted....")
            qd.exit_(1)

    IM_Cred_tmp = download_cred('IM.Cred')
    password_txt_tmp = download_cred('password.txt')

    # move credentials
    # ensure the real credentials folder exists
    finalCredentialsDir = media.get_im_keyring_file_location()
    qd.echo_info("Move credentials files to [%s]" % finalCredentialsDir)
    qd.check_and_create_folder(finalCredentialsDir)
    qd.move_file(IM_Cred_tmp, os.path.join(finalCredentialsDir, qd.basename(IM_Cred_tmp)))
    qd.move_file(password_txt_tmp, os.path.join(finalCredentialsDir, qd.basename(password_txt_tmp)))
    qd.chmod_777(qd.path_join(finalCredentialsDir, '*'))


def check_db2_id(db2_id):
    '''
    Check that the DB2 id (user or group) meets the following rules:

    1. Must be
        * Windows: any-case letters (A-Za-z), numbers (0-9), and the underscore character(_)
        * Linux : lowercase letters (a-z), numbers (0-9), and the underscore character(_)
    2. Must not start with a number or underscore
    3. Cannot be longer than
        * Windows: 30 characters
        * Linux : 8 characters
    4. Cannot begin with IBM, SYS, SQL, or a number
    5. Can not be USERS, ADMINS, GUESTS, PUBLIC, LOCAL

    :param db2_id: db2 administrator user ID. e.g. db2inst1/db2admin
    '''
    qd.echo_info("checking db2_id [%s]" % db2_id)
    # use  flag at the end to determine the final result
    db2_id_valid = True

    # rule 1. Must be lowercase letters (a-z), numbers (0-9), and the underscore character(_)
    # linux : Must be lowercase letters (a-z), numbers (0-9), and the underscore character(_)
    # windows : Must be anycase letters (A-Za-z), numbers (0-9), and the underscore character(_)
    # re_obj = re.compile('[^a-z0-9_]')
    # if re_obj.search(db2_id):
    if qd.is_linux():
        if re.search('[^a-z0-9_]', db2_id):
            qd.echo_error("db2_id [%s]"
                          " contains characters outside range [a-z0-9_] --" % db2_id)
            db2_id_valid = False
        else:
            qd.echo_info("passed only contains characters in range [a-z0-9_] test")
    else:
        if re.search('[^A-Za-z0-9_]', db2_id):
            qd.echo_error("db2_id [%s]"
                          " contains characters outside range [A-Za-z0-9_] --" % db2_id)
            db2_id_valid = False
        else:
            qd.echo_info("passed only contains characters in range [A-Za-z0-9_] test")

    # rule 2. Must not start with a number or underscore
    if re.match('[0-9_]', db2_id):
        qd.echo_error("db2_id [%s]"
                      " must not start with a number or underscore [0-9_] --" % db2_id)
        db2_id_valid = False
    else:
        qd.echo_info("passed does not start with a number or underscore [0-9_] test")

    # rule 3. Cannot be longer than eight characters (linux), 30 characters (windows)
    if qd.is_linux():
        test_length = 8
    else:
        test_length = 30
    if len(db2_id) > test_length:
        qd.echo_error("db2_id [%s]"
                      " cannot be longer than [%s] characters --" % (db2_id, test_length))
        db2_id_valid = False
    else:
        qd.echo_info("passed valid length test")

    # rule 4. Cannot begin with IBM, SYS, SQL, or a number
    # (number already tested in rule 2
    pattern_list = ['IBM', 'SYS', 'SQL']
    for pattern in pattern_list:
        if re.match(pattern, db2_id):
            qd.echo_error("db2_id [%s] must not start with %s --"
                          % (db2_id, str(pattern_list)))
            db2_id_valid = False
    # else:
    qd.echo_info("passed does not start with %s test" % str(pattern_list))

    # rule 5. Can not be USERS, ADMINS, GUESTS, PUBLIC, LOCAL
    reserved_keywords = ['USERS', 'ADMINS', 'GUESTS', 'PUBLIC', 'LOCAL']
    if db2_id in reserved_keywords:
        # reserved_keywords = 'USERS, ADMINS, GUESTS, PUBLIC, LOCAL'
        qd.echo_error("db2_id [%s] cannot be a reserved keyword %s --"
                      % (db2_id, str(reserved_keywords)))
        db2_id_valid = False
    else:
        qd.echo_info("passed not a keyword %s test" % str(reserved_keywords))

    return db2_id_valid


def get_os_user_primary_group(os_user):
    '''
    If Linux, get the primary user group for the given user, otherwise return None.

    :param os_user: user whose primary group we want
    :return: if Linux the primary group, otherwise None
    '''
    if qd.is_linux():
        # need to get the primary group and not assume it is os_user_group
        id_output = qd.check_output("id " + os_user, shell=True)
        # process the return
        # eg uid=1001(clmadmin) gid=1001(clmadmin) groups=1001(clmadmin),100(users)
        groups = id_output.split(' ')[2]
        primary_group = groups.split('(')[1].split(')')[0]
    else:
        qd.echo_warning("Windows operating system, so no primary group")
        primary_group = None

    return primary_group


def create_new_os_user(user_props):
    '''
    Using the entries held in {user_props} create the user OSUserName,
    with password OSUserPassword and ensure it is part of group OSUserGroup.
    Once created ensure the users's primary group is then stored as OSUserPrimaryGroup.

    :param user_props: user properties
    '''

    # Create local user for the installation
    user_name, user_password, user_group = qd_props.get_new_os_users_creds_and_grp(user_props)
    install_util.ensure_os_user_and_group(user_name, user_password, user_group)
    # need to get the primary group and not assume it is os_user_group
    primary_group = get_os_user_primary_group(user_name)
    if primary_group:
        qd.echo_info("returned primary_group [%s]" % primary_group)
    else:
        qd.echo_info("cannot determine the primary group, so use OSUserGroup [%s]" % user_group)
        primary_group = user_group

    # update {user_props} for other to use later
    qd_props.set_new_os_users_primary_group(user_props, primary_group)


def main():
    '''Main function; see module documentation.'''
    # load env.properties
    qd.chmod_777(qd_props.DEFAULT_ENV_PROPS_LOCATION)
    env_props = qd_props.get_default_env_props()
    topo = qd_topo.QDTopo(env_props.get_values(qd_props.UCD_PN_GLOBAL_TAG_LISTING))
    thost = topo.get_host(env_props.get_values('localHostname', localHostname=qd.get_host_name()))

    # set AUTO_PREP
    working_dir = os.getcwdu()
    env_props.set_value(qd_props.PN_AUTO_PREP, working_dir)

    # ensure OSUser is created
    user_props = qd_props.get_user_props(env_props)
    create_new_os_user(user_props)

    # ensure the UCD ROOT DIR exists, it will either have been created prior to running QD
    # or will have been created by the UCD process itself
    qd.check_folder_exists(env_props.get_values(qd_props.PN_QD_ROOT))

    # ensure umask setting is correct level
    # The user created file should be readable and executable by everyone
    if qd.is_linux():
        ensure_umask_set()
        ensure_home_dir_permission()
        ensure_qdroot_dir_permission(env_props.get_values(qd_props.PN_QD_ROOT))

    # ensure domain name exists
    # SSO configuration requires to have the domain name
    domain_fqn = socket.getfqdn()
    name_list = domain_fqn.split(".", 1)
    if len(name_list) < 2:
        qd.echo_error('Host does not have a domain name [%s]'
                      % domain_fqn)
        qd.exit_(1)

    # ensure that the links are in place
    # to keep the rest of the install scripts working
    # for windows, ensure that the target folders are in place
    process_link_dirs(env_props, thost)

    # Check if iFix is enabled that iFix version and base version match
    check_ifix_version(env_props)

    # Check RH preReq packages are installed
    if qd.is_linux():
        check_rh_packages_installed(thost)

    # Ensure properties files
    ensure_properties_files(env_props, thost)

    # Ensure media properties have been set, also download IM credentials for pipeline
    ensure_media_properties_set(env_props, thost)

    # source media properties
    media_props = qd_props.get_media_props(env_props)
    media_properties_type = env_props.get_values('mediaPropertiesType')

    # download pipeline credentials
    # if (media_properties_type == 'pipeline') or (media_properties_type == 'ftp'):
    if media_properties_type == 'pipeline':
        download_pipeline_media_credentials(media_props)

    ################################################
    # any other checks go in here
    ################################################

    if qd.is_linux():
        qd.check_call(['sh', 'config_xvfb.sh', '/etc'])


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main()
    exit(0)
