#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2016, 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Touch (properties) file.  See :func:`main` function.

Example::

    touch_props.py prop_filename [-e][-s][PROP_NAME=VALUE]*
    touch_props.py filename

Originally used to make sure a properties file existed and optionally add new properties
information to it.  Can also be used to 'touch' any file.
'''
import os
import sys

# find qdlib from parent paths and add the first one to the Python search path
# pylint: disable=line-too-long
sys.path.append(next(p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and (p + '/qdlib') not in sys.path))
# pylint: enable=line-too-long
# pylint: disable=wrong-import-position
from qdlib import qd
# pylint: enable=wrong-import-position


def main(argv):
    '''
    Main code for touching (properties) files.

    :param prop_filename: file to touch and/or update
    :param -e: (optional) -e=Encode, encode the property values if specified (affects all values)
    :param -s: (optional) -s=Skip empty, property=value pairs will not be added to the file
                          if the value is empty
    :param property=value: (optional) multiple property=value pairs, indicating
                           which properties should be added/modified in the file

    .. note::
      * passing in a property that already exists will overwrite the existing value
      * using -s just makes sure that empty values are not written to the properties file; it
        cannot be used to remove a property.  To remove a property's value, you'd have to pass in
        propertyToRemove=, without -s.  That would put the property in the file, but with no value.
    '''

    # check inputs
    qd.process_inputs(argv[:3], 'file to update,[-e=Encode],[-s=Skip empty]', 2, False)

    env_file, argv = argv[1], argv[2:]
    b_encode, argv = (True, argv[1:]) if argv[:1] == ['-e'] else (False, argv)
    b_skip_empty, argv = (True, argv[1:]) if argv[:1] == ['-s'] else (False, argv)
    props = argv

    # Normally, let process_inputs to output the info about each argument.  But with
    # optional arguments, the output gets more difficult.  Do it here, now that the
    # values are set.
    qd.echo_info('%s: %s' % ("file to update", env_file))
    qd.echo_info('%s: %s' % ("[-e=Encode]", b_encode))
    qd.echo_info('%s: %s' % ("[-s=Skip empty]", b_skip_empty))

    # create file if none
    if not os.path.exists(env_file):
        qd.echo_info('Property file [%s] does not exist. Creating...' % env_file)
        with open(env_file, 'wt'):
            pass  # nothing to add
    else:
        qd.echo_info('Property file [%s] does exist.' % env_file)

    for prop_item in props:
        # The following line will raise a ValueError if = is not included
        property_name, property_value = prop_item.split('=', 1)

        # Validate that there is a property_name; raise ValueError on e.g., =b
        if not property_name:
            raise ValueError("No property name in argument '%s'" % prop_item)

        if b_skip_empty and not property_value:
            qd.echo_info('  Skipping property [%s] because value is empty...' % property_name)
            continue

        # encode value if required
        if b_encode:
            enc_val = qd.encode_base64(property_value)
        else:
            enc_val = property_value
        qd.echo_info('  Adding property [%s]: [%s]...' % (property_name, enc_val))
        qd.QDEnv.set_value_to_prop_file(env_file, property_name, enc_val)


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main(sys.argv)
    exit(0)
