#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Deploy Jazz applications to WebSphere application server. See :func:`main` function.
'''
import imp
import os
import sys

# find qdlib from parent paths and add the first one to the Python search path
# pylint: disable=line-too-long
sys.path.append(next(p for p in ['..', '../..', '../../..'] if os.path.exists(p + '/qdlib') and (p + '/qdlib') not in sys.path))
# pylint: enable=line-too-long

from qdlib import qd, qd_props, install_util


def main():
    '''
    Deploy Jazz applications to WebSphere application server.

    1. Update LQE-specific JVM properties
    2. Call install_applications.jy to install each application into WebSphere
    3. Restart the WebSphere application server process
    '''
    # pylint: disable=too-many-locals
    working_dir = os.getcwdu()
    env_props = qd_props.get_default_env_props()
    user_props = qd_props.get_user_props(env_props)
    os_user_name, os_user_password = qd_props.get_os_users_creds(user_props)
    was_common_dir = env_props.get_values(qd_props.PN_WAS_COMMON_DIR)

    clm_property_file = qd.path_join(was_common_dir, 'was.properties')
    ldap_property_file = qd.path_join(was_common_dir, qd_props.DEFAULT_LDAP_REL_PATH)
    qd.check_file_exists(clm_property_file)
    qd.check_file_exists(ldap_property_file)

    apps = getattr(env_props.get_values(qd_props.PN_APPS), 'split')(',')

    for app in apps:
        timer = qd.start_clock('Installing App [%s]' % app)
        if app == 'lqe':
            qd.echo_info("Add LQE specific JVM properties to WAS appserver.....")
            lqe_was_property = qd.path_join(working_dir, 'lqewas.properties')
            qd.check_file_exists(lqe_was_property)
            # we need convert "\\" to "/" in the lqewas.properties,
            # since the path passed into WASadmin need to be c:/IBMQD/JazzTeamServer
            # str(jazz_home) is required to go from the tuple to a string!
            jazz_home = str(env_props.get_values(qd_props.PN_CLM_HOME_PREP))
            replacement_list = [('_CLM_HOME_', jazz_home.replace('\\', '/'))]
            qd.sed_in_place(lqe_was_property, replacement_list)
            cmd = [qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                                qd_props.DEFAULT_REL_WAS_ROOT, 'bin',
                                qd_props.DEFAULT_WSADMIN_EXE), '-lang', 'jython', '-f',
                   qd.path_join(working_dir, 'update_jvm_lqe.jy'), lqe_was_property]
            install_util.su_call_if_necessary('--', (os_user_name, os_user_password),
                                              cmd, call_method=qd.check_call)
            qd.echo_info("LQE specific JVM properties update succeeded")

        # Installing each application to WAS
        qd.echo_info("Install Application [%s] to WAS....." % app)
        cmd = [qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                            qd_props.DEFAULT_REL_WAS_ROOT, 'bin',
                            qd_props.DEFAULT_WSADMIN_EXE), '-lang', 'jython', '-f',
               qd.path_join(working_dir, 'install_application.jy'),
               app, clm_property_file, ldap_property_file]
        install_util.su_call_if_necessary('--', (os_user_name, os_user_password),
                                          cmd, call_method=qd.check_call)
        qd.echo_info("Successfully installed application [%s] to WAS application server" % app)
        qd.stop_clock(timer)

    # dynamic load the module
    was_restart = imp.load_source('single_was_start_stop',
                                  qd.path_join(was_common_dir, 'single_was_start_stop.py'))
    was_restart.main(['restart'])


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main()
    exit(0)
