#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Execute repotool script to create database tables.  See :func:`main` function.
'''
import imp
import os
import sys

# find qdlib from parent paths and add the first one to the Python search path
# pylint: disable=line-too-long
sys.path.append(next(p for p in ['..', '../..', '../../..'] if os.path.exists(p + '/qdlib') and (p + '/qdlib') not in sys.path))
# pylint: enable=line-too-long

from qdlib import qd, qd_props, install_util


def change_owner(file_path):
    '''
    Change the file or folder's owner id and group id to be the same as the parent folder.

    :param file_path: absolute file path to be modified
    '''
    if qd.is_linux():
        # when move to a new file, make sure the file owner is the same as the folder
        # The temporary file could be owned by root, and make duplicated file owner to
        # be root that could cause JTS setup problem.
        folder_st = os.stat(qd.dirname(file_path))
        '''
        we have to disable pylint check for os.chown, the code worked on target machine
        however failed pylint and pydev code check
        pylint: complained ID:no-member Module 'os' has no 'chown' memeber
        pydev: complained Undefined variable from import: chown
        So both pylint and pydev format disable are required in following code
        '''
        # pylint: disable=maybe-no-member
        os.chown(file_path, folder_st.st_uid, folder_st.st_gid)  # @UndefinedVariable


def set_db_driver(env_props):
    '''
    Copy the db driver file to the target folder

    :param env_props: env.properties
    :return dictionary {'ORACLE_JDBC_DRIVER_FILE'\\: db_driver_dest_file_path}:
        * **ORACLE_JDBC_DRIVER_FILE**: The property name of JDBC driver
        * **db_driver_dest_file_path**: The path for JDBC driver file name
    '''
    jazz_home = env_props.get_values(qd_props.PN_CLM_HOME_PREP)
    db_driver_file_name = env_props.get_values(qd_props.PN_DB_DRIVER_FILE)
    db_driver_dest_folder = qd.path_join(jazz_home, 'server', 'oracle')
    db_driver_dest_file_path = qd.path_join(db_driver_dest_folder, db_driver_file_name)
    qd.check_and_create_folder(db_driver_dest_folder)
    change_owner(db_driver_dest_folder)
    qd.copy_file(qd.path_join(env_props.get_values(qd_props.PN_INTERNAL_MEDIA_PROPS),
                              'drivers', db_driver_file_name), db_driver_dest_folder)
    change_owner(db_driver_dest_file_path)
    return {'ORACLE_JDBC_DRIVER_FILE': db_driver_dest_file_path}


def run_repotool(env_props, app, env_dict, os_user_name, os_user_password):
    '''
    Execute repotool based on the application

    :param env_props: env.properties
    :param app: application name, such as jts, ccm, qm
    :param env_dict: environment dictionary with added jdbc driver
    :param os_user_name: the os user that should run the repotool command
    :param os_user_password: the os user password
    '''
    jazz_home = env_props.get_values(qd_props.PN_CLM_HOME_PREP)
    repotool_dir = qd.path_join(jazz_home, 'server')
    teamserver_path = qd.path_join(repotool_dir, 'conf', app, 'teamserver.properties')

    repotool_exe_prefix = ''.join(['repotools-', app])
    if qd.is_linux():
        repotool_exe = ''.join(['./', repotool_exe_prefix, '.sh'])
        repotool_cmd = [repotool_exe, '-createTables',
                        ''.join(['teamserver.properties=', teamserver_path]),
                        ''.join(['logFile=repotools_', app, '.log'])]
    else:
        repotool_exe = ''.join([repotool_exe_prefix, '.bat'])
        repotool_cmd = ['cmd', '/c', repotool_exe, '-createTables',
                        ''.join(['teamserver.properties=', teamserver_path]),
                        ''.join(['logFile=repotools_', app, '.log'])]

    install_util.su_call_if_necessary('--', (os_user_name, os_user_password), repotool_cmd,
                                      call_method=qd.check_call, cwd=repotool_dir, env=env_dict)
    if app in 'jts':
        if qd.is_linux():
            datawarehouse_cmd = [repotool_exe, '-createWarehouse',
                                 ''.join(['teamserver.properties=', teamserver_path]),
                                 ''.join(['logFile=repotools_dw.log'])]
        else:
            datawarehouse_cmd = ['cmd', '/c', repotool_exe, '-createWarehouse',
                                 ''.join(['teamserver.properties=', teamserver_path]),
                                 ''.join(['logFile=repotools_dw.log'])]
        install_util.su_call_if_necessary('--', (os_user_name, os_user_password),
                                          datawarehouse_cmd, call_method=qd.check_call,
                                          cwd=repotool_dir, env=env_dict)


def replace_teamserver_properties(env_props, app, need_db_connection_properties, replacement_list):
    '''
    Update teamserver.properties or dbconnection.properties.  Copy the updated properties file to
    production directory

    :param env_props: env.properties
    :param app: application
    :param need_db_connection_proeprties: the product that need update dbconneciton properties
    :param replacement_list: the substitution list based on the application
    '''
    do_not_skip = True
    need_general_template = ['jts', 'qm']
    need_specific_template = ['rm', 'ccm', 'dm', 'relm', 'gc', 'dcc']
    jazz_home = env_props.get_values(qd_props.PN_CLM_HOME_PREP)
    template_dir = qd.path_join(os.getcwdu(), 'AppConfigScripts')
    if app in need_specific_template:
        template_file = qd.path_join(template_dir, app + '_template_teamserver.properties')
        temp_file = ''.join(['temp_', app, '_teamserver.properties'])
    elif app in need_general_template:
        template_file = qd.path_join(template_dir, 'template_teamserver.properties')
        temp_file = ''.join(['temp_', app, '_teamserver.properties'])
    elif app in need_db_connection_properties:
        template_file = qd.path_join(template_dir, 'lqe_template_dbconnection.properties')
        temp_file = ''.join(['temp_', app, '_dbconnection.properties'])
        if app == 'lqe':
            set_lqe_properties(qd.path_join(jazz_home, 'server', 'conf', 'lqe', 'lqe.properties'))
    elif app in 'rs':
        qd.echo_info("[%s] does not require to run repo tool " % app)
        do_not_skip = False
    else:
        qd.exit_with_err_msg("currently do not support this component [%s]" % app)
    if do_not_skip:
        qd.duplicate_file(template_file, temp_file)
        set_new_properties(env_props, app, need_general_template + need_specific_template,
                           qd.path_join(template_dir, temp_file))
        qd.sed_in_place(qd.path_join(template_dir, temp_file), replacement_list)
        qd.copy_file(qd.path_join(template_dir, temp_file),
                     qd.path_join(jazz_home, 'server', 'conf', app))
        if app in need_specific_template or app in need_general_template:
            app_properties_file = 'teamserver.properties'
            # back up the original properties file
            qd.duplicate_file(qd.path_join(jazz_home, 'server', 'conf', app, app_properties_file),
                              'teamserver_backup.properties')
        else:
            app_properties_file = 'dbconnection.properties'
        # copy updated file to Jazz home
        qd.duplicate_file(qd.path_join(jazz_home, 'server', 'conf', app, temp_file),
                          app_properties_file)
        # python script run as root, the file is created as root
        # and we need to change owner as the folder owner
        change_owner(qd.path_join(jazz_home, 'server', 'conf', app, app_properties_file))
        qd.echo_info("Completed updating application [%s] properties file" % app)


def set_new_properties(env_props, app, ntp_list, temp_file):
    '''
    Add new properties to teamserver.properties file

    :param env_props: env.properties
    :param app: application name
    :param ntp_list: the application list that require to add ntpServer property
    :param cm_enable_list: the application list that require to add the activation key
    :param temp_file: a file name that all properties to be added
    '''
    need_cm_enabled = ['qm', 'rm', 'dm']
    if app in ntp_list:
        qd.QDEnv.set_value_to_prop_file(temp_file, 'com.ibm.team.repository.cluster.isInCluster',
                                        'false')
        if env_props.get_values('EnableNTPclock') == 'TRUE':
            # pylint: disable=line-too-long
            qd.QDEnv.set_value_to_prop_file(temp_file, 'com.ibm.team.repository.diagnostic.systemClockDiagnostic.ntpServerAddress', env_props.get_values('NTPclock'))
            # pylint: enable=line-too-long
    if env_props.get_values('EnableCM') == 'TRUE' and app in need_cm_enabled:
        qd.echo_info("Cm enable ------- app is [%s]" % app)
        qd.QDEnv.set_value_to_prop_file(temp_file, 'com.ibm.team.repository.vvc.activationKey',
                                        env_props.get_values('CMActivationKey'))


def set_lqe_properties(lqe_file):
    '''
    Add some additional lqe properties in Jazz_home/server/conf/lqe/lqe.properties

    :param env_props: env.properties
    :param lqe_file: Full path to the lqe.properties
    '''
    qd.QDEnv.set_value_to_prop_file(lqe_file, 'lqe.authtype.delegated', 'true')
    qd.QDEnv.set_value_to_prop_file(lqe_file, 'lqe.quadFilter.ignoreAccessControl', 'false')
    qd.QDEnv.set_value_to_prop_file(lqe_file, 'lqe.sparql.resultSetLimit', '1000')
    qd.QDEnv.set_value_to_prop_file(lqe_file, 'lqe.configMode', 'true')


def set_common_section(env_props, app):
    '''
    Common properties that all teamserver.properties needed

    :param env_props: env.properties
    :param app: application name
    :return list replacement_list:
        * **replacement_list**: The list include [('to be replaced string', 'replacement string')]
    '''
    replacement_list = [('_CLM_COMPONENT_', app)]
    replacement_list += [('_WEB_SERVER_HOSTNAME_', env_props.get_values(qd_props.PN_IHS_HOSTNAME))]
    return replacement_list


def set_ldap_section(env_props, app):
    '''
    Update all ldap related parameters

    :param env_props: env.properties
    :param app: application name
    :return list replacement_list:
        * **replacement_list**: The list include [('to be replaced string', 'replacement string')]
    '''
    need_ldap_section = ['jts', 'ccm', 'qm', 'dcc']
    if app in need_ldap_section:
        ldap_env = qd_props.get_was_ldap_props(env_props,
                                               env_props.get_values(qd_props.PN_WAS_COMMON_DIR))
        ldap_dict = ldap_env.get_dict()
        # pylint: disable=line-too-long
        replacement_list = [('BASE_GROUP_DN', ldap_dict['jazz_ldap_baseGroupDN'])]
        replacement_list += [('BASE_USER_DN', ldap_dict['jazz_ldap_baseUserDN'])]
        replacement_list += [('REGISTRY_LOCATION', 'ldap://' + ldap_dict['jazz_userrealm'])]
        replacement_list += [('REGISTRY_USERNAME', ldap_dict['jazz_userrole_RegistryUsername'])]
        replacement_list += [('REGISTRY_PASSWORD', qd.decode_base64(ldap_dict['jazz_userrole_RegistryPassword']))]
        replacement_list += [('GROUP_MAPPING', ldap_dict['jazz_ldap_groupMapping'])]
        replacement_list += [('MEMBERS_OF_GROUP', ldap_dict['jazz_ldap_membersOfGroup'])]
        replacement_list += [('FIND_GROUPS_QUERY', ldap_dict['jazz_ldap_findGroupsForUserQuery'])]
        replacement_list += [('USER_ATTRIBUTE_MAPPING', ldap_dict['jazz_ldap_userAttributesMapping'])]
        # pylint: enable=line-too-long
        if ldap_dict['jazz_ldap_vendor'] == 'MSAD':
            replacement_list += [('_FIND_USER_QUERY_PROP_NAME_', 'findUsersByUserIdQuery')]
        else:
            replacement_list += [('_FIND_USER_QUERY_PROP_NAME_', 'findGroupsForUserQuery')]
        return replacement_list
    return []


def set_db_section(env_props, app):
    '''
    Update database related properties

    :param env_props: env.properties
    :param app: application name
    :return list replacement_list:
        * **replacement_list**: The list include [('to be replaced string', 'replacement string')]

    '''
    need_db_dw_properties = ['jts', 'ccm', 'qm', 'rm', 'dcc', 'gc']
    dw_setup = True if app in need_db_dw_properties else False

    def add_addition_slashes(file_path):
        '''
            add additional_slashes for windows platform
        '''
        return file_path.replace('\\', '\\\\\\')

    db_vendor = env_props.get_values('DB_VENDOR')
    replacement_list = [('_DBVENDOR_', db_vendor)]
    user_props = qd_props.get_user_props(env_props)
    replacement_list += [('_MOD_DB_HOSTNAME_', env_props.get_values('DB_HOSTNAME'))]
    if db_vendor == 'db2':
        temp_db2inst_pwd = qd.decode_base64(user_props.get_values('ibmQseInst_PW'))
        replacement_list += [('#_DB2_CONNECTION_STRING_', '')]
        replacement_list += [('_MOD_DB_PORT_', env_props.get_values('DB2PORT'))]
        replacement_list += [('_MOD_PASSWORD_', temp_db2inst_pwd)]
        replacement_list += [('_MOD_ADMIN_USER_', user_props.get_values('ibmQseInst_User'))]
        replacement_list += [('_MOD_TABLE_NAME_', env_props.get_values(app + 'DBName'))]
        if dw_setup is True:
            replacement_list += [('_DW_DB_HOSTNAME_', env_props.get_values('DB_HOSTNAME'))]
            replacement_list += [('_DATAWAREHOUSEVENDOR_', 'db2')]
            replacement_list += [('#_DB2_DW_CONNECTION_STRING_', '')]
            replacement_list += [('_DW_DB_PORT_', env_props.get_values('DB2PORT'))]
            replacement_list += [('_DW_TABLE_NAME_', env_props.get_values('dwDBName'))]
            replacement_list += [('_DW_PASSWORD_', temp_db2inst_pwd)]
            replacement_list += [('_DW_ADMIN_USER_', user_props.get_values('ibmQseInst_User'))]
            # We need an add extra '\\' in the path, so that when substituted in
            # teamserver.properties windows path will have double '\\' in the file
            dw_folder_path = add_addition_slashes(qd.path_join(env_props.get_values('QD_Root'),
                                                               qd_props.DEFAULT_DW_FOLDER))
            replacement_list += [('_DW_BASE_FOLDER_', dw_folder_path)]
    elif db_vendor == 'oracle':
        db_env = qd_props.get_database_props(env_props)
        temp_oracle_users_pwd = qd.decode_base64(user_props.get_values('oracleDBUsersPassword'))
        replacement_list += [('#_ORACLE_CONNECTION_STRING_', '')]
        replacement_list += [('_MOD_DB_PORT_', db_env.get_values('oraclePort'))]
        replacement_list += [('_ORACLE_SERVICE_NAME_', db_env.get_values('oracleInstanceName'))]
        replacement_list += [('_MOD_ORACLE_DB_USER_', db_env.get_values('oracle' + app + 'User'))]
        replacement_list += [('_MOD_PASSWORD_', temp_oracle_users_pwd)]
        if dw_setup is True:
            replacement_list += [('_DW_DB_HOSTNAME_', env_props.get_values('DB_HOSTNAME'))]
            replacement_list += [('_DATAWAREHOUSEVENDOR_', 'oracle')]
            replacement_list += [('#_ORACLE_DW_CONNECTION_STRING_', '')]
            replacement_list += [('_DW_DB_PORT_', db_env.get_values('oraclePort'))]
            replacement_list += [('_DW_ORACLE_USER_', db_env.get_values('oracledwUser'))]
            replacement_list += [('_DW_PASSWORD_', temp_oracle_users_pwd)]
            # We need an add extra '\\' in the path, so that when substituted in
            # teamserver.properties windows path will have double '\\' in the file
            dw_folder_path = add_addition_slashes(qd.path_join(env_props.get_values('QD_Root'),
                                                               qd_props.DEFAULT_DW_FOLDER))
            replacement_list += [('_DW_BASE_FOLDER_', dw_folder_path)]
    else:
        qd.exit_with_err_msg("Unsupported database vendor [%s]" % db_vendor)
    return replacement_list


def main():
    '''
    Entry point to run repotool for all necessary applications

    1. Get the replacement list based on the application
    2. Update the teamserver.properties/dbconnection.properties using the replacement list
    3. Copy the update properties file to Jazz product install directory
    4. Run the repotool for applications
    '''
    env_props = qd_props.get_default_env_props()
    user_props = qd_props.get_user_props(env_props)
    os_user_name, os_user_password = qd_props.get_os_users_creds(user_props)
    all_apps = env_props.get_values(qd_props.PN_APPS)
    apps = getattr(all_apps, 'split')(',')

    # rs does not need to run repo tools, so no need to update teamserver.properties file
    need_ntp_property = ['jts', 'qm', 'rm', 'ccm', 'dm', 'relm', 'gc', 'dcc']
    need_db_connection_properties = ['lqe', 'ldx']

    # stop WAS before repotool run
    was_restart = imp.load_source('single_was_start_stop',
                                  qd.path_join(env_props.get_values(qd_props.PN_WAS_COMMON_DIR),
                                               'single_was_start_stop.py'))
    was_restart.main(['stop'])

    vendor = env_props.get_values(qd_props.PN_DB_VENDOR)
    env_dict = os.environ.copy()
    if vendor == 'oracle':
        env_dict.update(set_db_driver(env_props))
    elif vendor == 'db2':
        qd.echo_info("Do not add db driver path for vendor [%s]" % vendor)
    else:
        qd.exit_with_err_msg("Unsupported DB vendor [%s]" % vendor)

    for app in apps:
        replacement_list = []
        # Get replacement_list
        if app in need_ntp_property:
            replacement_list = set_db_section(env_props, app)
            replacement_list += set_common_section(env_props, app)
            replacement_list += set_ldap_section(env_props, app)
        elif app in need_db_connection_properties:
            qd.echo_info("Do not need to run repo tools for application [%s]" % app)
            replacement_list = set_db_section(env_props, app)
        elif app in 'rs':
            qd.echo_info("[%s] does not require to run repo tool " % app)
        else:
            qd.exit_with_err_msg("Unsupport application [%s] to run repotools" % app)
        if replacement_list:
            replace_teamserver_properties(env_props, app, need_db_connection_properties,
                                          replacement_list)

    for app in apps:
        if app in need_ntp_property:
            run_repotool(env_props, app, env_dict, os_user_name, os_user_password)
    # start websphere
    was_restart.main(['start'])


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main()
    exit(0)
