#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************

'''
Configure IBM HTTP server and plugin-cfg.xml files.  See :func:`main` function.
'''
import os
import shutil
import sys
import time

import build_plugin_file
import enable_ssl


# find qdlib from parent paths and add the first one to the Python search path
# pylint: disable=line-too-long
sys.path.append(next(p for p in ['..', '../..', '../../..'] if os.path.exists(p + '/qdlib') and (p + '/qdlib') not in sys.path))
# pylint: enable=line-too-long

from qdlib import qd, qd_props


def get_local_app_name(env_props, count):
    '''Get the application list based on the number from env.properties, NODE1_AppList

    :param env_props: environment properties
    :param count: node number
    :return: list of applications on that node'''
    return env_props.get_values(''.join(['NODE', str(count), '_AppList']))


def get_node_hostname(env_props, count):
    '''Get the host name based on the number from env.properties, NODE1_HostName

    :param env_props: environment properties
    :param count: node number
    :return: hostname of the node'''
    return env_props.get_values(''.join(['NODE', str(count), '_HostName']))


def copy_certificate_linux(env_props, count, qd_temp_dir, os_user_name):
    '''
    - Copy serverroot.cer from all WAS node to IHS node
    - Copy jtsLPTA key to IHS node

    :param env_props: environment properties
    :param count: node number
    :param qd_temp_dir: temporary directory
    :param os_user_name: user name
    '''
    node_host = get_node_hostname(env_props, count)
    node_app_list = get_local_app_name(env_props, count)
    # copy root certificate from each node to IHS
    copy_cmd = ['scp', ''.join([os_user_name, '@', node_host, ':',
                                qd.path_join(qd_temp_dir, 'serverroot.cer')]),
                qd.path_join(qd_temp_dir, ''.join(['serverroot_', str(count), '.cer']))]
    qd.check_call(copy_cmd)

    if 'jts' in node_app_list:
        qd.echo_info('copy jts LPTA key to IHS node')
        copy_cmd = ['scp', ''.join([os_user_name, '@', node_host, ':',
                                    qd.path_join(qd_temp_dir, 'jtsLTPAkey')]), qd_temp_dir]
        qd.check_call(copy_cmd)


def prepare_ihs_config_linux(env_props):
    '''
    Copy files between IHS node and WAS node.

    - rootserver.cer
    - jtsLTPAkey

    :param env_props: environment properties
    '''
    user_props = qd_props.get_user_props(env_props)
    os_user_name, os_user_password = qd_props.get_os_users_creds(user_props)
    qd_temp_dir = qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                               qd_props.DEFAULT_REL_TMP_DIR)
    total_nodes = int(env_props.get_values(qd_props.PN_CUSTOM_NODES))
    count = 1
    while count <= total_nodes:
        setup_ssh(get_node_hostname(env_props, count), os_user_name, os_user_password)
        copy_certificate_linux(env_props, count, qd_temp_dir, os_user_name)
        count = count + 1
    count = 1

    while count <= total_nodes:
        node_app_list = get_local_app_name(env_props, count)
        if 'jts' in node_app_list:
            pass
        else:
            node_host = get_node_hostname(env_props, count)
            copy_ltpa_key_linux(env_props, node_host, qd_temp_dir, os_user_name)
        count = count + 1


def setup_ssh(host_name, os_user, os_user_pwd):
    '''
    Setup remote access on Linux platform.

    :param host_name: the remote host name that needs set up ssh keyless entry
    :param os_user: the user that need access the remote host
    :param os_user_pwd: the user password
    '''
    ssh_script = qd.path_join(os.getcwdu(), 'sshScripts', 'setupSSH')
    qd.chmod_777(ssh_script, 'o+x')
    ssh_setup_cmd = [ssh_script, '-p', os_user_pwd, '-d', ''.join([os_user, '@', host_name])]
    qd.check_call(ssh_setup_cmd)


def copy_ltpa_key_linux(env_props, remote_server, qd_temp_dir, os_user_name):
    '''
    Copy ltpa key to the WAS node and import the key to WAS.

    :param env_props: environment property file instance
    :param remote_server: Copy key file to the remote server
    :param qd_temp_dir: The location where the key file will be stored
    :param os_user_name: the user that will ssh to remote server
    '''
    qd.echo_info('copy jtsLTPAkey to other WAS node')
    copy_cmd = ['scp', qd.path_join(qd_temp_dir, 'jtsLTPAkey'),
                ''.join([os_user_name, '@', remote_server, ':', qd_temp_dir])]
    qd.check_call(copy_cmd)
    # Get the WAS_COMMON_DIR path from the WAS node, since we do not know the value on IHS node
    get_was_dir_cmd = ['ssh', ''.join([os_user_name, '@', remote_server]),
                       'grep', '-m', '1', qd_props.PN_WAS_COMMON_DIR,
                       qd_props.DEFAULT_ENV_PROPS_LOCATION,
                       '|', 'cut', '-d', '=', '-f2']
    temp_was_dir = qd.check_output_strip(get_was_dir_cmd)

    # temporary turn on read mode for userCreds.properties
    user_props_path = qd.path_join(env_props.get_values(qd_props.PN_AUTO_PREP),
                                   qd_props.DEFAULT_USER_CREDS_REL_PATH)
    chmod_read = ['ssh', '-t', '-t', '-o', 'ConnectTimeout=300',
                  ''.join([os_user_name, '@', remote_server]),
                  'sudo', 'chmod', 'o+r', user_props_path]
    qd.check_call(chmod_read)

    # import jtsLTPAkey on other WAS node
    python_path = sys.executable
    import_cmd = ['ssh', '-t', '-t', '-o', 'ConnectTimeout=300',
                  ''.join([os_user_name, '@', remote_server]),
                  ''.join(['sudo -- sh -c ', '"cd ', ''.join([temp_was_dir, ';']),
                           ' ', python_path, ' enable_sso.py', ' import', '"'])]
    qd.check_call(import_cmd)

    # turn off the read mode for userCreds.properties
    chmod_revert = ['ssh', '-t', '-t', '-o', 'ConnectTimeout=300',
                    ''.join([os_user_name, '@', remote_server]),
                    'sudo', 'chmod', '600', user_props_path]
    qd.check_call(chmod_revert)


def main():
    '''
    Configure IBM HTTP server

    1. Create a directory on IHS node to hold all files. (QD_Root/qd_tmp)
    2. Exchange files between IHS node, WAS node (Linux only, windows is completed in UCD processes)
    3. Modify httpd.conf file
    4. Create IHS key store
    5. Create self-signed certificate in ihs key store
    6. Create SSL handshake
    7. Restart WebServer
    '''
    working_dir = os.getcwdu()
    env_props = qd_props.get_default_env_props()
    user_props = qd_props.get_user_props(env_props)

    total_nodes = int(env_props.get_values(qd_props.PN_CUSTOM_NODES))
    # get the qd_tmp directory for storing the root certificate files
    qd_temp_dir = qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                               qd_props.DEFAULT_REL_TMP_DIR)
    qd.check_and_create_folder(qd_temp_dir)

    if qd.is_linux():
        prepare_ihs_config_linux(env_props)

    qd.echo_info('Enable SSL from httpd.conf')
    enable_ssl.main()

    qd.echo_info('create ihsserverkey key store')
    ihs_keystore_pwd = qd.decode_base64(user_props.get_values(qd_props.PN_IHS_KEYSTORE_PASSWORD))
    http_root = qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                             qd_props.DEFAULT_REL_IHS_HOME)

    gsk_cmd = qd.path_join(http_root, 'bin', qd.os_switch('gskcmd', 'gskcmd.bat'))
    create_key_db = [gsk_cmd, '-keydb', '-create', '-db',
                     qd.path_join(http_root, 'ihsserverkey.kdb'),
                     '-pw', ihs_keystore_pwd, '-expire', '365', '-stash', '-type', 'cms']
    qd.check_call(create_key_db)

    qd.echo_info('Create self sign certificate')
    cert_cmd = [gsk_cmd, '-cert', '-create', '-db',
                qd.path_join(http_root, 'ihsserverkey.kdb'),
                '-label', 'ihsserver', '-expire', '365', '-dn',
                ''.join(['CN=', env_props.get_values(qd_props.PN_IHS_HOSTNAME)]),
                '-default_cert', 'yes', '-pw', ihs_keystore_pwd]
    qd.check_call(cert_cmd)

    qd.echo_info('List cerfiticate')
    ls_cert_cmd = [gsk_cmd, '-cert', '-list', '-db',
                   qd.path_join(http_root, 'ihsserverkey.kdb'), '-pw', ihs_keystore_pwd]
    qd.check_output(ls_cert_cmd)

    qd.echo_info('add server root certificates to plugin, SSL hand shake')
    count = 1
    while count <= total_nodes:
        add_cert_cmd = [gsk_cmd, '-cert', '-add', '-db',
                        qd.path_join(working_dir, 'webserver_template', 'plugin-key.kdb'),
                        '-label', ''.join(['server_root_', str(count)]), '-pw', 'WebAS', '-file',
                        qd.path_join(qd_temp_dir, ''.join(['serverroot_', str(count), '.cer']))]
        qd.check_call(add_cert_cmd)
        count = count + 1

    qd.echo_info("Merge Plugin-cfg.xml and put it in webservermerge directory")
    plugin_root = qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                               qd_props.DEFAULT_REL_PLUGINS_HOME)
    qd.check_and_create_folder(qd.path_join(plugin_root, 'logs', 'webservermerge'))
    shutil.copytree(qd.path_join(working_dir, 'webserver_template'),
                    qd.path_join(plugin_root, 'config', 'webservermerge'))

    qd.echo_info("Merge Plugin-cfg.xml and put it on webservermerge directory")
    build_plugin_file.main()

    qd.copy_file(qd.path_join(working_dir, 'plugin-cfg.xml'),
                 qd.path_join(plugin_root, 'config', 'webservermerge'))

    # restart webserver
    # restart option does not work properly on Linux platform,
    # IHS doesn't connect for the first time restart.
    # Only stop, wait, and start process worked
    # so we have differentiate restart option for windows and linux
    if qd.is_linux():
        qd.check_call([qd.path_join(http_root, 'bin', 'apachectl'), 'stop'])
        time.sleep(30)
        qd.check_call([qd.path_join(http_root, 'bin', 'apachectl'), 'start'])
    elif qd.is_windows():
        qd.check_call([qd.path_join(http_root, 'bin', 'apache.exe'), '-k', 'restart'])
    else:
        qd.exit_with_err_msg("Unsupported platform")


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main()
    exit(0)
