#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Set up WebSphere application server LDAP authentication.  See :func:`main` function.
'''
import os
import sys

import single_was_start_stop

# find qdlib from parent paths and add the first one to the Python search path
# pylint: disable=line-too-long
sys.path.append(next(p for p in ['..', '../..', '../../..'] if os.path.exists(p + '/qdlib') and (p + '/qdlib') not in sys.path))
# pylint: enable=line-too-long

from qdlib import qd, qd_props, install_util


def main():
    '''
    Main function to enable LDAP authentication.

    1. Update the LDAP property template file with correct username and password
    2. Call enable_LDAP.jy file to set LDAP authentication, test the connection
    3. Set security enablement = true.  Add WAS administrator username and password in
       soap.client.props.  Any future wsadmin call does not require to pass in username and
       password.
    4. Restart WebSphere to make sure authentication set up correctly
    '''
    env_props = qd_props.get_default_env_props()
    user_props = qd_props.get_user_props(env_props)

    os_user_name, os_user_password = qd_props.get_os_users_creds(user_props)
    # was_admin_password is decrypted
    was_admin_user, was_admin_password = qd_props.get_was_admin_users(user_props)
    default_absolute_ldap_file = qd.path_join(env_props.get_values(qd_props.PN_WAS_COMMON_DIR),
                                              qd_props.DEFAULT_LDAP_REL_PATH)
    default_ldap_dir = qd.dirname(default_absolute_ldap_file)
    default_ldap_file_name = qd.basename(default_absolute_ldap_file)

    # Move CustomLDAP property file to LDAP folder
    if env_props.get_values(qd_props.PN_LDAP_PROPS) == 'CustomLDAP':
        qd.check_and_create_folder(default_ldap_dir)
        source_ldap_template = qd.path_join(env_props.get_values(qd_props.PN_AUTO_PREP),
                                            'CustomLDAP', 'Custom_LDAPSecurity.properties')
        qd.check_file_exists(source_ldap_template)
        qd.move_file(source_ldap_template, qd.path_join(default_ldap_dir,
                                                        'CustomLDAP' + '_LDAPSecurity.properties'))

    # Move the specific LDAP template file to the default LDAP property file name
    ldap_props = env_props.get_values(qd_props.PN_LDAP_PROPS) + '_LDAPSecurity.properties'
    qd.check_file_exists(qd.path_join(default_ldap_dir, ldap_props))
    qd.duplicate_file(qd.path_join(default_ldap_dir, ldap_props), default_ldap_file_name)
    # Update the username and password in LDAPSecurity.properties file
    replacement_list = [('_LDAP_ADMIN_USER_', was_admin_user),
                        ('_LDAP_ADMIN_PASSWORD_',
                         user_props.get_values(qd_props.PN_USER_WAS_USER_PASSWORD))]
    qd.sed_in_place(default_absolute_ldap_file, replacement_list)

    qd.echo_info("enable LDAP and sync nodes")
    cmd = [qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                        qd_props.DEFAULT_REL_WAS_PROFILE, 'bin',
                        qd_props.DEFAULT_WSADMIN_EXE),
           '-lang', 'jython', '-f', qd.path_join(os.getcwdu(), 'enable_ldap.jy'),
           env_props.get_values(qd_props.PN_LDAP_PROPS), default_absolute_ldap_file]
    install_util.su_call_if_necessary('--', (os_user_name, os_user_password),
                                      cmd, call_method=qd.check_call)

    # Set the credentials so WAS_USER_NAME can start/stop WAS
    # Note if WAS_USER_NAME is different than the owner of WAS/bin files,
    # Then they need to set credentials to be able to execute wsadmin
    # in their java user.home directory which overrides soap.client.props
    # SOAPPROPS=~${WAS_USER_NAME}/wsadmin.properties

    # If manually adding passwords in soap.client.props, then encrypt password with
    # $WAS_PROFILE_ROOT/bin/PropFilePasswordEncoder.sh $SOAPPROPS com.ibm.SOAP.loginPassword
    qd.echo_info("----- Setting up soap properties for [%s]" % was_admin_user)
    soap_props = qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                              qd_props.DEFAULT_REL_WAS_PROFILE, 'properties', 'soap.client.props')

    qd.check_file_exists(soap_props)
    qd.QDEnv.set_value_to_prop_file(soap_props, 'com.ibm.SOAP.securityEnabled', 'true')
    qd.QDEnv.set_value_to_prop_file(soap_props, 'com.ibm.SOAP.loginUserid', was_admin_user)
    qd.QDEnv.set_value_to_prop_file(soap_props, 'com.ibm.SOAP.loginPassword', was_admin_password)
    qd.QDEnv.set_value_to_prop_file(soap_props, 'com.ibm.SOAP.requestTimeout', '360')
    # Encode the password in soap.client.props file
    exe_name = qd.os_switch('PropFilePasswordEncoder.sh', 'PropFilePasswordEncoder.bat')
    cmd = [qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                        qd_props.DEFAULT_REL_WAS_PROFILE, 'bin',
                        exe_name), soap_props, 'com.ibm.SOAP.loginPassword']
    install_util.su_call_if_necessary('--', (os_user_name, os_user_password),
                                      cmd, call_method=qd.check_call)
    single_was_start_stop.main(['restart'])


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main()
    exit(0)
