#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Update WebSphere configuration:

1. Set up appropriate heapsize based on the total physical memory size
2. Set up JVM properties and custom properties
3. Set up Threadpool properties
4. Set up session manager properties
5. Set up LDAP authentication
6. Configure single sign on option - export/import jts LTPA key
'''
import os
import re
import sys

import enable_ldap
import enable_sso

# find qdlib from parent paths and add the first one to the Python search path
# pylint: disable=line-too-long
sys.path.append(next(p for p in ['..', '../..', '../../..'] if os.path.exists(p + '/qdlib') and (p + '/qdlib') not in sys.path))
# pylint: enable=line-too-long

from qdlib import qd, qd_props, install_util


def get_memory_size_on_linux():
    '''
    Get the total memory size on Linux platform.

    :return: memory size in MB
    '''
    meminfo = open('/proc/meminfo').read()
    matched = re.search(r'^MemTotal:\s+(\d+)', meminfo)
    mem_total_kb = 0
    if matched:
        mem_total_kb = int(matched.groups()[0])
        # return the memory size in MB
        return mem_total_kb / 1000
    else:
        qd.echo_error("Having trouble to get the memory size")
        qd.exit_(1)


def get_memory_size_on_windows():
    '''
    Get the total memory size on Windows platform.

    :return: memory size in MB
    '''
    process = os.popen('wmic memorychip get capacity')
    result = process.read()
    process.close()
    mem_total_b = 0
    for mem in result.split("  \r\n")[1:-1]:
        mem_total_b += int(mem)
    if mem_total_b == 0:
        qd.exit_with_err_msg("Failed to get the total memory size on windows,"
                             " please make sure wmic utility is enabled")
    # return the memory size in MB
    return mem_total_b / (1000**2)


def get_heapsize_replacement():
    '''
    Determine appropriate heap size arguments, based on total physical memory size.

    :return: list of heapsize arguments (Heapsize, -Xmx, -Xms, etc.), based on memory size
    '''
    if qd.is_linux():
        mem_total = get_memory_size_on_linux()
    elif qd.is_windows():
        mem_total = get_memory_size_on_windows()
    else:
        qd.echo_error("Unsupported platform")
        qd.exit_(1)
    if mem_total < 4000:
        qd.echo_info("The system RAM size is less than 4GB")
        replacement_list = [('HeapSize=4096', 'HeapSize=' + str(mem_total)),
                            ('-Xmx4g -Xms4g', '-Xmx2g -Xms2g')]
    elif mem_total < 6000:
        qd.echo_info("The system RAM size is 4gb, do nothing")
        replacement_list = []
    elif mem_total < 7500:
        qd.echo_info("The system RAM size is less than 7GB")
        replacement_list = [('HeapSize=4096', 'HeapSize=' + str(mem_total)),
                            ('-Xmx4g -Xms4g', '-Xmx6g -Xms6g'), ('-Xmn512m', '-Xmn1g')]
    else:
        qd.echo_info("The system RAM size more than 8gb, allocate 75% of total memory $RAMSize")
        allocate_mem = int(mem_total * 0.75)
        heapsize = allocate_mem / 1000
        qd.echo_info("Allocate memory is [%s]" % allocate_mem)
        replacement_list = [('HeapSize=4096', 'HeapSize=' + str(allocate_mem)),
                            ('-Xmx4g -Xms4g',
                             '-Xmx' + str(heapsize) + 'g' + ' -Xms' + str(heapsize) + 'g'),
                            ('-Xmn512m', '-Xmn1g')]
    return replacement_list


def execute_wsadmin_command(envprops, jython_file_full_path, arg_option):
    '''
    Execute wasadmin.sh or wsadmin.bat to run a jython file

    :param envprops: environment property, to get the QD_Root directory
    :param jython_file_full_path: absolute jython file path
    :param arg_option: absolute WASCommon directory path
    '''
    user_props = qd_props.get_user_props(envprops)
    os_user_name, os_user_password = qd_props.get_os_users_creds(user_props)

    cmd = [qd.path_join(envprops.get_values(qd_props.PN_QD_ROOT),
                        qd_props.DEFAULT_REL_WAS_ROOT, 'bin',
                        qd_props.DEFAULT_WSADMIN_EXE), '-lang', 'jython',
           '-f', jython_file_full_path, arg_option]
    install_util.su_call_if_necessary('--', (os_user_name, os_user_password),
                                      cmd, call_method=qd.check_call)


def main():
    '''Main function; see module documentation.'''
    # set up local environment for this script
    working_dir = os.getcwdu()
    # make sure others can execute any file under the folder, especially the shell scripts
    qd.chmod_r_777(working_dir, mode='755')
    env_props = qd_props.get_default_env_props()
    env_props.set_value(qd_props.PN_WAS_COMMON_DIR, working_dir)
    local_tags = env_props.get_values('localTags')

    qd.echo_info("was_common_dir is [%s]" % working_dir)
    # we need convert "\\" to "/" in the was.properties,
    # since the path passed into WASadmin need to be c:/IBMQD/JazzTeamServer
    # str(jazz_home) is required to go from the tuple to a string!
    jazz_home = str(env_props.get_values(qd_props.PN_CLM_HOME_PREP))
    converted_clm_home = jazz_home.replace('\\', '/')
    replacement_list = [('_CLM_HOME_', converted_clm_home)]
    replacement_list += get_heapsize_replacement()

    if env_props.get_values(qd_props.PN_DB_VENDOR) == 'oracle':
        qd.QDEnv.set_value_to_prop_file(qd.path_join(working_dir, 'was.properties'),
                                        'ORACLE_JAR_NAME',
                                        env_props.get_values(qd_props.PN_DB_DRIVER_FILE))
        replacement_list += [('_JAR_FILE_', env_props.get_values(qd_props.PN_DB_DRIVER_FILE))]
    # Update all appropriate values in was.properties file
    qd.sed_in_place(qd.path_join(working_dir, 'was.properties'), replacement_list)

    qd.echo_info("Setting up JVM properties")
    execute_wsadmin_command(env_props, qd.path_join(working_dir, 'set_jvm_props.jy'), working_dir)
    qd.echo_info("Setting up Threadpool properties")
    execute_wsadmin_command(env_props, qd.path_join(working_dir, 'set_threadpool_props.jy'),
                            working_dir)
    qd.echo_info("Setting up session manager properties")
    execute_wsadmin_command(env_props, qd.path_join(working_dir, 'set_session_manager_props.jy'),
                            working_dir)

    if ('DM' in local_tags or 'RM' in local_tags) and qd.is_linux():
        add_display = [('Bootstrap values ...',
                        'ADDED FOR CONVERTER CLUSTERING SUPPORT\nDISPLAY=:3.0\n'
                        'export DISPLAY\n\n# Bootstrap values ...')]
        qd.sed_in_place(qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                        qd_props.DEFAULT_REL_WAS_ROOT, 'bin', 'startServer.sh'), add_display)
        qd.sed_in_place(qd.path_join(env_props.get_values(qd_props.PN_QD_ROOT),
                        qd_props.DEFAULT_REL_WAS_ROOT, 'bin', 'startNode.sh'), add_display)

    enable_ldap.main()
    if 'jts' in env_props.get_values(qd_props.PN_APPS):
        enable_sso.main(['export'])
    qd.echo_info("Successfully completed {scriptname} execution")


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main()
    exit(0)
