#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Script does the following:
    1. Checks rules to confirm it is valid to install the iFix
    2. Download / assemble the iFix media (zip) ready for the "engine to use"
    3. Dynamically import the correct engine to install the iFix
    4. Use "engine" to install the ifix

Assumptions:
    1. There is already an install done [eg 6.0 GA]
    2. Nothing has been deployed into WAS
    3. No inputs, control comes from /etc/env.properties and the QD tagging system
    4. iFix rules have already been checked

Also see :func:`main` function.
'''
import sys
import os
import imp


# find qdlib from parent paths and add the first one to the Python search path
# pylint: disable=line-too-long
sys.path.append(next(p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and (p + '/qdlib') not in sys.path))
# pylint: enable=line-too-long
# pylint: disable=wrong-import-position
from qdlib import qd, qd_props, media, qd_topo

# ---------- CONSTANTS ------------
PN_IFIX_MEDIA_TYPE = 'iFixMediaType'  #:
''' Property name - ifix media type, in environment, supplied by media.properties '''
PN_CLM_IFIX_MEDIA_ROOT = 'CLMiFixMediaRoot'  #:
''' Property name - clm ifix media root, in environment, supplied by media.properties '''
PN_DM_IFIX_MEDIA_ROOT = 'DMiFixMediaRoot'  #:
''' Property name - dm ifix media root, in environment, supplied by media.properties '''
PN_DMMS_IFIX_MEDIA_ROOT = 'DMMSiFixMediaRoot'  #:
''' Property name - dmms ifix media root, in environment, supplied by media.properties '''
PN_IFIX_FILE = 'iFixFile'  #:
''' Property name - ifix file, in environment, supplied by media.properties '''
PN_CLM_PATCH_ZIP_SUFFIX = 'clmPatchZipSuffix'  #:
''' Property name - CLM patch zip suffix, supplied by media.properties '''

PN_CLM_BUILD_ID = 'CLM_BUILD_ID'  #:
''' Property name - CLM build id, in environment, supplied by clm UCD component '''
PN_CLM_BUILDROOTURL = 'CLM_BUILDROOTURL'  #:
''' Property name - CLM build root URL, in environment, supplied by clm UCD component '''

PN_APP_MEDIA_ROOT = 'appMediaRoot'  #:
''' Property name - app media root, in environment, supplied by UCD gui '''
PN_CLM_IFIX_VERSION = 'CLMiFixVersion'  #:
''' Property name - clm ifix version, in environment, supplied by UCD gui '''
PN_DM_IFIX_VERSION = 'DMiFixVersion'  #:
''' Property name - dm ifix version, in environment, supplied by UCD gui '''
PN_DMMS_IFIX_VERSION = 'DMMSiFixVersion'  #:
''' Property name - dmms ifix version, in environment, supplied by UCD gui '''

CLM_IFIXABLE_APPS = ['gc', 'dcc', 'relm', 'qm', 'rm', 'ccm', 'jts', 'lqe', 'ldx', 'rs', 'dm']  #:
''' List of apps that can be patched by a CLM iFix '''
DM_IFIXABLE_APPS = ['dm']  #:
''' List of apps that can be patched by a DM iFix '''
DMMS_IFIXABLE_APPS = ['dm']  #:
''' List of apps that can be patched by a DMMS (Rhapsody Design Manager) iFix '''

# ----- files names -----
LOCAL_ENGINE_CONFIG_FILE = 'QDiFixConfig.properties'
LOCAL_ENGINE_NAME = 'install_ifix_engine.py'
DOWNLOADED_ENGINE_NAME = 'downloaded_ifix_engine.py'
MEDIA_SERVER_ENGINE_CONFIG_FILE_TAIL = '_QDconfig.properties'

CLM_SERVER_PATCH_PARTIAL_NAME = 'CLM_server_patch_'
CLM_ADDITIONAL_FILES_PARTIAL_NAME = 'AdditionalFiles_'
CLM_CONSTRUCTED_ZIP_NAME = "CLM_constructed_iFix.zip"

# ----- working folders -----
REL_IFIX_DIR = "iFixPatches"
REL_IFIX_ZIP_DIR = qd.path_join(REL_IFIX_DIR, "ZIP")
REL_CLM_IFIX_ADDITIONAL_DIR = qd.path_join(REL_IFIX_DIR, "CLM/downloads", "additional")
REL_CLM_IFIX_TEMP_DIR = qd.path_join(REL_IFIX_DIR, "CLM/downloads", "temp")
REL_CLM_IFIX_FILES_DIR = qd.path_join(REL_IFIX_DIR, "CLM/downloads", "files")


# pylint: disable=too-many-locals,too-many-statements,too-many-branches
def check_ifix_version_and_scope(env_props):
    '''
    Check if iFixing is enabled and that the server contains at least one app in scope for iFixing;
    assuming all other iFix rules have been checked in core_automation_prep.py

    :param env_props: instance of env.properties

    :return tuple (clm_ifix_settings_valid, dm_ifix_settings_valid, dmms_ifix_settings_valid):
        * **clm_ifix_settings_valid**: (boolean) True if in scope for a CLM ifix, False if not
        * **dm_ifix_settings_valid**: (boolean) True if in scope for a DM ifix, False if not
        * **dmms_ifix_settings_valid**: (boolean) True if in scope for a DMMS ifix, False if not
    '''

    # get the properties to check [iFixEnabled][CLMiFixVersion][DMiFixVersion][DMMSiFixVersion]
    ifix_enabled = env_props.get_values('iFixEnabled')
    clm_ifix_version = env_props.get_values('CLMiFixVersion', CLMiFixVersion=None)
    dm_ifix_version = env_props.get_values('DMiFixVersion', DMiFixVersion=None)
    dmms_ifix_version = env_props.get_values('DMMSiFixVersion', DMMSiFixVersion=None)

    # only continue if "iFixing is enabled"
    if qd.parse_boolean(ifix_enabled, 'ifix_enabled'):
        qd.echo_info("potentially installing iFix as iFixEnabled [%s]" % ifix_enabled)
    else:
        qd.exit_with_warning_msg("not installing iFix as iFixEnabled [%s]" % ifix_enabled)

    # only continue if versions have been set
    assert clm_ifix_version, "clm_ifix_version [%s] not set!" % clm_ifix_version
    assert dm_ifix_version, "dm_ifix_version [%s] not set!" % dm_ifix_version
    assert dmms_ifix_version, "dmms_ifix_version [%s] not set!" % dmms_ifix_version

    # use these flags at the end to determine the final result
    clm_ifix_settings_valid = None
    dm_ifix_settings_valid = None
    dmms_ifix_settings_valid = None

    def check_scope(ifix_type, apps, ifixable_apps):
        '''
        Determine if the apps on the current server are in scope for patching with the ifix
        inputs

        :param ifix_type: {CLM, DM, DMMS}
        :param apps: list of apps (tags) on the server eg ['jts', 'rm', 'dm]
        :param ifixable_apps: list of apps in scope for an {ifix_type} patch

        :return ifix_settings_valid (boolean):
            * True - one (or more) of the apps are in scope for ifix patching
            * False - no apps are in scope
        '''
        ifix_settings_valid = False
        for app in apps:
            if app in ifixable_apps:
                ifix_settings_valid = True
        if ifix_settings_valid:
            qd.echo_info("potentially installing [%s] patches as at least one tag in [%s]"
                         " matches one in the iFixable list [%s]"
                         % (ifix_type, server_apps, ifixable_apps))
        else:
            ifix_settings_valid = False
            qd.echo_info("not installing any [%s] patches as no tag in [%s]"
                         " matches one in the iFixable list [%s]"
                         % (ifix_type, server_apps, ifixable_apps))
        return ifix_settings_valid

    # get topology and current server tags information
    topo = qd_topo.QDTopo(env_props.get_values(qd_props.UCD_PN_GLOBAL_TAG_LISTING))
    thost = topo.get_host(env_props.get_values('localHostname', localHostname=qd.get_host_name()))
    server_apps = thost.get_local_apps()

    if clm_ifix_version == 'NONE':
        qd.echo_info("not installing any CLM iFixes as CLMiFixVersion [%s]" % clm_ifix_version)
        clm_ifix_settings_valid = False
    else:
        clm_ifix_settings_valid = check_scope('CLM', server_apps, CLM_IFIXABLE_APPS)

    if dm_ifix_version == 'NONE':
        qd.echo_info("not installing any DM iFixes as DMiFixVersion [%s]" % dm_ifix_version)
        dm_ifix_settings_valid = False
    else:
        dm_ifix_settings_valid = check_scope('DM', server_apps, DM_IFIXABLE_APPS)

    if dmms_ifix_version == 'NONE':
        qd.echo_info("not installing any DMMS iFixes as DMMSiFixVersion [%s]" % dmms_ifix_version)
        dmms_ifix_settings_valid = False
    else:
        dmms_ifix_settings_valid = check_scope('DMMS', server_apps, DMMS_IFIXABLE_APPS)

    # exit cleanly
    return clm_ifix_settings_valid, dm_ifix_settings_valid, dmms_ifix_settings_valid


# pylint: disable=too-many-public-methods,too-many-instance-attributes
class BaseIFixMedia(media.QDMedia):
    '''Class giving access to the iFix media'''

    def __init__(self, media_props):
        super(BaseIFixMedia, self).__init__(media_props)

        self.a_app_media_root_prop = PN_APP_MEDIA_ROOT
        self.a_media_type_prop = PN_IFIX_MEDIA_TYPE

        # place holders for populating later
        self.script_folder = None
        self.ifix_engine = None  # path to the ifix engine module
        self.ifix_engine_config = None  # # path to the ifix engine config file
        self.ifix_zip = None  # # path to the ifix zip file

        self.a_file_prop = PN_IFIX_FILE  # set to "directory" to use the media server contents

        # blank but required
        self.ifix_zip_root_dir = None
        self.ifix_zip_dir = None
        self.ifix_media_root = None
        self.ifix_media_version = None

        # blank but required for internal use ==CLM==
        self.clm_ifix_additional_dir = None
        self.clm_ifix_temp_dir = None
        self.clm_ifix_files_dir = None
        self.clm_ifix_download_dir = None
        self.ifix_int_maint_file_prefix = None

        # populated for internal use ==CLM==
        self.a_ifix_build_root_url_prop = PN_CLM_BUILDROOTURL
        self.a_ifix_url_suffix_prop = PN_CLM_PATCH_ZIP_SUFFIX
        self.ifix_build_id_prop = PN_CLM_BUILD_ID

        # populated for internal use ==DM==
        # none as not currently supported

        # populated for internal use ==DMMS==
        # none as not currently supported

        # populated for external use ==CLM==
        self.a_clm_media_root_prop = PN_CLM_IFIX_MEDIA_ROOT  # media property
        self.a_clm_ifix_version_prop = PN_CLM_IFIX_VERSION  # env property
        # populated for external use ==DM==
        self.a_dm_media_root_prop = PN_DM_IFIX_MEDIA_ROOT  # media property
        self.a_dm_ifix_version_prop = PN_DM_IFIX_VERSION  # env property
        # populated for external use ==DMMS==
        self.a_dmms_media_root_prop = PN_DMMS_IFIX_MEDIA_ROOT  # media property
        self.a_dmms_ifix_version_prop = PN_DMMS_IFIX_VERSION  # env property

    def locate_ifix_engine(self):
        '''
        Locate the path to the ifix engine module. Depending on the value of {app_media_type}
        the module may need downloaded or copied to a local folder.

        :return: full path to the module
        '''

        ifix_engine = ""
        app_media_type = self.pget(self.a_media_type_prop)
        if app_media_type in ['CLM']:
            qd.echo_info("[%s] [%s] engine in the same folder as this module"
                         % (self.a_media_type_prop, app_media_type))
            qd.echo_info("workout location")
            target_folder = self.get_script_folder()
            ifix_engine = qd.path_join(target_folder, LOCAL_ENGINE_NAME)

        elif app_media_type in ['ZIP']:
            qd.echo_info("[%s] [%s] engine is on media server"
                         % (self.a_media_type_prop, app_media_type))
            self.set_download_dir(self.get_ifix_zip_dir())
            zip_dir = self.prepare_media()
            downloaded_ifix_engine = qd.path_join(zip_dir, LOCAL_ENGINE_NAME)
            qd.check_file_exists(downloaded_ifix_engine)
            # need to move the file to the "QD folder" so it can be loaded (imported) by python
            ifix_engine = qd.copy_file_new_name(downloaded_ifix_engine, self.get_script_folder(),
                                                DOWNLOADED_ENGINE_NAME)

        else:
            qd.exit_with_err_msg("unknown media type $%s [%s]"
                                 % (self.a_media_type_prop, app_media_type))

        qd.check_file_exists(ifix_engine)
        self.ifix_engine = ifix_engine
        return ifix_engine

    def get_engine_data(self):
        '''
        Locate the path to the ifix engine data files {qd_config_file} {ifix_zipfile}.
        Depending on the value of {app_media_type} the files may need downloaded
        or copied to a local folder.

        :return tuple (qd_config_file, ifix_zipfile):
            * **qd_config_file**: full path to the engine config file
            * **ifix_zipfile**: full path to the ifix zip file
        '''

        ifix_engine_config = ""
        ifix_zip = ""
        app_media_type = self.pget(self.a_media_type_prop)
        if app_media_type in ['CLM']:
            qd.echo_info("[%s] [%s] config file in the same folder as this module"
                         % (self.a_media_type_prop, app_media_type))
            target_folder = self.get_script_folder()
            ifix_engine_config = qd.path_join(target_folder, LOCAL_ENGINE_CONFIG_FILE)
            qd.check_file_exists(ifix_engine_config)

            # get the CLM_server_patch zip file
            qd.echo_info("zip file urls have to be constructed")
            self.set_ifix_int_maint_file_prefix(CLM_SERVER_PATCH_PARTIAL_NAME)
            self.set_download_dir(self.get_clm_ifix_files_dir())
            self.prepare_media()

            # get the additional_files zip file
            self.set_ifix_int_maint_file_prefix(CLM_ADDITIONAL_FILES_PARTIAL_NAME)
            self.set_download_dir(self.get_clm_ifix_additional_dir())
            additional_files_zipfile = self.prepare_media()

            qd.echo_info("unzip and extract the required parts")
            # extract contents
            qd.unzip_zip(additional_files_zipfile, self.get_clm_ifix_temp_dir())

            # parse each extracted file in case it needs extracted again
            for f_temp in qd.list_files_in_folder(self.get_clm_ifix_temp_dir()):
                f_with_path = qd.path_join(self.get_clm_ifix_temp_dir(), f_temp)
                if qd.is_zipfile(f_with_path):
                    names = qd.list_zip(f_with_path)
                    if len(names) == 1:
                        qd.unzip_zip(f_with_path, self.get_clm_ifix_files_dir())
                    else:
                        qd.copy_file_with_stat(f_with_path, self.get_clm_ifix_files_dir())
                else:
                    qd.copy_file_with_stat(f_with_path, self.get_clm_ifix_files_dir())

            # zip the contents of "get_clm_ifix_files_dir()" to "get_ifix_zip_dir()"
            ifix_zip = qd.path_join(self.get_ifix_zip_dir(), CLM_CONSTRUCTED_ZIP_NAME)
            qd.echo_info("construct into the required zip [%s]" % ifix_zip)
            qd.zipup_dir(self.get_clm_ifix_files_dir(), ifix_zip)

        elif app_media_type in ['ZIP']:
            qd.echo_info("[%s] [%s] config and zip file in the same folder but on media server"
                         % (self.a_media_type_prop, app_media_type))
            self.set_download_dir(self.get_ifix_zip_dir())
            zip_dir = self.prepare_media()
            config_file = self.get_ifix_media_version() + MEDIA_SERVER_ENGINE_CONFIG_FILE_TAIL
            ifix_engine_config = qd.path_join(zip_dir, config_file)
            ifix_zip = qd.path_join(zip_dir, self.get_ifix_media_version() + ".zip")

        else:
            qd.exit_with_err_msg("unknown media type $%s [%s]"
                                 % (self.a_media_type_prop, app_media_type))

        self.ifix_engine_config = ifix_engine_config
        self.ifix_zip = ifix_zip

        return self.ifix_engine_config, self.ifix_zip

    def set_script_folder(self, script_folder):
        '''
        Set the internal property [script_folder].
        It is the folder containing this python script

        :param script_folder: new property value
        '''
        self.script_folder = script_folder

    def get_script_folder(self):
        '''
        Returns the folder containing this python script

        :return: the internal property [script_folder]
        '''
        return self.script_folder

    def create_working_folders(self, qd_media_root):
        '''
        Create the working folders required for downloading / creating the ifix zip files

        :param qd_media_root: root folder for all QD media
        '''

        # make the root dir
        self.ifix_zip_root_dir = qd.path_join(qd_media_root, REL_IFIX_DIR)
        # and all the children off it
        self.ifix_zip_dir = qd.path_join(qd_media_root, REL_IFIX_ZIP_DIR)
        self.clm_ifix_additional_dir = qd.path_join(qd_media_root, REL_CLM_IFIX_ADDITIONAL_DIR)
        self.clm_ifix_temp_dir = qd.path_join(qd_media_root, REL_CLM_IFIX_TEMP_DIR)
        self.clm_ifix_files_dir = qd.path_join(qd_media_root, REL_CLM_IFIX_FILES_DIR)

        # create the folders if they do not exist
        qd.check_and_create_folder(self.ifix_zip_root_dir)
        qd.check_and_create_folder(self.ifix_zip_dir)
        qd.check_and_create_folder(self.clm_ifix_additional_dir)
        qd.check_and_create_folder(self.clm_ifix_temp_dir)
        qd.check_and_create_folder(self.clm_ifix_files_dir)

    def delete_working_folders(self):
        '''
        Delete the working folders required for downloading / creating the ifix zip files
        '''

        qd.delete_folder_if_exists(self.ifix_zip_dir)
        qd.delete_folder_if_exists(self.clm_ifix_additional_dir)
        qd.delete_folder_if_exists(self.clm_ifix_temp_dir)
        qd.delete_folder_if_exists(self.clm_ifix_files_dir)
        # and finally the parent
        qd.delete_folder_if_exists(self.ifix_zip_root_dir)

    def cleanup(self):
        '''
        *Override the media.QDMedia method*

        Delete any temporary files created / downloaded
        '''

        self.delete_working_folders()
        # if the engine was downloaded it will have been temporarily copied
        duplicate_ifix_engine = qd.path_join(self.get_script_folder(), DOWNLOADED_ENGINE_NAME)
        qd.delete_file_if_exists(duplicate_ifix_engine)

    def get_ifix_zip_root_dir(self):
        '''
        Get the parent folder containing all the working directories used by the script.
        e.g., ``/IBMQD/iFixPatches``

        :return: the internal property [ifix_zip_root_dir]
        '''
        return self.ifix_zip_root_dir

    def get_ifix_zip_dir(self):
        '''
        Get the working directory containing the assembled ifix zip file.
        e.g., ``/IBMQD/iFixPatches/ZIP``

        :return: the internal property [ifix_zip_dir]
        '''
        return self.ifix_zip_dir

    def get_clm_ifix_additional_dir(self):
        '''
        Get the working directory containing the downloaded additional files zip.
        e.g., ``/IBMQD/iFixPatches/ZIP/CLM/downloads/additional``

        :return: the internal property [clm_ifix_additional_dir]
        '''
        return self.clm_ifix_additional_dir

    def get_clm_ifix_temp_dir(self):
        '''
        Get the working directory containing the content of the downloaded additional files zip.
        e.g., ``/IBMQD/iFixPatches/ZIP/CLM/downloads/temp``

        :return: the internal property [clm_ifix_temp_dir]
        '''
        return self.clm_ifix_temp_dir

    def get_clm_ifix_files_dir(self):
        '''
        Get the working directory containing the files to be zipped into the assembled ifix zip.
        e.g., ``/IBMQD/iFixPatches/ZIP/CLM/downloads/files``

        :return: the internal property [clm_ifix_files_dir]
        '''
        return self.clm_ifix_files_dir

    def get_download_dir(self):
        '''
        *Override the media.QDMedia method*

        Get the folder to download into.

        :return: the internal property [clm_ifix_download_dir]
        '''
        return self.clm_ifix_download_dir

    def set_download_dir(self, download_dir):
        '''
        *Override the media.QDMedia method*

        Set the internal property [clm_ifix_download_dir].
        It is the folder to download into.

        :param download_dir: the new folder path
        '''
        self.clm_ifix_download_dir = download_dir

    def set_ifix_int_maint_file_prefix(self, new_prefix):
        '''
        Set the internal property [ifix_int_maint_file_prefix].
        It is the current prefix for constructing the zip file name to download.

        :param new_prefix: new property value
        '''
        self.ifix_int_maint_file_prefix = new_prefix

    def get_ifix_int_maint_file_prefix(self):
        '''
        Get the current prefix for constructing the zip file name to download.
        eg ``CLM_server_patch_`` or ``AdditionalFiles_``

        :return: the internal property [ifix_int_maint_file_prefix]
        '''
        return self.ifix_int_maint_file_prefix

    def set_ifix_media_root(self, new_media_root):
        '''
        Set the internal property [ifix_media_root].
        It is the root folder containing the ifix to download.

        :param new_media_root: new property value
        '''
        self.ifix_media_root = new_media_root

    def get_ifix_media_root(self):
        '''
        Get the root folder containing the ifix to download.
        Defined in media.properties and different for CLM, DM and DMMS

        :return: the internal property [ifix_media_root]
        '''
        return self.ifix_media_root

    def set_ifix_media_version(self, new_media_version):
        '''
        Set the internal property [ifix_media_version]
        It is the version selected to be installed for CLM, DM or DMMS eg ``CLM_601_iFix010``.

        :param new_media_version: new property value
        '''
        self.ifix_media_version = new_media_version

    def get_ifix_media_version(self):
        '''
        Get the media version for CLM, DM or DMMS iFix eg ``CLM_601_iFix010``.

        :return: the internal property [ifix_media_version]
        '''
        return self.ifix_media_version

    def construct_internal_url_prefix(self):
        '''
        Construct front part of the url to the internal ifix location
        using {ifix_build_root_url} and {ifix_url_suffix}

        :return: the constructed front part of the url
        '''
        ifix_build_root_url = self.pget(self.a_ifix_build_root_url_prop)
        ifix_url_suffix = self.pget(self.a_ifix_url_suffix_prop)
        # only continue if urls have been set
        assert ifix_build_root_url, "ifix_build_root_url [%s] not set!" % ifix_build_root_url
        assert ifix_url_suffix, "ifix_url_suffix [%s] not set!" % ifix_url_suffix
        qd.echo_info("constructing url prefix from url root [%s] and url suffix [%s]"
                     % (ifix_build_root_url, ifix_url_suffix))
        url_prefix = qd.url_join(ifix_build_root_url, ifix_url_suffix)
        qd.echo_info("calculated url [%s]" % url_prefix)
        return url_prefix

    def construct_int_ifix_filename(self):
        '''
        Construct iFix filename eg ``CLM_server_patch_6.0.2.0-CALM603M-I20170223-0933.zip``
        using {app_media_root} {ifix_file_prefix} {ifix_stream_id} and {ifix_build_id}

        :return: the constructed filename
        '''
        app_media_root = self.pget(self.a_app_media_root_prop)
        ifix_build_id = self.pget(self.ifix_build_id_prop)
        assert app_media_root, "app_media_root [%s] not set!" % app_media_root
        assert ifix_build_id, "ifix_build_id [%s] not set!" % ifix_build_id

        ifix_stream_id = media.get_ifix_stream(app_media_root)
        assert ifix_stream_id, "ifix_stream_id [%s] not set!" % ifix_stream_id

        # calculate the file name
        ifix_file_prefix = self.get_ifix_int_maint_file_prefix()
        qd.echo_info("ifix_file_prefix is [%s], ifix_stream_id [%s] and ifix_build_id [%s]"
                     % (ifix_file_prefix, ifix_stream_id, ifix_build_id))

        ifix_file = ifix_file_prefix + ifix_stream_id + '-' + ifix_build_id + '.zip'

        qd.echo_info("constructing maintenance zip file name using "
                     "app_media_root [%s], ifix_file_prefix [%s], ifix_stream_id [%s]"
                     " and ifix_build_id [%s]"
                     % (app_media_root, ifix_file_prefix, ifix_stream_id, ifix_build_id))

        qd.echo_info("calculated filename [%s]" % ifix_file)
        return ifix_file

    def get_local_directory_location(self, **kwargs):
        '''
        *Override the media.QDMedia method as {ifix_media_root} is different for CLM/DM/DMMS*

        Returns the path to the local directory.

        Note: The version and OSdependent values are left in for future compatability.

        :param kwargs: dictionary which can be empty containing
            {"OSdependent": '<value>', "version": '<value>'}
        :return: path to local directory
        '''
        media_root = self.pget(self.media_root_prop)
        ifix_media_root = self.get_ifix_media_root()
        if "version" in kwargs:
            a_version = kwargs["version"]
        else:
            a_version = self.pget(self.a_version_prop)
        if "OSdependent" in kwargs:
            return qd.path_join(media_root, ifix_media_root, a_version, kwargs["OSdependent"])
        else:
            return qd.path_join(media_root, ifix_media_root, a_version)

    def get_remote_directory_location(self, **kwargs):
        '''
        *Override the media.QDMedia method as {ifix_media_root} is different for CLM/DM/DMMS*

        Builds and returns a URL to the directory.

        Note: The version and OSdependent values are left in for future compatability.

        :param kwargs: dictionary which can be empty containing
            {"OSdependent": '<value>', "version": '<value>'}
        :return: url to remote directory
        '''
        media_server = self.pget(self.media_server_prop)
        ifix_media_root = self.get_ifix_media_root()
        if "version" in kwargs:
            a_version = kwargs["version"]
        else:
            a_version = self.pget(self.a_version_prop)
        if "OSdependent" in kwargs:
            return qd.url_join(media_server, ifix_media_root, a_version, kwargs["OSdependent"])
        else:
            return qd.url_join(media_server, ifix_media_root, a_version)

    def _custom_prepare_media(self):
        '''
        *Override the media.QDMedia method*

        Only supports
            * **CLM** returns the path to the download file
            * **ZIP** returns the path to the folder containing the downloaded files
            * all other types result in exit_with_err_msg()

        :return: **[CLM]** the path to the downloaded file
            or **[ZIP]** the path to the folder containing all the files downloaded
        '''

        app_media_type = self.pget(self.a_media_type_prop)
        if app_media_type == 'CLM':
            # using CLM maintenance repo
            ifix_file_url_prefix = self.construct_internal_url_prefix()
            ifix_file_name = self.construct_int_ifix_filename()
            ifix_file_url = qd.url_join(ifix_file_url_prefix, ifix_file_name)
            qd.echo_info("ifix file url to be downloaded [%s]" % ifix_file_url)

            assert 'notUsed' not in ifix_file_url

            qd.echo_debug("download_or_copy [%s]" % ifix_file_url)
            timer_get = qd.start_clock("Ready to get ifix file [%s] of type [%s]"
                                       % (ifix_file_url, app_media_type))
            downloaded_file = self.download_or_copy(ifix_file_url)
            qd.stop_clock(timer_get)

            return downloaded_file

        elif app_media_type == 'ZIP':
            mydict = {"version": ""}  # to avoid breaking anything
            # Install media directory
            src_zip_dir = self.get_directory_location(**mydict)
            qd.echo_info('ifix file url to be downloaded, %s#src_zip_dir [%s]'
                         % (type(self).__name__, src_zip_dir))
            assert 'notUsed' not in src_zip_dir

            # get list of all files and remove all that do not contain the substring
            file_list = self.get_file_list(src_zip_dir)
            sub = self.get_ifix_media_version()
            match_list = [f_temp for f_temp in file_list if sub in f_temp]

            # download only those that match
            qd.echo_info("only downloading the files that match [%s]" % sub)
            match_count = 0
            for f_match in match_list:
                qd.echo_debug("download_or_copy [%s] [%s]" % (src_zip_dir, f_match))
                timer_get = qd.start_clock("Ready to get ifix file [%s] of type [%s]"
                                           % (f_match, app_media_type))

                media_access = self.pget(self.media_access_prop)
                if media_access == 'FILE':
                    remote_src_zip = qd.path_join(src_zip_dir, f_match)
                elif media_access in ['HTTP', 'HTTPS', 'FTP']:
                    remote_src_zip = qd.url_join(src_zip_dir, f_match)
                else:
                    raise Exception('unknown $%s [%s]' % (self.media_access_prop, media_access))

                local_src_zip = self.download_or_copy(remote_src_zip)
                qd.stop_clock(timer_get)
                qd.echo_debug("downloaded [%s]" % local_src_zip)
                match_count = match_count + 1

            if match_count == 0:
                qd.exit_with_err_msg("match_count [%s], no files matched" % match_count)
            else:
                qd.echo_info("downloaded [%s] files" % match_count)
                # return the folder containing the files downloaded
                return qd.dirname(local_src_zip)

        elif app_media_type in ['DM'] or app_media_type in ['DMMS']:
            # placeholder section for future internal expansion
            qd.exit_with_err_msg("at present do not support media type" % app_media_type)
        else:
            qd.exit_with_err_msg("do not support media type" % app_media_type)


def main(argv):
    '''
    Main function; see module documentation.

    1. Checks rules to confirm it is valid to install the iFix
    2. Download / assemble the iFix media (zip) ready for the "engine to use"
    3. Dynamically import the correct engine to install the iFix
    4. Use "engine" to install the ifix
    '''
    # load local environment
    script_folder = os.path.abspath(qd.dirname(argv[0]))

    env_props = qd_props.get_default_env_props()
    media_props = qd_props.get_media_props(env_props)  # get media properties

    # check rules
    qd.echo_info("establish if any iFix patches can be installed")
    clm_ifix_valid, dm_ifix_valid, dmms_ifix_valid = check_ifix_version_and_scope(env_props)
    if clm_ifix_valid or dm_ifix_valid or dmms_ifix_valid:
        qd.echo_info("at least one iFix patch can be installed")
    else:
        qd.exit_with_warning_msg("no tags match any of the iFix patch lists")

    # create the media object to do the work
    ifix_med = BaseIFixMedia(media_props)
    ifix_med.set_script_folder(script_folder)
    qd_media_root = env_props.get_values(qd_props.PN_INTERNAL_MEDIA_PROPS)
    ifix_med.create_working_folders(qd_media_root)

    if clm_ifix_valid or dm_ifix_valid or dmms_ifix_valid:

        app_media_type = ifix_med.pget(ifix_med.a_media_type_prop)
        assert app_media_type, "app_media_type [%s] not set!" % app_media_type

        clm_media_root = ifix_med.pget(ifix_med.a_clm_media_root_prop)
        assert clm_media_root, "clm_media_root [%s] not set!" % clm_media_root
        ifix_med.set_ifix_media_root(clm_media_root)

        # hijack {ifix_media_version} and use it to contain the engine name
        ifix_med.set_ifix_media_version(LOCAL_ENGINE_NAME)

        # engine is common to all iFixes
        ifix_engine_file = ifix_med.locate_ifix_engine()
        qd.echo_info("ifix_engine [%s]" % ifix_engine_file)

    if clm_ifix_valid:
        clm_media_root = ifix_med.pget(ifix_med.a_clm_media_root_prop)
        assert clm_media_root, "clm_media_root [%s] not set!" % clm_media_root
        ifix_med.set_ifix_media_root(clm_media_root)

        clm_ifix_version = ifix_med.pget(ifix_med.a_clm_ifix_version_prop)
        assert clm_ifix_version, "clm_ifix_version [%s] not set!" % clm_ifix_version
        ifix_med.set_ifix_media_version(clm_ifix_version)

        # access the 2 CLM files required by the iFix install engine
        ifix_clm_config_file, ifix_clm_zipfile = ifix_med.get_engine_data()
        qd.echo_info("ifix_clm_config_file [%s], ifix_clm_zipfile [%s]"
                     % (ifix_clm_config_file, ifix_clm_zipfile))

    if dm_ifix_valid:
        dm_media_root = ifix_med.pget(ifix_med.a_dm_media_root_prop)
        assert dm_media_root, "dm_media_root [%s] not set!" % dm_media_root
        ifix_med.set_ifix_media_root(dm_media_root)

        dm_ifix_version = ifix_med.pget(ifix_med.a_dm_ifix_version_prop)
        assert dm_ifix_version, "dm_ifix_version [%s] not set!" % dm_ifix_version
        ifix_med.set_ifix_media_version(dm_ifix_version)

        # access the 2 DM files required by the iFix install engine
        ifix_dm_config_file, ifix_dm_zipfile = ifix_med.get_engine_data()
        qd.echo_info("ifix_dm_config_file [%s], ifix_dm_zipfile [%s]"
                     % (ifix_dm_config_file, ifix_dm_zipfile))

    if dmms_ifix_valid:
        dmms_media_root = ifix_med.pget(ifix_med.a_dmms_media_root_prop)
        assert dmms_media_root, "dmms_media_root [%s] not set!" % dmms_media_root
        ifix_med.set_ifix_media_root(dmms_media_root)

        dmms_ifix_version = ifix_med.pget(ifix_med.a_dmms_ifix_version_prop)
        assert dmms_ifix_version, "dmms_ifix_version [%s] not set!" % dmms_ifix_version
        ifix_med.set_ifix_media_version(dmms_ifix_version)

        # access the 2 CLM files required by the iFix install engine
        ifix_dmms_config_file, ifix_dmms_zipfile = ifix_med.get_engine_data()
        qd.echo_info("ifix_dmms_config_file [%s], ifix_dmms_zipfile [%s]"
                     % (ifix_dmms_config_file, ifix_dmms_zipfile))

    # dynamically load the correct engine module
    qd.echo_info("dynamically import the correct ifix install engine module")
    split_name = os.path.splitext(qd.basename(ifix_engine_file))
    m_name = split_name[0]
    qd.echo_info("module [%s] path [%s]" % (m_name, ifix_engine_file))
    ifix_engine = imp.load_source(m_name, ifix_engine_file)

    # run the iFix engine
    if clm_ifix_valid:
        qd.echo_info("install CLM iFix")
        debug_command = "ifix_engine.main([" + ifix_clm_config_file + ", " + ifix_clm_zipfile + "])"
        qd.echo_debug("command is [%s]" % debug_command)
        ifix_engine.main([ifix_clm_config_file, ifix_clm_zipfile])

    if dm_ifix_valid:
        qd.echo_info("install DM iFix")
        debug_command = "ifix_engine.main([" + ifix_dm_config_file + ", " + ifix_dm_zipfile + "])"
        qd.echo_debug("command is [%s]" % debug_command)
        ifix_engine.main([ifix_dm_config_file, ifix_dm_zipfile])

    if dmms_ifix_valid:
        qd.echo_info("install DMMS iFix")
        debug_command = "ifix_engine.main([" + ifix_dmms_config_file
        debug_command = debug_command + ", " + ifix_dmms_zipfile + "])"
        qd.echo_debug("command is [%s]" % debug_command)
        ifix_engine.main([ifix_dmms_config_file, ifix_dmms_zipfile])

    # tidy up by deleting anything created, downloaded or unzipped
    ifix_med.cleanup()


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main(sys.argv)
    exit(0)
