#!/usr/bin/python
# pylint: disable=too-many-lines
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************


'''
Install CLM / DM iFix fix pack.

    :param argv[0]: the ifix config file that tells the engine what to do
    :param argv[1]: the ifix zip file containing the patch


This module will be called dynamically from another module.
This means that the location of the module can change depending on UCD being pipeline or external
modes.

To call dynamically the calling code will contain something like::

    import imp

    # dynamically load the module
    ifix_engine_dir = "full path of the folder containing the module"
    module_name = 'install_maintenance_files_engine'
    module_path = qd.path_join(ifix_engine_dir, module_name + '.py')
    ifix_engine = imp.load_source(module_name, module_path)

    ifix_config_file = "full path to the ifix config file"
    ifix_zip_file = "full path to the ifix zip file"
    ifix_engine.main([ifix_config_file, ifix_zip_file])

'''
import sys
import os
import glob
import time
import math
import tempfile
import logging

# find qdlib from parent paths and add the first one to the Python search path
# pylint: disable=line-too-long
sys.path.append(next(p for p in ['..', '../..'] if os.path.exists(p + '/qdlib') and (p + '/qdlib') not in sys.path))
# pylint: enable=line-too-long
# pylint: disable=wrong-import-position
from qdlib import qd, qd_props, qd_topo
# pylint: enable=wrong-import-position

# GLOBAL TO THIS SCRIPT
TIMESTAMP = str(int(math.floor(time.time())))

VALID_TOKEN_COUNT = 7

GROUP1_ACTIONS = ['backup', 'delete', 'backupDelete']  #:
''' actions referred to as group 1 '''
GROUP2_ACTIONS = ['copy', 'replace']  #:
''' actions referred to as group 2 '''
GROUP3_ACTIONS = ['unzip', 'unzipSameDir', 'untar', 'untarSameDir']  #:
''' actions referred to as group 3 '''
GROUP4_ACTIONS = ['unzipCopy', 'unzipReplace', 'untarCopy', 'untarReplace']  #:
''' actions referred to as group 4 '''
GROUP5_ACTIONS = ['modifyTspKey', 'deleteTspKey']  #:
''' actions referred to as group 5 '''

ALL_ACTIONS = GROUP1_ACTIONS + GROUP2_ACTIONS + GROUP3_ACTIONS + GROUP4_ACTIONS + GROUP5_ACTIONS

VALID_OPERATION_TYPES = ['file', 'folder', 'files']

VALID_OPERATING_SYSTEMS = ['win', 'linux']


def get_unzip_files_folder(zip_file):
    '''
    Return the path to the folder that the zip_file will be unzipped into.
    The path will be the full path to the zip_file plus the folder {Files}

    :param zip_file: file that will be unzipped
    :return: folder to contain the uinzipped files
    '''
    folder = qd.dirname(zip_file)
    unzip_files_folder = qd.path_join(folder, "Files")
    return unzip_files_folder


def get_apps(env_props):
    '''
    Get the APPS previously installed on the current server.
    Will exit the script if no apps exist on the server.

    :param env_props: instance of env.properties
    :return: list of apps on current server
    '''

    # get the APPS previously installed on the server
    server_hostname = env_props.get_values('localHostname', localHostname=qd.get_host_name())
    global_tag_list = env_props.get_values(qd_props.UCD_PN_GLOBAL_TAG_LISTING)
    topo = qd_topo.QDTopo(global_tag_list)
    thost = topo.get_host(server_hostname)
    apps = thost.get_local_apps()

    # === exit if no apps are selected
    if not apps:
        qd.echo_info("No apps are selected to install. Exiting...")
        qd.exit_(0)

    return apps


def matching_app(server_apps, tokens):
    '''
    Check the config file tag list (tokens[1]) against the apps on the current server.
    Return True if a match is found.

    token list = os:tag:operation:type:source:destination:string

    :param server_apps: list of apps on current server
    :param tokens: config_line split into tokens
    :return:
        * True - one (or more) of the tags is found in the server apps
        * False - no match found
    '''
    match_val = False

    # tags is a csv string stored in tokens[1], so convert to a list
    tag_list = tokens[1].split(",")

    for t_val in tag_list:
        if t_val in server_apps:
            match_val = True

    if match_val:
        qd.echo_debug("one or more apps in tag list [%s] found in the list of server apps [%s]"
                      % (tag_list, server_apps))

    return match_val


def matching_os(tokens):
    '''
    Check the config file os_list (tokens[0]) against the server operating system.
    Return True if a match is found.

    config_line format = os:tag:operation:type:source:destination:string

    :param tokens: config_line split into tokens
    :return:
        * True - one (or more) of the {os_list} entries matches the server operating system
        * False - no match found
    '''
    match_val = False

    # os_list is a csv string stored in tokens[0], so convert to a list
    os_list = tokens[0].split(",")

    if qd.is_windows():
        if "win" in os_list:
            qd.echo_debug("[win] match found in [%s]" % os_list)
            match_val = True

    elif qd.is_linux():
        if "linux" in os_list:
            qd.echo_debug("[linux] match found in [%s]" % os_list)
            match_val = True

    else:
        qd.exit_with_err_msg("Unsupported operating system")

    return match_val


def valid_config_line(config_line):
    '''
    Parse the line of text to determine whether to process or not,
    will only process lines that after stripping whitespace are.
    - not blank, nor commented out by a "#" character

    :param config_line: the string to check
    :return:
        * True - config line valid
        * False - config line NOT valid
    '''

    # remove leading / trailing whitespace
    config_line = config_line.strip()

    # is config_line blank
    if config_line:
        if config_line.startswith("#"):
            # config_line starts with a "#" character
            ret_val = False
        else:
            # config_line not blank, not a "#" line
            ret_val = True
    else:
        # config_line blank
        ret_val = False

    return ret_val


def tokenise_config_line(config_line):
    '''
    Tokenise the line of text, stripping whitespace from each token
    and returning a list containing the tokens

    :param config_line: the string to tokenise
    :return: list comprised of config_line split into tokens
    '''

    # split line into tokens and remove leading / trailing whitespace
    tokens = config_line.split(":")
    for enum_count, tok_val in enumerate(tokens):
        tokens[enum_count] = tok_val.strip()

    return tokens


def valid_config_line_format(tokens):
    # pylint: disable=too-many-branches
    '''
    Parse the list of tokens to determine it contains the correct number of entries,
    has valid {tags}, valid operation action {operation} and a valid operation type {type}

    token list = os:tag:operation:type:source:destination:string

    :param tokens: list containing the config_line split into individual tokens
    :return:
        * True - config line format valid
        * False - config line format NOT valid
    '''

    # set return value [True], only change to [False] if line is invalid
    ret_val = True

    if len(tokens) == VALID_TOKEN_COUNT:
        qd.echo_info("passed [%s] tokens test" % VALID_TOKEN_COUNT)
    else:
        # not valid
        ret_val = False
        qd.echo_error("line [%s] expected [%s] tokens, not [%s]" %
                      (tokens, VALID_TOKEN_COUNT, len(tokens)))

    # os is a csv string stored in tokens[0], so convert to a list and check they are valid
    # use a temp_flag as there are potentially many os's to check
    temp_flag = True
    os_list = tokens[0].split(",")
    for o_val in os_list:
        if o_val not in VALID_OPERATING_SYSTEMS:
            temp_flag = False
            qd.echo_error("os [%s] from [%s], is not in VALID_OPERATING_SYSTEMS [%s]"
                          % (o_val, os_list, VALID_OPERATING_SYSTEMS))
    if temp_flag:
        qd.echo_info("passed valid os list test as all in [%s] "
                     "contained in VALID_OPERATING_SYSTEMS [%s]"
                     % (os_list, VALID_OPERATING_SYSTEMS))
    else:
        ret_val = False

    # tags is a csv string stored in tokens[1], so convert to a list and check they are valid
    # use a temp_flag as there are potentially many tags to check
    temp_flag = True
    tag_list = tokens[1].split(",")
    for t_val in tag_list:
        if t_val not in qd_topo.ALL_APP_TAGS:
            temp_flag = False
            qd.echo_error("tag [%s] from [%s], is not in ALL_APP_TAGS [%s]"
                          % (t_val, tag_list, qd_topo.ALL_APP_TAGS))
    if temp_flag:
        qd.echo_info("passed valid tag list test as all in [%s] contained in ALL_APP_TAGS [%s]"
                     % (tag_list, qd_topo.ALL_APP_TAGS))
    else:
        ret_val = False

    # check that the operation action tokens[2] is valid
    operation_action = tokens[2]
    if operation_action in ALL_ACTIONS:
        qd.echo_info("passed valid operation actions test as [%s] contained in "
                     "ALL_ACTIONS [%s]" % (operation_action, ALL_ACTIONS))
    else:
        qd.echo_error("operation action [%s] is not in ALL_ACTIONS [%s]"
                      % (operation_action, ALL_ACTIONS))
        ret_val = False

    # check that the operation type tokens[3] is valid
    operation_type = tokens[3]
    if operation_type in VALID_OPERATION_TYPES:
        qd.echo_info("passed valid operation type test as [%s] contained in "
                     "VALID_OPERATION_TYPES [%s]" % (operation_type, VALID_OPERATION_TYPES))
    else:
        qd.echo_error("operation type [%s] is not in VALID_OPERATION_TYPES [%s]"
                      % (operation_type, VALID_OPERATION_TYPES))
        ret_val = False

    return ret_val


def old_get_folder_name_match(partial_name, target_folder):
    '''
    *KEEPING FOR NOW AS HISTORY, the problem is that some filenames contain multiple "dots"*

    Returns a list of names (files / folders) that match the {partial_name} in the {target_folder}.
    This is required as the {partial_name}
    1. could be partially supplied eg CLM_server_patch_ instead of CLM_server_patch_<datecode>
    2. could contain a relative path eg server/filename or server/web

    :param partial_name: fragment of file or folder potentially including a relative path
    :param target_folder: folder containing the {partial_name}
    :return: list of file / folder names that match {partial_name} in the {target_folder}

    '''
    # allow for potential of the {partial_name} containing a partial path
    head = qd.dirname(partial_name)
    if head:
        # {partial_name} contains a partial path so update target_folder
        target_folder = qd.path_join(target_folder, head)

    tail = qd.basename(partial_name)
    if tail:
        match_name = qd.path_join(target_folder, tail)
        qd.echo_info("checking for matches against [%s]" % match_name)
        # get list of matches for {partial_name}
        # use "glob" to do the partial name matching
        '''
        ALLOW FOR 2 CONDITIONS
        1. filename has an extension
           here assume it is an exact name and look for it
        2. filename has no extension
           here assume it could be partial name so add a *
           eg {partname} becomes {partname*}
        '''
        p_name, p_extension = os.path.splitext(match_name)
        qd.echo_debug("name [%s] extension [%s]" % (p_name, p_extension))
        if not p_extension:
            match_name = match_name + "*"
        qd.echo_info("checking for matches against [%s]" % match_name)
        match_list = glob.glob(match_name)

    else:
        qd.exit_with_err_msg("partial_name [%s] does not contain a file or folder (tail) part."
                             " Exiting..." % partial_name)

    return match_list


def get_file_folder_name_match(partial_name, target_folder):
    '''
    Returns a list of names (files / folders) that match the {partial_name} in the {target_folder}.

    This is required as the {partial_name}
        1. could be partially supplied eg CLM_server_patch_ instead of CLM_server_patch_<datecode>
        2. could contain a relative path eg server/filename or server/web

    :param partial_name: fragment of file or folder potentially including a relative path
    :param target_folder: folder containing the {partial_name}
    :return: list of file / folder names that match {partial_name} in the {target_folder}

    '''
    # allow for potential of the {partial_name} containing a partial path
    head = qd.dirname(partial_name)
    if head:
        # {partial_name} contains a partial path so update target_folder
        target_folder = qd.path_join(target_folder, head)

    tail = qd.basename(partial_name)
    if tail:
        match_name = qd.path_join(target_folder, tail)
        # get list of matches for {partial_name}
        # use "glob" to do the partial name matching
        # add wildcard to {match_name} to enable partial name matching
        match_name = match_name + "*"
        qd.echo_info("checking for matches against [%s]" % match_name)
        match_list = glob.glob(match_name)

    else:
        qd.exit_with_err_msg("partial_name [%s] does not contain a file or folder (tail) part."
                             " Exiting..." % partial_name)

    return match_list


def get_full_filename(partial_name, target_folder):
    '''
    Calculates the filename and the full path to the target folder.

    This is required as the {partial_name}
        1. could be partially supplied eg CLM_server_patch_ instead of CLM_server_patch_<datecode>
        2. could contain a relative path eg server/filename

    :param partial_name: fragment of filename potentially including a relative path
    :param target_folder: folder containing the {partial_name}
    :return tuple (target_name, target_folder):
        * **target_name**: [filename to action]
        * **target_folder**: [path to target file]

    '''

    # get list of potential files
    match_list = get_file_folder_name_match(partial_name, target_folder)
    # remove any folders that were matched
    match_list = [entry for entry in match_list if os.path.isfile(entry)]
    if len(match_list) == 1:
        full_path_filename = match_list[0]
        qd.check_file_exists(full_path_filename)
    else:
        qd.exit_with_err_msg("Partial filename [%s] did not return only 1 match [%s]"
                             % (partial_name, match_list))

    target_folder = qd.dirname(full_path_filename)
    target_name = qd.basename(full_path_filename)
    return target_name, target_folder


def get_full_foldername(partial_name, target_folder):
    '''
    Calculates the foldername and the full path to the target folder.

    This is required as the {partial_name}
        1. could be partially supplied eg folder_ instead of folder_<datecode>
        2. could contain a relative path eg server/foldername

    :param partial_name: fragment of foldername potentially including a relative path
    :param target_folder: folder containing the {partial_name}
    :return tuple (target_name, target_folder):
        * **target_name**: [folder to action]
        * **target_folder**: [path to target folder]

    '''

    # get list of potential files
    match_list = get_file_folder_name_match(partial_name, target_folder)
    # remove any files that were matched
    match_list = [entry for entry in match_list if os.path.isdir(entry)]
    if len(match_list) == 1:
        full_path_foldername = match_list[0]
        qd.check_folder_exists(full_path_foldername)
    else:
        qd.exit_with_err_msg("Partial foldername [%s] did not return only 1 match [%s]"
                             % (partial_name, match_list))

    target_folder = qd.dirname(full_path_foldername)
    target_name = qd.basename(full_path_foldername)
    return target_name, target_folder


def get_save_in_folder_name(saved_files_folder_head, partial_name, relative_target_folder):
    '''
    Create the path to the {save_in_folder} based on {saved_files_folder_head} and
    {relative_target_folder} and optionally the [head] of {partial_name}

    ref InstallMaintenanceFilesEngine.sh : updateDestinationFoldertail()

    :param saved_files_folder_headfolder: the folder to contain
        any files/folders saved by a future action
    :param partial_name: fragment of filename potentially including a relative path
    :param relative_target_folder: target_folder relative to {clm_home}
    :return: full path to the folder to save any backup files / folders

    ..  warning:: not doing any "partial path processing" until it can be justified,

        returning save_in_folder = qd.path_join(saved_files_folder_head, relative_target_folder)

    '''
    # initial {save_in_folder}
    save_in_folder = qd.path_join(saved_files_folder_head, relative_target_folder)

    # WARNING not doing any "partial path processing" until it can be justified
    skip_partial_name_processing = False
    if skip_partial_name_processing:
        # allow for potential of the {partial_name} containing a partial path
        head = qd.dirname(partial_name)
        if head:
            # {partial_name} contains a partial path so update save_in_folder
            save_in_folder = qd.path_join(save_in_folder, head)

    return save_in_folder


def prepare_group1(clm_home, tokens, saved_files_folder_head):
    '''
    Preparation for the GROUP1_ACTIONS ['backup', 'delete', 'backupDelete']

    token list = os:tag:operation:type:source:destination:string

    Calculates the paths to the target folder and the folder to save any backup content.
    Supports type [file and folder], will exit the script if type is unsupported.

    :param clm_home: location of {JazzTeamServer} folder
    :param tokens: os:tag:operation:type:source:destination:string
    :param saved_files_folder_head: folder to contain any files/folders saved by a future action
    :return tuple (target_type, target_name, target_folder, save_in_folder):
        * **target_type**: [file][folder][does not support files]
        * **target_name**: [file: filename to action] [folder: folder to action]
        * **target_folder**: [path to target file or folder]
        * **save_in_folder**: [file: save file {target_folder/target_name} to {save_in_folder}]
          [folder: save folder {target_folder/target_name} to "{save_in_folder/target_name}]

    '''

    operation_action = tokens[2]
    target_type = tokens[3]
    initial_target_name = tokens[4]
    relative_target_folder = tokens[5]

    # the target_folder is always relative to {clm_home}
    intial_target_folder = qd.path_join(clm_home, relative_target_folder)

    # need to know the full name and path to the target file/folder
    if target_type == "file":
        target_name, target_folder = get_full_filename(initial_target_name, intial_target_folder)
    elif target_type == "folder":
        target_name, target_folder = get_full_foldername(initial_target_name, intial_target_folder)
    else:
        qd.exit_with_err_msg("Operation [%s] on type [%s] is not supported. Exiting..."
                             % (operation_action, target_type))

    # calculate the the path to the folder to save any backup files/folders
    save_in_folder = get_save_in_folder_name(saved_files_folder_head, initial_target_name,
                                             relative_target_folder)

    return target_type, target_name, target_folder, save_in_folder


def prepare_group2(clm_home, tokens, local_ifix_head_folder, saved_files_folder_head):
    '''
    Calculates the paths to the source folder, the target folder and
    the folder to save any backup content.

    Supports type [file and folder], will exit the script if type is unsupported.

    preparation for the GROUP2_ACTIONS ['copy', 'replace']

    token list = os:tag:operation:type:source:destination:string

    :param clm_home: location of {JazzTeamServer} folder
    :param tokens: os:tag:operation:type:source:destination:string
    :param local_ifix_head_folder: folder containing any files/folders used by a future action
    :param saved_files_folder_head: folder to contain any files/folders saved by a future action
    :return tuple (target_type, source_name, source_folder, target_folder, save_in_folder):
        * **target_type**: [file][folder][does not support files]
        * **source_name**: [file: filename to action] [folder: folder to action]
        * **source_folder**: [path to source file or folder]
        * **target_folder**: [path to target file or folder]
        * **save_in_folder**: [file: save file {target_folder/target_name} to {save_in_folder}]
          [folder: save folder {target_folder/target_name} to "{save_in_folder/target_name}]
    '''

    operation_action = tokens[2]
    target_type = tokens[3]
    initial_target_name = tokens[4]
    relative_target_folder = tokens[5]

    # the target_folder is always relative to {clm_home}
    target_folder = qd.path_join(clm_home, relative_target_folder)

    # the source folder is always relative to {local_ifix_head_folder}
    # need to know the full name and path to the source file/folder
    if target_type == "file":
        source_name, source_folder = get_full_filename(initial_target_name, local_ifix_head_folder)
    elif target_type == "folder":
        source_name, source_folder = get_full_foldername(initial_target_name,
                                                         local_ifix_head_folder)
    else:
        qd.exit_with_err_msg("Operation [%s] on type [%s] is not supported. Exiting..."
                             % (operation_action, target_type))

    # calculate the the path to the folder to save any backup files/folders
    save_in_folder = get_save_in_folder_name(saved_files_folder_head, initial_target_name,
                                             relative_target_folder)

    return target_type, source_name, source_folder, target_folder, save_in_folder


def prepare_group3(clm_home, tokens, local_ifix_head_folder):
    '''
    Calculates the paths to the source folder and the target folder.

    Supports type [file], will exit the script if type is unsupported.

    preparation for the GROUP3_ACTIONS ['unzip', 'unzipSameDir', 'untar', 'untarSameDir']

    token list = os:tag:operation:type:source:destination:string

    :param clm_home: location of {JazzTeamServer} folder
    :param tokens: os:tag:operation:type:source:destination:string
    :param local_ifix_head_folder: folder containing any files/folders used by a future action
    :return tuple (target_type, source_name, source_folder, target_folder):
        * **target_type**: [file][does not support files | folder]
        * **source_name**: [file: filename to action]
        * **source_folder**: [path to source file]
        * **target_folder**: [path to target folder]
    '''

    operation_action = tokens[2]
    target_type = tokens[3]
    initial_target_name = tokens[4]
    relative_target_folder = tokens[5]

    # the target_folder is always relative to {clm_home}
    target_folder = qd.path_join(clm_home, relative_target_folder)

    # the source folder is always relative to {local_ifix_head_folder}
    # need to know the full name and path to the source file/folder
    if target_type == "file":
        source_name, source_folder = get_full_filename(initial_target_name, local_ifix_head_folder)
    else:
        qd.exit_with_err_msg("Operation [%s] on type [%s] is not supported. Exiting..."
                             % (operation_action, target_type))

    return target_type, source_name, source_folder, target_folder


def prepare_group4(clm_home, tokens, local_ifix_head_folder, saved_files_folder_head):
    # pylint: disable=line-too-long
    '''
    Calculates the paths to the source folder, the target folder
    and the folder to save any backup content.

    Supports type [file, files and folder], will exit the script if type is unsupported.

    preparation for the GROUP4_ACTIONS ['unzipCopy', 'unzipReplace', 'untarCopy', 'untarReplace']

    token list = os:tag:operation:type:source:destination:string

    :param clm_home: location of {JazzTeamServer} folder
    :param tokens: os:tag:operation:type:source:destination:string
    :param local_ifix_head_folder: folder containing any files/folders used by a future action
    :param saved_files_folder_head: folder to contain any files/folders saved by a future action
    :return tuple (target_type, source_name, source_folder, unzip_folder, target_folder, save_in_folder):
        * **target_type**: [file][files][folder]
        * **source_name**: [file: filename to action]
        * **source_folder**: [path to source file]
        * **unzip_folder**: [path to folder to temporarily unzip into]
        * **target_folder**: [path to target folder]
        * **save_in_folder**: [file: save file {target_folder/target_name} to {save_in_folder}]
          [folder: save folder {target_folder/target_name} to "{save_in_folder/target_name}]
    '''
    # pylint: enable=line-too-long

    operation_action = tokens[2]
    target_type = tokens[3]
    initial_target_name = tokens[4]
    relative_target_folder = tokens[5]

    # the target_folder is always relative to {clm_home}
    target_folder = qd.path_join(clm_home, relative_target_folder)

    # the source folder is always relative to {local_ifix_head_folder}
    # need to know the full name and path to the source file/folder
    if (target_type == "file") or (target_type == "files") or (target_type == "folder"):
        source_name, source_folder = get_full_filename(initial_target_name, local_ifix_head_folder)
    else:
        qd.exit_with_err_msg("Operation [%s] on type [%s] is not supported. Exiting..."
                             % (operation_action, target_type))

    # calculate the the path to the temporary folder to initially unzip into
    # it is always relative to {local_ifix_head_folder}
    unzip_folder = qd.path_join(local_ifix_head_folder, "Temp", relative_target_folder)

    # calculate the the path to the folder to save any backup files/folders
    save_in_folder = get_save_in_folder_name(saved_files_folder_head, initial_target_name,
                                             relative_target_folder)

    return target_type, source_name, source_folder, unzip_folder, target_folder, save_in_folder


def key_value_pair_valid(tsp_string):
    '''
    Confirm that the key value pair contains a key and a value.
    Will exit the script if a key and a value are not found

    :param tsp_string: key=value pair
    :return tuple (key, value):
        * **key**: the {key} part of the {tsp_string}
        * **value**: the {value} part of the {tsp_string}
    '''
    # check it is a k=v pair
    split_str = tsp_string.split("=", 1)
    if len(split_str) == 2:
        key, value = tsp_string.split("=", 1)
        if key and value:
            qd.echo_info("tsp_string [%s] contains key[%s] value[%s]" % (tsp_string, key, value))
        else:
            qd.exit_with_err_msg("tsp_string [%s] does not contain a key=value pair" % tsp_string)
    else:
        qd.exit_with_err_msg("tsp_string [%s] does not contain a key=value pair" % tsp_string)

    return key, value


def prepare_group5(clm_home, tokens, saved_files_folder_head):
    '''
    preparation for the GROUP5_ACTIONS ['modifyTspKey', 'deleteTspKey']

    token list = os:tag:operation:type:source:destination:string

    Calculates the paths to the target folder and the folder to save any backup content.

    Supports type [file], will exit the script if type is unsupported.

    :param clm_home: location of {JazzTeamServer} folder
    :param tokens: os:tag:operation:type:source:destination:string
    :param saved_files_folder_head: folder to contain any files/folders saved by a future action
    :return tuple (target_type, target_name, target_folder, save_in_folder):
        * **target_type**: [file][does not support folder | files]
        * **target_name**: [file: filename of teamserver.properties file to action]
        * **target_folder**: [path to target file]
        * **save_in_folder**: [file: save file {target_folder/target_name} to {save_in_folder}]
        * **tsp_string**: [key=value pair in teamserver.properties file to update]

    '''

    operation_action = tokens[2]
    target_type = tokens[3]
    initial_target_name = tokens[4]
    relative_target_folder = tokens[5]
    tsp_string = tokens[6]

    # the target_folder is always relative to {clm_home}
    intial_target_folder = qd.path_join(clm_home, relative_target_folder)

    # need to know the full name and path to the target file/folder
    if target_type == "file":
        target_name, target_folder = get_full_filename(initial_target_name, intial_target_folder)
    else:
        qd.exit_with_err_msg("Operation [%s] on type [%s] is not supported. Exiting..."
                             % (operation_action, target_type))

    # calculate the the path to the folder to save any backup files/folders
    save_in_folder = get_save_in_folder_name(saved_files_folder_head, initial_target_name,
                                             relative_target_folder)

    # confirm that there is a key and a value in {tsp_string}
    key_value_pair_valid(tsp_string)

    return target_type, target_name, target_folder, save_in_folder, tsp_string


def do_delete_action(target_type, target_name, target_folder):
    '''
    ref InstallMaintenanceFilesEngine.sh : doDeleteAction()

    Deletes the target
        [target_type = file] delete file target_folder/target_name

        [target_type = folder] deletes the contents and the folder target_folder/target_name

    :param target_type: [file or folder]
    :param target_name: name of file or folder to delete
    :param target_folder: full path to the folder containing the {target_name}
    '''

    if target_type == "file":
        actual_target_file = qd.path_join(target_folder, target_name)
        qd.delete_file(actual_target_file)

    elif target_type == "folder":
        actual_target_folder = qd.path_join(target_folder, target_name)
        qd.delete_folder(actual_target_folder)

    else:
        qd.exit_with_err_msg("Unsupported target_type [%s]" % target_type)


def do_backup_action(target_type, target_name, target_folder, save_in_folder):
    '''
    ref InstallMaintenanceFilesEngine.sh : doBackupAction()

    Backs up the target
        [target_type = file] backs up file target_folder/target_name
        to save_in_folder/target_name

        [target_type = folder] backs up the contents of folder target_folder/target_name
        to save_in_folder/target_name

    :param target_type: [file or folder]
    :param target_name: name of file or folder to back up
    :param target_folder: full path to the folder containing the {target_name}
    :param save_in_folder: full path to the folder to save any backup files / folders
    '''

    if target_type == "file":
        actual_target_file = qd.path_join(target_folder, target_name)
        actual_saved_file = qd.path_join(save_in_folder, target_name)
        # get file attributes of existing {actual_target_file} for use later
        file_stat = get_file_folder_stat(actual_target_file)
        # backup the original file
        do_copy_file_action(target_name, target_folder, save_in_folder)
        # update the file attributes of the saved file
        set_file_folder_stat(actual_saved_file, file_stat, False)

    elif target_type == "folder":
        actual_target_folder = qd.path_join(target_folder, target_name)
        actual_save_in_folder = qd.path_join(save_in_folder, target_name)
        # get folder attributes of existing {actual_target_folder} for use later
        file_stat = get_file_folder_stat(actual_target_folder)
        # backup the original folder
        do_copy_folder_action(actual_target_folder, actual_save_in_folder)
        # recursively update the folder attributes of the saved folder
        set_file_folder_stat(actual_save_in_folder, file_stat, True)

    else:
        qd.exit_with_err_msg("Unsupported target_type [%s]" % target_type)


def do_backup_delete_action(target_type, target_name, target_folder, save_in_folder):
    '''
    ref InstallMaintenanceFilesEngine.sh : doBackupAction() & doDeleteAction()

    Backs up the target then deletes the original
        [target_type = file] backs up file target_folder/target_name
        to save_in_folder/target_name, then deletes file target_folder/target_name

        [target_type = folder] backs up the contents and the folder target_folder/target_name
        to save_in_folder/target_name, then deletes folder target_folder/target_name

    :param target_type: [file or folder]
    :param target_name: name of file or folder to back up
    :param target_folder: full path to the folder containing the {target_name}
    :param save_in_folder: full path to the folder to save any backup files / folders
    '''

    # back up fist
    do_backup_action(target_type, target_name, target_folder, save_in_folder)
    # then delete
    do_delete_action(target_type, target_name, target_folder)


def do_copy_file_action(source_name, source_folder, target_folder):
    '''
    ref InstallMaintenanceFilesEngine.sh : doCopyFileAction()

    Copies file source_folder/source_name to target_folder/source_name
    Will create full path to target_folder if it does not already exist

    :param source_name: name of file to copy
    :param source_folder: name of the folder containing {source_name}
    :param target_folder: full path to the folder containing the {source_name} to be overwritten
    '''

    actual_source_file = qd.path_join(source_folder, source_name)
    actual_target_file = qd.path_join(target_folder, source_name)

    qd.check_file_exists(actual_source_file)
    # delete target file if it already exists?
    qd.delete_file_if_exists(actual_target_file)
    # need to create the target folder if it does not already exist
    qd.check_and_create_folder(target_folder)
    # copy {actual_source_file} to {actual_target_file}
    copied_file = qd.copy_file_with_stat(actual_source_file, target_folder)
    if copied_file:
        if copied_file == actual_target_file:
            qd.check_file_exists(copied_file)
        else:
            qd.exit_with_err_msg("copied file [%s] does not have the expected name [%s]"
                                 % (copied_file, actual_target_file))
    else:
        qd.exit_with_err_msg("failed to copy file [%s]" % actual_source_file)


def do_copy_folder_action(source_folder, target_folder):
    '''
    ref InstallMaintenanceFilesEngine.sh : doCopyFolderAction()

    Copies contents of folder {source_folder} to {target_folder}

    :param source_folder: full path to the the folder containing content to be copied
    :param target_folder: full path to the folder getting overwritten
    '''

    qd.check_folder_exists(source_folder)
    # delete target folder if it already exists?
    qd.delete_folder_if_exists(target_folder)
    # no need to create the target folder?
    copied_folder = qd.copy_folder_with_stat(source_folder, target_folder)
    if copied_folder:
        if copied_folder == target_folder:
            qd.check_folder_exists(copied_folder)
        else:
            qd.exit_with_err_msg("copied folder [%s] does not have the expected name [%s]"
                                 % (copied_folder, target_folder))
    else:
        qd.exit_with_err_msg("failed to copy folder [%s]" % source_folder)


def do_copy_action(target_type, source_name, source_folder, target_folder):
    '''
    ref InstallMaintenanceFilesEngine.sh : doCopyAction()

    Copies the source to the target, overwriting if required
        [target_type = file] copies file source_folder/source_name to target_folder/source_name

        [target_type = folder] copies contents of folder source_folder/source_name
        to target_folder/source_name

    :param target_type: [file or folder]
    :param source_name: name of file or folder to copy
    :param source_folder: name of the folder containing {source_name}
    :param target_folder: full path to the folder containing the {source_name} to be overwritten
    '''

    if target_type == "file":
        actual_target_file = qd.path_join(target_folder, source_name)
        # get file attributes of existing {actual_target_file} for use later
        file_stat = get_file_folder_stat(actual_target_file)
        # do action on target file
        do_copy_file_action(source_name, source_folder, target_folder)
        # update the file attributes of the new target file
        set_file_folder_stat(actual_target_file, file_stat, False)

    elif target_type == "folder":
        actual_source_folder = qd.path_join(source_folder, source_name)
        actual_target_folder = qd.path_join(target_folder, source_name)
        # get folder attributes of existing {actual_target_folder} for use later
        file_stat = get_file_folder_stat(actual_target_folder)
        # do action on target folder
        do_copy_folder_action(actual_source_folder, actual_target_folder)
        # recursively update the folder attributes of the new target folder
        set_file_folder_stat(actual_target_folder, file_stat, True)

    else:
        qd.exit_with_err_msg("Unsupported target_type [%s]" % target_type)


def do_replace_action(target_type, source_name, source_folder, target_folder, save_in_folder):
    '''
    ref InstallMaintenanceFilesEngine.sh : doReplaceAction()

    Backs up the target then copies the source to the target, overwriting if required
        [target_type = file] backs up file target_folder/source_name
        to save_in_folder/source_name, then copies file source_folder/source_name
        to target_folder/source_name

        [target_type = folder] backs up the contents and the folder target_folder/source_name
        to save_in_folder/source_name, then copies contents of folder source_folder/source_name
        to target_folder/source_name

    :param target_type: [file or folder]
    :param source_name: name of file or folder to copy
    :param source_folder: name of the folder containing {source_name}
    :param target_folder: full path to the folder containing the {source_name} to be overwritten
    :param save_in_folder: full path to the folder to save any backup files / folders
    '''

    if target_type == "file":
        actual_target_file = qd.path_join(target_folder, source_name)
        actual_saved_file = qd.path_join(save_in_folder, source_name)
        # get file attributes of existing {actual_target_file} for use later
        file_stat = get_file_folder_stat(actual_target_file)
        # backup the original file
        do_copy_file_action(source_name, target_folder, save_in_folder)
        # update the file attributes of the saved file
        set_file_folder_stat(actual_saved_file, file_stat, False)
        # overwrite with the new file
        do_copy_file_action(source_name, source_folder, target_folder)
        # update the file attributes of the new target file
        set_file_folder_stat(actual_target_file, file_stat, False)

    elif target_type == "folder":
        actual_source_folder = qd.path_join(source_folder, source_name)
        actual_target_folder = qd.path_join(target_folder, source_name)
        actual_save_in_folder = qd.path_join(save_in_folder, source_name)
        # get folder attributes of existing {actual_target_folder} for use later
        file_stat = get_file_folder_stat(actual_target_folder)
        # backup the original folder
        do_copy_folder_action(actual_target_folder, actual_save_in_folder)
        # recursively update the folder attributes of the saved folder
        set_file_folder_stat(actual_save_in_folder, file_stat, True)
        # overwrite with the new folder contents
        do_copy_folder_action(actual_source_folder, actual_target_folder)
        # recursively update the folder attributes of the new target folder
        set_file_folder_stat(actual_target_folder, file_stat, True)

    else:
        qd.exit_with_err_msg("Unsupported target_type [%s]" % target_type)


def do_extract_copy_action(target_type, source_unzip_folder, target_folder):
    '''
    ref InstallMaintenanceFilesEngine.sh : doExtractReplaceAction()

    Copies the content from {source_unzip_folder} to the target {target_folder},
    overwriting if required

        [target_type = file] expects {source_unzip_folder} to contain a single file {filename}.
        Copies file {source_unzip_folder}/{filename} to {target_folder}/{filename}

        [target_type = files] expects {source_unzip_folder} to contain many files.
        Loops through each file {filename}
        copies each file {source_unzip_folder}/{filename} to {target_folder}/{filename}

        [target_type = folder]
        Copies the contents of folder {source_unzip_folder} to {target_folder}

    :param target_type: [file or files or folder]
    :param source_unip_folder: name of folder containing the unzipped contents
    :param target_folder: full path to the folder whose content(s) are to be overwritten
    '''

    if target_type == "file":

        # can only be one file in the folder
        file_list = qd.list_files_in_folder(source_unzip_folder)
        if len(file_list) == 1:
            source_name = file_list[0]
            actual_target_file = qd.path_join(target_folder, source_name)
            # get file attributes of existing {actual_target_file} for use later
            file_stat = get_file_folder_stat(actual_target_file)
            # overwrite the new file
            do_copy_file_action(source_name, source_unzip_folder, target_folder)
            # update the file attributes of the new target file
            set_file_folder_stat(actual_target_file, file_stat, False)
        else:
            qd.exit_with_err_msg("Folder [%s] contains more or less than 1 file [%s]"
                                 % (source_unzip_folder, file_list))

    elif target_type == "files":

        file_list = list_files_in_folder_unless_prefixed(source_unzip_folder, "._")
        # check there is at least 1 file in the folder
        if len(file_list) >= 1:
            for source_name in file_list:
                actual_target_file = qd.path_join(target_folder, source_name)
                # get file attributes of existing {actual_target_file} for use later
                file_stat = get_file_folder_stat(actual_target_file)
                # overwrite the new file
                do_copy_file_action(source_name, source_unzip_folder, target_folder)
                # update the file attributes of the new target file
                set_file_folder_stat(actual_target_file, file_stat, False)
        else:
            qd.exit_with_err_msg("Folder [%s] contains no files [%s]"
                                 % (source_unzip_folder, file_list))

    elif target_type == "folder":
        # get folder attributes of existing {target_folder} for use later
        file_stat = get_file_folder_stat(target_folder)
        # overwrite with folder contents
        do_copy_folder_action(source_unzip_folder, target_folder)
        # recursively update the folder attributes of the new target folder
        set_file_folder_stat(target_folder, file_stat, True)

    else:
        qd.exit_with_err_msg("Unsupported target_type [%s]" % target_type)


def do_extract_replace_action(target_type, source_unzip_folder, target_folder, save_in_folder):
    '''
    ref InstallMaintenanceFilesEngine.sh : doExtractReplaceAction()

    Backs up the target then copies the content from {source_unzip_folder}
    to the target {target_folder}, overwriting if required

        [target_type = file] expects {source_unzip_folder} to contain a single file {filename}.
        Backs up file target_folder/{filename} to save_in_folder/{filename},
        then copies file {source_unzip_folder}/{filename} to {target_folder}/{filename}

        [target_type = files] expects {source_unzip_folder} to contain many files.
        Loops through each file {filename}, backs up file {target_folder}/{filename}
        to save_in_folder/{filename},
        then copies each file {source_unzip_folder}/{filename} to {target_folder}/{filename}

        [target_type = folder] Backs up the contents of target_folder to {save_in_folder}
        then copies the contents of folder {source_unzip_folder} to {target_folder}

    :param target_type: [file or files or folder]
    :param source_unip_folder: name of folder containing the unzipped contents
    :param target_folder: full path to the folder whose content(s) are to be overwritten
    :param save_in_folder: full path to the folder to save any backup files / folders
    '''

    if target_type == "file":

        # can only be one file in the folder
        file_list = qd.list_files_in_folder(source_unzip_folder)
        if len(file_list) == 1:
            source_name = file_list[0]
            actual_target_file = qd.path_join(target_folder, source_name)
            actual_saved_file = qd.path_join(save_in_folder, source_name)
            # get file attributes of existing {actual_target_file} for use later
            file_stat = get_file_folder_stat(actual_target_file)
            # backup the original file
            do_copy_file_action(source_name, target_folder, save_in_folder)
            # update the file attributes of the saved file
            set_file_folder_stat(actual_saved_file, file_stat, False)
            # overwrite with the new file
            do_copy_file_action(source_name, source_unzip_folder, target_folder)
            # update the file attributes of the new target file
            set_file_folder_stat(actual_target_file, file_stat, False)
        else:
            qd.exit_with_err_msg("Folder [%s] contains more or less than 1 file [%s]"
                                 % (source_unzip_folder, file_list))

    elif target_type == "files":

        file_list = list_files_in_folder_unless_prefixed(source_unzip_folder, "._")
        # check there is at least 1 file in the folder
        if len(file_list) >= 1:
            for source_name in file_list:
                actual_target_file = qd.path_join(target_folder, source_name)
                actual_saved_file = qd.path_join(save_in_folder, source_name)
                # get file attributes of existing {actual_target_file} for use later
                file_stat = get_file_folder_stat(actual_target_file)
                # backup the original file
                do_copy_file_action(source_name, target_folder, save_in_folder)
                # update the file attributes of the saved file
                set_file_folder_stat(actual_saved_file, file_stat, False)
                # overwrite with the new file
                do_copy_file_action(source_name, source_unzip_folder, target_folder)
                # update the file attributes of the new target file
                set_file_folder_stat(actual_target_file, file_stat, False)
        else:
            qd.exit_with_err_msg("Folder [%s] contains no files [%s]"
                                 % (source_unzip_folder, file_list))

    elif target_type == "folder":
        # get folder attributes of existing {target_folder} for use later
        file_stat = get_file_folder_stat(target_folder)
        # backup the original folder
        do_copy_folder_action(target_folder, save_in_folder)
        # recursively update the folder attributes of the saved folder
        set_file_folder_stat(save_in_folder, file_stat, True)
        # overwrite with the new folder contents
        do_copy_folder_action(source_unzip_folder, target_folder)
        # recursively update the folder attributes of the new target folder
        set_file_folder_stat(target_folder, file_stat, True)

    else:
        qd.exit_with_err_msg("Unsupported target_type [%s]" % target_type)


def do_unzip_action(source_name, source_folder, target_folder):
    '''
    ref InstallMaintenanceFilesEngine.sh : unzipZip()

    Extracts the the content from {source_folder}/{source_name}
    to the target {target_folder}, overwriting if required

    :param source_name: filename containing the .zip file
    :param source_folder: full path to the folder containing the {filename}
    :param target_folder: full path to the folder where {filename} content(s) are extracted to
    '''

    zip_file = qd.path_join(source_folder, source_name)
    qd.unzip_zip(zip_file, target_folder)


def do_untar_action(source_name, source_folder, target_folder):
    '''
    ref InstallMaintenanceFilesEngine.sh : untarTar()

    Extracts the the content from {source_folder}/{source_name}
    to the target {target_folder}, overwriting if required

    :param source_name: filename containing the .tar or .tar.gz file
    :param source_folder: full path to the folder containing the {filename}
    :param target_folder: full path to the folder where {filename} content(s) are extracted to
    '''

    tar_file = qd.path_join(source_folder, source_name)
    qd.unzip_zip(tar_file, target_folder)


def do_unzip_extract_copy_action(target_type, source_name, source_folder, unzip_folder,
                                 target_folder):
    '''
    Extracts the the content from {source_folder}/{source_name} to temp folder {unip_folder},
    then copies the contents of {unip_folder} to the target {target_folder},
    overwriting if required

        [target_type = file] expects {unzip_folder} to contain a single file {filename}.
        Copies file {unzip_folder}/{filename} to {target_folder}/{filename}

        [target_type = files] expects {unzip_folder} to contain a many files.
        Loops through each file {filename}
        copies each file {unzip_folder}/{filename} to {target_folder}/{filename}

        [target_type = folder]
        Copies the contents of folder {unzip_folder} to {target_folder}

    :param target_type: [file or files or folder]
    :param source_name: filename containing the .zip file
    :param source_folder: full path to the folder containing the {filename}
    :param unzip_folder: name of folder to extract the unzipped contents into
    :param target_folder: full path to the folder where {filename} content(s) are extracted to

    '''

    do_unzip_action(source_name, source_folder, unzip_folder)
    do_extract_copy_action(target_type, unzip_folder, target_folder)


def do_unzip_extract_replace_action(target_type, source_name, source_folder, unzip_folder,
                                    target_folder, save_in_folder):
    # pylint: disable=too-many-arguments
    '''
    Extracts the the content from {source_folder}/{source_name} to temp folder {unip_folder},
    then copies the contents of {unip_folder} to the target {target_folder},
    overwriting if required

        [target_type = file] expects {unzip_folder} to contain a single file {filename}.
        Copies file {unzip_folder}/{filename} to {target_folder}/{filename}

        [target_type = files] expects {unzip_folder} to contain a many files.
        Loops through each file {filename}
        copies each file {unzip_folder}/{filename} to {target_folder}/{filename}

        [target_type = folder]
        Copies the contents of folder {unzip_folder} to {target_folder}

    :param target_type: [file or files or folder]
    :param source_name: filename containing the .zip file
    :param source_folder: full path to the folder containing the {filename}
    :param unzip_folder: name of folder to extract the unzipped contents into
    :param target_folder: full path to the folder where {filename} content(s) are extracted to
    :param save_in_folder: full path to the folder to save any backup files / folders
    '''

    do_unzip_action(source_name, source_folder, unzip_folder)
    do_extract_replace_action(target_type, unzip_folder, target_folder, save_in_folder)


def do_untar_extract_copy_action(target_type, source_name, source_folder, untar_folder,
                                 target_folder):
    '''
    Extracts the the content from {source_folder}/{source_name} to temp folder {untar_folder},
    then copies the contents of {untar_folder} to the target {target_folder},
    overwriting if required

        [target_type = file] expects {untar_folder} to contain a single file {filename}.
        Copies file {untar_folder}/{filename} to {target_folder}/{filename}

        [target_type = files] expects {untar_folder} to contain a many files.
        Loops through each file {filename}
        copies each file {unzip_folder}/{filename} to {target_folder}/{filename}

        [target_type = folder]
        Copies the contents of folder {unzip_folder} to {target_folder}

    :param target_type: [file or files or folder]
    :param source_name: filename containing the .tar file
    :param source_folder: full path to the folder containing the {filename}
    :param untar_folder: name of folder to extract the untarred contents into
    :param target_folder: full path to the folder where {filename} content(s) are extracted to

    '''

    do_untar_action(source_name, source_folder, untar_folder)
    do_extract_copy_action(target_type, untar_folder, target_folder)


def do_untar_extract_replace_action(target_type, source_name, source_folder, untar_folder,
                                    target_folder, save_in_folder):
    # pylint: disable=too-many-arguments
    '''
    Extracts the the content from {source_folder}/{source_name} to temp folder {untar_folder},
    then copies the contents of {untar_folder} to the target {target_folder},
    overwriting if required

        [target_type = file] expects {untar_folder} to contain a single file {filename}.
        Copies file {untar_folder}/{filename} to {target_folder}/{filename}

        [target_type = files] expects {untar_folder} to contain a many files.
        Loops through each file {filename}
        copies each file {untar_folder}/{filename} to {target_folder}/{filename}

        [target_type = folder]
        Copies the contents of folder {untar_folder} to {target_folder}

    :param target_type: [file or files or folder]
    :param source_name: filename containing the .zip file
    :param source_folder: full path to the folder containing the {filename}
    :param untar_folder: name of folder to extract the unzipped contents into
    :param target_folder: full path to the folder where {filename} content(s) are extracted to
    :param save_in_folder: full path to the folder to save any backup files / folders
    '''

    do_untar_action(source_name, source_folder, untar_folder)
    do_extract_replace_action(target_type, untar_folder, target_folder, save_in_folder)


def backup_tsp_file(target_name, target_folder, save_in_folder):
    '''
    Backs up the target file.
    If the file is a teamserver.properties file then a TSP_history file is created.

    :param target_name: name of file to modify
    :param target_folder: full path to the folder containing the {target_name}
    :param save_in_folder: full path to the folder to save the backup file
    '''

    actual_target_file = qd.path_join(target_folder, target_name)
    actual_saved_file = qd.path_join(save_in_folder, target_name)
    # get file attributes of existing {actual_target_file} for use later
    file_stat = get_file_folder_stat(actual_target_file)
    # backup the original file
    do_copy_file_action(target_name, target_folder, save_in_folder)
    # update the file attributes of the saved file
    set_file_folder_stat(actual_saved_file, file_stat, False)
    # if TSP file create a history file
    if target_name == "teamserver.properties":
        tsp_history_filename = "teamserver-iFix-" + TIMESTAMP + ".properties"
        qd.duplicate_file(actual_target_file, tsp_history_filename)
        # update the file attributes of the duplicated file
        actual_tsp_history_file = qd.path_join(target_folder, tsp_history_filename)
        set_file_folder_stat(actual_tsp_history_file, file_stat, False)


def do_modify_tsp_key_action(target_type, target_name, target_folder, save_in_folder, tsp_string):
    '''
    ref InstallMaintenanceFilesEngine.sh : doDeleteTspKey()

    Backs up the target file then modifies a key in it. If the file is a teamserver.properties file
    then a TSP_history file is created prior to editing it.
    If the key does not exist it will be added

    :param target_type: [file]
    :param target_name: name of file to modify
    :param target_folder: full path to the folder containing the {target_name}
    :param save_in_folder: full path to the folder to save the backup file
    :param tsp_string: key=value pair to delete
    '''

    if target_type == "file":
        actual_target_file = qd.path_join(target_folder, target_name)
        backup_tsp_file(target_name, target_folder, save_in_folder)
        # confirm that there is a key and a value in {tsp_string}
        key, value = key_value_pair_valid(tsp_string)
        # modify (or add) the key in the file
        qd.echo_info("In file [%s] modify key[%s] to contain value[%s]"
                     % (actual_target_file, key, value))
        qd.QDEnv.set_value_to_prop_file(actual_target_file, key, value)

    else:
        qd.exit_with_err_msg("Unsupported target_type [%s]" % target_type)


def do_delete_tsp_key_action(target_type, target_name, target_folder, save_in_folder, tsp_string):
    '''
    ref InstallMaintenanceFilesEngine.sh : doDeleteTspKey()

    Backs up the target file then delete a key from it. If the file is a teamserver.properties file
    then a TSP_history file is created prior to editing it.

        [target_type = folder] backs up the contents and the folder target_folder/target_name
        to save_in_folder/target_name, then deletes folder target_folder/target_name

    :param target_type: [file]
    :param target_name: name of file to modify
    :param target_folder: full path to the folder containing the {target_name}
    :param save_in_folder: full path to the folder to save the backup file
    :param tsp_string: key=value pair to delete
    '''

    if target_type == "file":
        actual_target_file = qd.path_join(target_folder, target_name)
        backup_tsp_file(target_name, target_folder, save_in_folder)
        # confirm that there is a key and a value in {tsp_string}
        key, value = key_value_pair_valid(tsp_string)
        # to delete the key set the value to None
        value = None
        qd.echo_info("In file [%s] delete key[%s]" % (actual_target_file, key))
        qd.QDEnv.set_value_to_prop_file(actual_target_file, key, value)

    else:
        qd.exit_with_err_msg("Unsupported target_type [%s]" % target_type)


def action_config_line(env_props, tokens, local_ifix_head_folder):
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-statements
    '''
    token list = os:tag:operation:type:source:destination:string

    process tokens[2] (operation) and carry out the required action.

    action done in 2 stages
        1. preparation *eg prepare_group2()*
        2. operation *eg do_copy_action()*

    :param env_props: instance of env.properties
    :param tokens: os:tag:operation:type:source:destination:string
    :param local_ifix_head_folder: folder containing any files/folders used by a future action
    '''

    clm_home = env_props.get_values(qd_props.PN_CLM_HOME_PREP)
    saved_files_folder_head = qd.path_join(clm_home, "server", "replacedByIFIX_" + TIMESTAMP)

    action = tokens[2]

    if action in GROUP1_ACTIONS:
        qd.echo_debug("do prepare_group1()")
        target_type, target_name, target_folder, save_in_folder = prepare_group1(
            clm_home, tokens, saved_files_folder_head)
        qd.echo_info("type [%s] name [%s] folder [%s] saveIn [%s]"
                     % (target_type, target_name, target_folder, save_in_folder))
        qd.echo_debug("do group1 actions")

        if action == "backup":
            qd.echo_debug("do backup action")
            do_backup_action(target_type, target_name, target_folder, save_in_folder)

        elif action == "delete":
            qd.echo_debug("do delete action")
            do_delete_action(target_type, target_name, target_folder)

        elif action == "backupDelete":
            qd.echo_debug("do backup delete action")
            do_backup_delete_action(target_type, target_name, target_folder, save_in_folder)

        else:
            qd.exit_with_err_msg("unsupported group 1 action [%s]" % action)

    elif action in GROUP2_ACTIONS:
        qd.echo_debug("do group2 prepare action")
        target_type, source_name, source_folder, target_folder, save_in_folder = prepare_group2(
            clm_home, tokens, local_ifix_head_folder, saved_files_folder_head)
        qd.echo_info("type [%s] name [%s] source folder [%s] target folder [%s] saveIn [%s]"
                     % (target_type, source_name, source_folder, target_folder, save_in_folder))
        qd.echo_debug("do group2 actions")

        if action == "copy":
            qd.echo_debug("do copy action")
            do_copy_action(target_type, source_name, source_folder, target_folder)

        elif action == "replace":
            qd.echo_debug("do replace action")
            do_replace_action(target_type, source_name, source_folder, target_folder,
                              save_in_folder)

        else:
            qd.exit_with_err_msg("unsupported group 2 action [%s]" % action)

    elif action in GROUP3_ACTIONS:
        qd.echo_debug("do group3 prepare action")
        target_type, source_name, source_folder, target_folder = prepare_group3(
            clm_home, tokens, local_ifix_head_folder)
        qd.echo_info("type [%s] name [%s] source folder [%s] target folder [%s]"
                     % (target_type, source_name, source_folder, target_folder))
        qd.echo_debug("do group3 actions")

        if action == "unzip":
            qd.echo_debug("do unzip action")
            do_unzip_action(source_name, source_folder, target_folder)

        elif action == "unzipSameDir":
            qd.echo_debug("do unzipSameDir action")
            do_unzip_action(source_name, source_folder, source_folder)

        elif action == "untar":
            qd.echo_debug("do untar action")
            do_untar_action(source_name, source_folder, target_folder)

        elif action == "untarSameDir":
            qd.echo_debug("do untarSameDir action")
            do_untar_action(source_name, source_folder, source_folder)

        else:
            qd.exit_with_err_msg("unsupported group 3 action [%s]" % action)

    elif action in GROUP4_ACTIONS:
        qd.echo_debug("do group4 prepare action")
        # pylint: disable=line-too-long
        target_type, source_name, source_folder, unzip_folder, target_folder, save_in_folder = prepare_group4(
            clm_home, tokens, local_ifix_head_folder, saved_files_folder_head)
        # pylint: enable=line-too-long
        qd.echo_info("type [%s] name [%s] source folder [%s] unzip_folder [%s] "
                     "target folder [%s] saveIn [%s]"
                     % (target_type, source_name, source_folder, unzip_folder,
                        target_folder, save_in_folder))
        qd.echo_debug("do group4 actions")

        if action == "unzipCopy":
            qd.echo_debug("do unzipCopy action")
            do_unzip_extract_copy_action(target_type, source_name, source_folder, unzip_folder,
                                         target_folder)

        elif action == "unzipReplace":
            qd.echo_debug("do unzipReplace action")
            do_unzip_extract_replace_action(target_type, source_name, source_folder, unzip_folder,
                                            target_folder, save_in_folder)

        elif action == "untarCopy":
            qd.echo_debug("do untarCopy action")
            do_untar_extract_copy_action(target_type, source_name, source_folder, unzip_folder,
                                         target_folder)

        elif action == "untarReplace":
            qd.echo_debug("do untarReplace action")
            do_untar_extract_replace_action(target_type, source_name, source_folder, unzip_folder,
                                            target_folder, save_in_folder)

        else:
            qd.exit_with_err_msg("unsupported group 4 action [%s]" % action)

    elif action in GROUP5_ACTIONS:
        qd.echo_debug("do group5 prepare action")
        target_type, target_name, target_folder, save_in_folder, tsp_string = prepare_group5(
            clm_home, tokens, saved_files_folder_head)
        qd.echo_info("type [%s] name [%s] folder [%s] saveIn [%s] tsp_string [%s]"
                     % (target_type, target_name, target_folder, save_in_folder, tsp_string))
        qd.echo_debug("do group5 actions")

        if action == "modifyTspKey":
            qd.echo_debug("do modifyTspKey action")
            do_modify_tsp_key_action(target_type, target_name, target_folder, save_in_folder,
                                     tsp_string)

        elif action == "deleteTspKey":
            qd.echo_debug("do deleteTspKey action")
            do_delete_tsp_key_action(target_type, target_name, target_folder, save_in_folder,
                                     tsp_string)

        else:
            qd.exit_with_err_msg("unsupported group 5 action [%s]" % action)

    else:
        qd.exit_with_err_msg("unsupported action [%s]" % action)


def get_file_folder_stat(path):
    '''
    return information about a file or folder.

    If file or folder does not exist returns None

    .. warning::
       On Windows, some return values may be filled with dummy values

    :param path: full path to the file or folder to "stat"
    :return tuple (user_id, group_id, mode):
        * user_id - config line format valid
        * group_id - config line format NOT valid
        * chmod_mode - protection bits (decimal) converted to chmod format (octal)
    '''
    if os.path.isdir(path) or os.path.isfile(path):
        user_id = os.stat(path).st_uid
        group_id = os.stat(path).st_gid
        mode = os.stat(path).st_mode
        size = os.stat(path).st_size
        # mode is decimal, it needs converted to chmod format, octal and then only last 3 RHS chars
        chmod_mode = str(oct(mode))[-3:]
        qd.echo_info("stat [%s] owner [%s] group [%s] mode [%s] chmod_mode [%s] size [%s]"
                     % (path, user_id, group_id, mode, chmod_mode, size))

        # For future information work out who owns the windows file/folder
        if qd.LOG_DEFAULT_LEVEL == logging.DEBUG:
            if qd.is_windows():

                if os.path.isdir(path):
                    # use folderpath to avoid having to match "."
                    folderpath = qd.dirname(path)

                    # dir command produces a multi line output, so parse via a list
                    foldername = qd.basename(path)
                    temp_out = qd.check_output_strip("dir /q %s" % folderpath, shell=True)
                    # qd.echo_debug("[dir /q %s] returned [%s]" % (folderpath, temp_out))
                    temp_list = temp_out.splitlines()
                    # remove all list entries except the 1 line containing the folder info
                    # typical line ['10/28/2016 02:35 PM <DIR> BUILTIN\Administrators RhapsodyMS']
                    new_list = [temp_line for temp_line in temp_list if foldername in temp_line]
                    win_user = qd.basename(new_list[0].split()[-2])
                    qd.echo_debug("for [%s] extracted user [%s] using [dir /q]" % (path, win_user))

                elif os.path.isfile(path):

                    # dir command produces a multi line output, so parse via a list
                    filename = qd.basename(path)
                    temp_out = qd.check_output_strip("dir /q %s" % path, shell=True)
                    # qd.echo_debug("[dir /q %s] returned [%s]" % (path, temp_out))
                    temp_list = temp_out.splitlines()
                    # remove all list entries except the 1 line containing the file info
                    # typical line ['10/28/2016 02:35 PM 45,158,722 BUILTIN\Administrators rs.war']
                    new_list = [temp_line for temp_line in temp_list if filename in temp_line]
                    win_user = qd.basename(new_list[0].split()[-2])
                    qd.echo_debug("for [%s] extracted user [%s] using [dir /q]" % (path, win_user))

                else:
                    qd.exit_with_err_msg("path [%s] does not refer to an existing file or folder"
                                         % path)

        return user_id, group_id, chmod_mode
    else:
        qd.echo_info("path [%s] does not refer to an existing file or folder" % path)
        return None


def set_file_folder_stat(path, stat, recursive_flag):
    '''
    Update file or folder stat information.
    Will recursively travel the folder tree if {recursive_flag} is True

    :param path: full path to the file or folder to "stat"
    :param stat: tuple (user_id, group_id, chmod_mode(octal))
        Note: if chmod_mode = None will only update (user_id, group_id)
    :param recursive_flag: (boolean)
        **True** - will travel the full folder tree
        **False** - will just do the single file / folder
    '''
    if os.path.isdir(path) or os.path.isfile(path):
        if stat:
            # valid entries
            user_id = stat[0]
            group_id = stat[1]
            chmod_mode = stat[2]

        else:
            qd.echo_warning("no stat information so will use OSUser information")
            env_props = qd_props.get_default_env_props()
            user_props = qd_props.get_user_props(env_props)
            user_id, group_id = qd_props.get_os_users_name_and_group(user_props)
            chmod_mode = None

        # update the file / folder information
        qd.echo_info("update [%s] owner [%s] group [%s] chmod_mode [%s]"
                     % (path, user_id, group_id, chmod_mode))
        if recursive_flag:
            qd.chown_r(path, user_id, group_id)
            if chmod_mode:
                qd.chmod_r_777(path, chmod_mode)
        else:
            qd.chown(path, user_id, group_id)
            if chmod_mode:
                qd.chmod_777(path, chmod_mode)

        # display the updated file properties
        get_file_folder_stat(path)
    else:
        qd.exit_with_err_msg("path [%s] does not refer to an existing file or folder" % path)


def list_files_in_folder_unless_prefixed(foldername, prefix):
    # pylint: disable=invalid-name
    '''
    Lists all the files contained in the folder {foldername}, unless they start with the {prefix}

    ...warning:: This is a workaround for unexpected files in the linux-x86-64.tar files supplied
    in the DNG602_OIT_PATCH_July2016.zip

    :param foldername: name of folder to list
        (if no path part will assume {foldername} is in the current folder)
    :param prefix: ignore all files whose name starts with "prefix"

    :return file_list: list of files in  folder {foldername}
    '''

    full_list = qd.list_files_in_folder(foldername)
    if len(full_list) >= 1:
        # filter out any starting with the prefix
        qd.echo_info("remove any files from the list that start with [%s]" % prefix)
        filtered_list = [f_name for f_name in full_list if not f_name[:len(prefix)] == prefix]
        qd.echo_info("filtered contents of folder [%s] [%s]" % (foldername, filtered_list))
        return filtered_list
    else:
        # list is empty so no need to further filter
        return full_list


def main(argv):
    '''
    Main function; see module documentation.

    Install CLM / DM / DMMS iFix fix pack.

    :param argv[0]: the ifix config file that tells the engine what to do
    :param argv[1]: the ifix zip file containing the patch
    '''

    if len(argv) == 2:
        config_file = argv[0]
        ifix_zip_file = argv[1]

    else:
        qd.echo_error("Passed in number of arguments [%s]" % len(argv))
        qd.exit_with_err_msg("SYNTAX  [arg1: ifix config file] [arg2: ifix zip file]")

    # check the files exist
    qd.echo_info("Check ifix config file exists")
    qd.check_file_exists(config_file)
    qd.echo_info("Check ifix zip file exists")
    qd.check_file_exists(ifix_zip_file)

    # get info about the current server
    env_props = qd_props.get_default_env_props()
    # get the APPS previously installed on the current server
    server_apps = get_apps(env_props)
    qd.echo_info("server_apps [%s]" % server_apps)

    # unzip the ifix zip file into a temp folder so that it is always writable
    unzip_dir = tempfile.mkdtemp(prefix='iFixUnzip_')
    qd.echo_info("temporary unzip dir is [%s]" % unzip_dir)

    qd.unzip_zip(ifix_zip_file, unzip_dir)
    # list the contents of the dir
    qd.list_files_in_folder(unzip_dir)

    # parse the ifix config file one line at a time and action the line
    with open(config_file) as file_handle:
        for line in file_handle:
            # remove leading / training whitespaces
            line = line.strip()
            if valid_config_line(line):
                tokens = tokenise_config_line(line)
                qd.echo_info("==== tokenised config line [%s] ====" % tokens)
                if valid_config_line_format(tokens):
                    if matching_app(server_apps, tokens):
                        if matching_os(tokens):
                            action_config_line(env_props, tokens, unzip_dir)
                        else:
                            qd.echo_info("config line contains no matching os's for current server")
                    else:
                        qd.echo_info("config line contains no matching apps for current server")
                else:
                    qd.exit_with_err_msg("Invalid config line")

    # delete the temporary unzip folder
    qd.delete_folder_if_exists(unzip_dir)

    qd.echo_info("=== engine has finished parsing the config from [%s] ===" % config_file)


if __name__ == '__main__':
    logging.basicConfig(level=qd.LOG_DEFAULT_LEVEL, format=qd.LOG_DEFAULT_FORMAT,
                        datefmt=qd.LOG_DEFAULT_DATEFMT)
    main(sys.argv[1:])
    exit(0)
