#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2016, 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Provides common functions for dealing with IBM Installation Manager.

- generate_imxml: generate an IM response XML from a template enabling elements and
  filling variables
- execute_im: execute IM to install products using a response XML file
- Concept of IM response components: see comments around IMRespCompo definition
'''
import collections
import re

from qdlib import qd, qd_props, install_util

DEFAULT_IM_EXEC = qd.os_switch('IBMIM', 'IBMIMc.exe')  #:
''' Default Installation Manager executable '''


##
# IM Functions
##
# pylint: disable=too-many-branches
def generate_imxml(im_template_file, im_xml, replacements=None, flags=None):
    '''Generate an IM response file (im_xml) from a given template (im_template_file).

    :param im_template_file: path to the IM template file
    :param im_xml: path to IM response file XML generated
    :param replacements: a dictionary or list of pairs of replacement KEY and the VALUE
    :param flags: a dictionary or list of tuples(KEY, BOOLEAN)

    **TEMPLATE PROCESS**
    The tags in IM response file template can be enabled/disabled using flags.

    1. Enables XMLs between BG_KEY and END_KEY when KEY is mapped to True in flags::

        <!--_BG_KEY_
        <tag1 .../><tag2../>
        _END_KEY_-->
    2. Enables a single tag (tagname) when KEY is mapped to True in flags::

        <!--_BG_KEY_tagname attr1="v"/_END_KEY_-->
        The above line is evaluated to
        <tagname attr1="v"/>
    3. Enables an open tag (tagname) when KEY is mapped to True in flags::

        <!--_BG_KEY_tagname ... _END_KEY_-->
    4. Enables a close tag (tagname) when KEY is mapped to True in flags::

        <!--_BG_KEY_/_END_KEY_-->
    5. Enables an element (tagname) and the children::

        <!--_BG_KEY_tagname ..>
            <tag1 ... /><tag2 ... />
        </_END_KEY_-->

    A replacement key REP_KEY is replaced by its value (rep_value)::

        <tag1 attr1="_REP_KEY_"/>
        The above line is evaluated to
        <tag1 attr1="rep_value"/>
        Please note that under score is necessary both before and after REP_KEY.

    **REPLACEMENT ORDER**
    For replacements and flags: When it is passed as dict, this method processes
    longer keys first to avoid conflict.
    '''

    # Initialize None, if necessary.
    if replacements is None:
        replacements = {}
    if flags is None:
        flags = []

    # load template
    with open(im_template_file) as file_in:
        im_xml_lines = file_in.readlines()

    # reorder
    if replacements.__class__ is dict:
        # sort by key's length
        replacements = sorted(replacements.items(), key=lambda e: -len(e[0]))
    if flags.__class__ is dict:
        # sort by key's length
        flags = sorted(getattr(flags, 'items'), key=lambda e: -len(e[0]))

    # process flags
    for flag_obj in flags:
        if isinstance(flag_obj, tuple):
            flag_key, flag_value = flag_obj
        else:
            flag_key, flag_value = flag_obj, True

        if not flag_value:
            continue
        # replace
        for i, src_line in enumerate(im_xml_lines):
            line = src_line
            # handle BEGIN tag
            line = re.sub(r'<!--_BG_%s_(\w[-A-Za-z0-9]+)' % flag_key, r'<\1', line)
            line = re.sub('<!--_BG_%s_' % flag_key, r'', line)
            # handle END tag
            line = re.sub('(/)_END_%s_([-A-Za-z0-9]+)--' % flag_key, r'\1\2', line)
            line = re.sub('_END_%s_([-A-Za-z0-9]+)--' % flag_key, r'\1', line)
            line = re.sub('(<.*)_END_%s_-->' % flag_key, r'\1>', line)
            line = re.sub('_END_%s_-->' % flag_key, r'', line)
            if src_line != line:
                im_xml_lines[i] = line
                # line %d is replaced

    # process rep
    for rep_key, rep_value in (replacements.items() if isinstance(replacements, dict)
                               else replacements):
        # replace
        for i, src_line in enumerate(im_xml_lines):
            line = re.sub('_%s_' % rep_key, rep_value, src_line)
            if src_line != line:
                im_xml_lines[i] = line
                # line %d is replaced

    # dump
    with open(im_xml, 'w') as file_out:
        for line in im_xml_lines:
            file_out.write(line)


# pylint: disable=too-many-arguments
def execute_im(im_xml, optional_args=None, secure_storage=None, password_file=None,
               im_location=None, im_usercreds=None):
    '''Execute IBMIM(c.exe) for the given IM response file (im_xml).

    :param im_xml: IM response file
    :param optional_args: (optional) list of optional args
    :param secure_storage: (optional) path to secure storage file
    :param password_file: (optional) path to the password file for the secure_storage
    :param im_location: (optional) IM installation directory (qd_props.DEFAULT_IM_ECLIPSE_DIR is
           used if not specified)
    :param im_username: (optional) user who runs the IM. When None, IBMIM executable owner is used
           in Linux and no-su when Windows.
    '''
    im_location = im_location or qd_props.DEFAULT_IM_ECLIPSE_DIR  # fixup im_location

    # build command
    im_exec = qd.path_join(im_location, DEFAULT_IM_EXEC)
    im_ini = qd.path_join(im_location, 'silent-install.ini')
    im_xml = qd.path_join(im_xml)
    cmd = [im_exec, '--launcher.ini', im_ini, '-input', im_xml, '-acceptLicense']

    # add options
    if secure_storage:
        assert password_file
        cmd.extend(['-secureStorageFile', secure_storage, '-masterPasswordFile', password_file])
    if optional_args:
        assert isinstance(optional_args, list)
        cmd.extend(optional_args)

    # run
    if qd.is_linux() and not im_usercreds:
        # drwxr-xr-x  2 clmadmin clmadmin     57 Oct 29 11:09 Desktop
        im_username = qd.check_output_strip("ls -l %s | cut -d' ' -f 3" % im_exec, shell=True)
        qd.echo_debug('Trying to run installation manager as [%s]' % im_username)

    qd.echo_info('IM command [%s]\nInstalling...' % ','.join(cmd))
    install_util.su_call_if_necessary('--', im_usercreds, cmd, call_method=qd.check_call,
                                      cwd=im_location)


##
# IM Response Components
##
# This is advanced concept for managing IM component's dependencies
# especially used in InstallApp.
#
IMRespCompo = collections.namedtuple('IMRespCompo', 'name depends flags repl data')
'''

    :param name: name of an install component (e.g. jts, lqe, ...)
    :param depends: comma-separated list of dependent IM_components
    :param flags: comma-separated list of flags that are activated on evaluating IM template file
    :param repl: key-value dict used for replacements on evaluating IM template file
    :param data: additional user data

    NOTE: for the **name** parameter
        - one starting '_' are internal and not exposed to user
        - one starting with '_base' is always considered
        - one starting with ! is applied when the tag is NOT present
          after all dependencies are considered

'''

# give default=None except for name
IMRespCompo.__new__.__defaults__ = (None,) * (len(IMRespCompo._fields) - 1)


def get_default_base_compo(repo, install_dir, im_shared_dir=None, repo_key='IBMQD_IM_REPO'):
    # pylint: disable=line-too-long
    '''Get a 'typical' base component with initial values for [flags] and [repl]

    The returned base component is a IMRespCompo collection with
        * initial values for [flags] : WIN_INSTALL, IBMQD_REPO
        * initial values for [repl] : locations for IM_INSTALL_DIR, IBMQD_IM_REPO, <os>_INSTALL_LOCATION, <os>_ECLIPSE_CACHE_LOCATION

    :param repo: repository location (either file path or url)
    :param install_dir: installation dir of the product
    :param im_shared_dir: (optional) IMShared directory, the default value
           install_manager.DEFAULT_IM_SHARED is used if not specified
    :return: IMRespCompo collection with initial values for [flags] and [repl]
    '''

    im_shared_dir = im_shared_dir or qd_props.DEFAULT_IM_SHARED  # fixup IMShared
    repo_loc = repo if qd.is_url(repo, accept_schemas=None) else qd.fileurl(repo)
    response = IMRespCompo('_base',
                           flags=qd.os_switch('LINUX_INSTALL', 'WIN_INSTALL') + ',IBMQD_REPO',
                           repl={qd.os_switch('LINUX_INSTALL_LOCATION', 'WIN_INSTALL_LOCATION'): install_dir,
                                 qd.os_switch('LINUX_ECLIPSE_CACHE_LOCATION', 'WIN_ECLIPSE_CACHE_LOCATION'): im_shared_dir,
                                 'IM_INSTALL_DIR': install_dir,
                                 repo_key: repo_loc
                                 })
    return response


def select_im_resp_compos(compo_names, all_components):
    '''
    Select a list of IMRespCompo collections from the {all_components} list of IMRespCompo
    collections, using the component_names and their dependencies.

    :param compo_names: a list of "app names"
    :param all_components: a list of all the available IMRespCompo collections

    :return tuple (selected_components, selected_names, unresolved_names):
        * **selected_components**: list of selected IMRespCompo(s)
        * **selected_names**: list of selected "app names"
        * **unresolved_names**: list of "apps names" that were not found in {all_components}

    User should go through 'unresolved_names' to see there is no error.
    '''

    if isinstance(compo_names, basestring):
        compo_names = qd.csv_split(compo_names)
    compo_names = compo_names[:]
    if '_base' not in compo_names:
        compo_names.append('_base')
    selected_names = []
    unresolved_names = []

    def is_internal_name(name):
        '''boolean test to determine if the {name} is an internal one starting with [_]'''
        return name and name.startswith('_')

    def is_component_selected(name):
        '''boolean test to determine if the {name} is already in the [selected_name] list'''
        return name in selected_names

    def select_component(name):
        ''' Select component recursively considering dependencies.'''

        compo = [component for component in all_components if component.name == name]
        if not compo:
            if name not in unresolved_names:
                unresolved_names.append(name)
        else:
            if not is_component_selected(name):
                selected_names.append(name)
                # add dependencies
                for dep_key in re.split(r'\s*,\s*', compo[0].depends) if compo[0].depends else []:
                    select_component(dep_key)

    # process component keys
    for name in compo_names:
        select_component(name)
    # add negative IM_components
    for component in [component for component in all_components if component.name.startswith('!')]:
        if not is_component_selected(component.name[1:]):
            select_component(component.name)

    # detect conflict (the case that both !key and key are selected)
    if any(((n.startswith('!') and n[1:] in selected_names) for n in selected_names)):
        assert False, 'Key conflict'

    # detect internal missing IM_components
    if any(is_internal_name(n) for n in unresolved_names):
        # pylint: disable=used-before-assignment
        assert False, 'Internal IMRespCompo dependency error. IMRespCompo[%s] is missing.' % n

    selected_components = [component for component in all_components
                           if component.name in selected_names]
    # filter internal components
    selected_names = [n for n in selected_names if not is_internal_name(n)]
    return selected_components, selected_names, unresolved_names


def get_im_flags_and_replacements(components):
    '''
    Get a list of enabled 'flags' and a dict of 'replacements' variables from IMRespCompos.

    :param components: list of IMRespCompo to be considered

    :return tuple (flags, replacements):
        * **flags**: list of flag names enabled by the component list
        * **replacements**: key-values of replacements from the component list
    '''

    def append_list(a_list, str_values):
        '''
        Populate list based on contents of string of values; if a value already in the list, it is
        not added again.

        :param a_list: list to be populated
        :param str_values: list of values to be added to a_list if not already there
        '''
        for value in (qd.csv_split(str_values) if isinstance(str_values, basestring)
                      else str_values):
            if value and value not in a_list:
                a_list.append(value)

    flags, repls = [], {}
    for compo in components:
        # add flags if defined
        if compo.flags:
            append_list(flags, compo.flags)
        # add replacement if defined
        if compo.repl:
            for key, value in compo.repl.items():
                if key not in repls:
                    repls[key] = []
                append_list(repls[key], value)

    # convert replacements values to string
    replacements = {}
    for key, value in repls.items():
        replacements[key] = ','.join(value)

    return flags, replacements
