#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2016, 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Provides common functions required by installation in Quick Deployer.

- User (access permission and 'su' (RunAs)) management
'''
import os
import StringIO

from qdlib import qd

# We have to disable pylint check for grp import error, the code worked on target machine
# however failed pylint and pydev code check.
# pylint: disable=import-error
if qd.is_linux():
    import grp  # @UnresolvedImport
# pylint: enable=import-error


def need_user_change(os_user_name):
    '''
    Return whether you need to change owner to get os_user_name or not.

    :param os_user_name: The target user name should be compared to current user
    :return: The target user name (the os_user_name argument) if necessary, None otherwise
    '''

    if not os_user_name:
        return None

    import getpass
    current_user = getpass.getuser()

    if os_user_name == current_user:  # root / Administrator
        if os_user_name not in ['root', 'Administrator']:
            qd.echo_warning('instuilt.need_user_change(%s) '
                            'identified no need to change user.' % os_user_name)
        return None

    return os_user_name


def need_user_inst(os_user_name):
    '''
    Return whether you need non-root install

    :param os_user_name: OSUserName defined in userCreds.properties
    :return: True if non-root or administrator user is defined in userCreds.proeprties file
    '''

    if qd.is_linux():
        return os_user_name not in ['root']

    if qd.is_windows():
        return os_user_name not in ['administrator', 'Administrator']

    return False  # not sure


def su_call_if_necessary(minus_opt, usercreds, cmd, call_method=None, **kwargs):
    '''
    Call the cmd with 'su' or directly depending on the username.

    :param minus_opt: '-' for new login shell, '--', or None otherwise
    :param usercreds: new su user neme in string (Linux) or a tuple (user,pass)
    :param cmd: command to be run
    :param call_method: available option - call, check_call, check_output, check_output_strip
    :param **kwargs: additional arguments that will be passed into call_method
    :return: return code of the command execution
    '''

    username = usercreds[0] if isinstance(usercreds, (list, tuple)) else usercreds
    su_creds = usercreds if need_user_change(username) else None
    return qd.su_call(minus_opt, su_creds, cmd, call_method, **kwargs)


# pylint: disable=too-many-branches
def ensure_os_user_and_group(os_user_name, os_user_password, os_user_group=None):
    '''
    Check OS username and group on the target machine

    - Create os_user_group if group does not exist on the target machine
    - Create os_user_name if user does not exist on the target machine
    - Add os_user_group to the existing user if the user does not belong to the os_user_group

    :param os_user_name: local user name on the target machine
    :param os_user_password: The password of os_user_name
    :param os_user_group: The os_user_name's group name

    .. note:: For Windows: unlikely to Linux 'su' command, the runas command doesn't work without \
    password even as Administrator. For compatibility, this program saves the password to \
    the OS credentials store using '/savecreds' option and uses the saved password later.
    '''

    if not need_user_change(os_user_name):
        qd.echo_info('SKip modifying users and groups for [%s].' % os_user_name)
        return

    # pylint: disable=misplaced-bare-raise
    if qd.is_linux():
        # Creating group if the group does not exist on the system
        if os_user_group:
            group_names = [g.gr_name for g in grp.getgrall()]
            if os_user_group in group_names:
                qd.echo_info('group [%s] exists' % os_user_group)
            else:
                qd.echo_info('creating group: [groupadd %s]' % os_user_group)
                qd.check_call(['groupadd', os_user_group])

        # Create user if missing
        if qd.call(['id', os_user_name]) == 0:
            qd.echo_info('user exists, check if it belongs to the usergroup')
            info = grp.getgrnam(os_user_group)
            if os_user_name in info.gr_mem:
                qd.echo_info('user belongs to group [%s]' % os_user_group)
            else:
                qd.echo_info('adding group [%s] to the user' % os_user_group)
                qd.check_call(['usermod', '-a', '-G', os_user_group, os_user_name])
        else:
            create_user_cmd = ['useradd', '-d', '/home/' + os_user_name, '-m', os_user_name]
            if os_user_group:
                create_user_cmd += ['-g', os_user_group]
            qd.echo_info("command to create user: [%s]" % ' '.join(create_user_cmd))
            qd.check_call(create_user_cmd)
            qd.check_call(['passwd', os_user_name, '--stdin'],
                          stdin=StringIO.StringIO(os_user_password))

    elif qd.is_windows():
        # Creating group if the group does not exist on the system
        if os_user_group:
            if qd.call(['net', 'localgroup', os_user_group]) == 0:
                qd.echo_info('group [%s] exists' % os_user_group)
            else:
                qd.echo_error('Group [%s] must exist beforehand in Windows.' % os_user_group)
                qd.exit_(1)

        # Create user if missing
        import subprocess
        if qd.call(['net', 'user', os_user_name]) == 0:
            qd.echo_info('user exists')
        else:
            qd.echo_info('creating a user %s using > net user [USER] [PASS] /add' % os_user_name)
            # use subprocess.check_call not to dump password
            subprocess.check_call(['net', 'user', os_user_name, os_user_password, '/add'])
        if os_user_group:
            if qd.call('net localgroup "%s" | findstr "%s"' % (os_user_group, os_user_name),
                       shell=True) != 0:
                qd.check_call(['net', 'localgroup', os_user_group, os_user_name, '/add'])

        # Validate the credentials
        qd.echo_info('Validating credentials...')
        if qd.su_call('-',
                      (os_user_name, os_user_password), 'echo OK', cwd=os.environ['windir']) != 0:
            qd.echo_error('Windows password [%s] does not match to the user [%s]'
                          % ('*' * len(os_user_password), os_user_name))
            qd.exit_(1)
    else:
        raise


def ensure_folder_and_set_permission(path, os_user_name, os_user_group=None):
    # Disable pylint invalid-name because function name is > 30 chars and want to keep it
    # pylint: disable=invalid-name
    '''
    Check the folder's owner, group and permission

    :param path: The folder path that should be checked
    :param os_user_name: The target owner of the folder
    :param os_user_group: The target owner's group
    '''

    # pylint: disable=misplaced-bare-raise
    # Ensure the target folder
    if not os.path.exists(path):
        qd.echo_warning('Missing directory [%s] for setting right permission. Creating...'
                        % path)
        os.makedirs(path)

    # Skip if necessary
    if not need_user_change(os_user_name):
        qd.echo_info('Skip modifying directory [%s] access permission for [%s].'
                     % (path, os_user_name))
        return

    # Set the folder permission
    qd.echo_info("Changing [%s] directory owner|group to be the %s | %s"
                 % (path, os_user_name, os_user_group))
    if qd.is_linux():
        qd.check_call(['chown', os_user_name, path])
        if os_user_group:
            qd.check_call(['chgrp', os_user_group, path])

    elif qd.is_windows():
        qd.check_call(['cacls', path, '/E', '/G', '%s:F' % os_user_name])
        if os_user_group:
            qd.check_call(['cacls', path, '/E', '/G', '%s:F' % os_user_group])

    else:
        raise
