#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2016, 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************'''
'''
Provides UCD TAG-based topology definition and access to the topology in various ways.

- Define TAGs for topology definition
- QDTopo: a class for representing a topology
- QDTopoHost: a class for representing a single host in a QDTopo
'''
from qdlib import qd

##
# TAGS for Quick Deployer topology definition
# - TODO:tags are case insensitive?
##

TAG_DB = 'db'  #:
''' tag name for the database resource '''
TAG_DERBY = 'derby'  #:
''' tag name for OOTB embedded liberty DB '''
TAG_IHS = 'ihs'  #:
''' tag name for the proxy resource '''
TAG_WAS = 'was'  #:
''' tag name for the WAS resource '''
TAG_LIBERTY = 'liberty'  #:
''' tag name for OOTB embedded liberty app server '''

# Apps
TAG_JTS = 'jts'  #:
''' JTS app tag '''
TAG_CCM = 'ccm'  #:
''' CCM app tag '''
TAG_QM = 'qm'  #:
''' QM app tag '''
TAG_RM = 'rm'  #:
''' RDNG/RM app tag '''
TAG_RS = 'rs'  #:
''' RS/Reporting app tag '''
TAG_DCC = 'dcc'  #:
''' DCC app tag '''
TAG_RELM = 'relm'  #:
''' RELM app tag '''
TAG_LDX = 'ldx'  #:
''' LDX app tag '''
TAG_GC = 'gc'  #:
''' GC app tag '''
TAG_LQE = 'lqe'  #:
''' LQE app tag '''
TAG_DM = 'dm'  #:
''' Rhapsody DM app tag '''

# Aggr # TODO:
TAG_CLM = 'clm'  #:
''' Aggregate tag for CLM '''
CLM_TAGS = 'jts,ccm,qm,rm,rs,dcc,relm,lqe,gc,ldx'  #:
''' Aggregate tag for all CLM apps '''
TAG_SSE = 'sse'  #:
''' Aggregate tag for SSE '''
SSE_TAGS = 'jts,ccm,qm,rm,rs,dcc,relm,lqe,gc,ldx,dm'  #:
''' Aggregate tag for all SSE apps '''


# TAG lists
ALL_DB_TAGS = (TAG_DB, TAG_DERBY)  #:
''' All database tags '''
ALL_APPSVR_TAGS = (TAG_WAS, TAG_LIBERTY)  #:
''' All app server tags '''
ALL_APP_TAGS = (TAG_JTS, TAG_CCM, TAG_QM, TAG_RM, TAG_RS,
                TAG_DCC, TAG_RELM, TAG_LDX, TAG_GC, TAG_LQE, TAG_DM)  #:
''' All app tags '''
ALL_TAGS = (TAG_IHS,) + ALL_DB_TAGS + ALL_APPSVR_TAGS + ALL_APP_TAGS  #:
''' All tags '''

# Tables
ALL_APP_TAGS_WITHOUT_TABLE = (TAG_RS,)  #:
''' All app tags, without table (e.g., RS does not have a table) '''


##
# TOPOLOGY CLASSES
##
class QDTopo(object):
    '''
    A Quick Deployer topology.

    - Provides access to information about the topology.

    NOTE: Currently this class assumes:

    - One IHS or Liberty in the topology
    - One JTS in the topology
    - Up to one app instance for each app

    Typical usage for accessing topology info::

        # Assumption: 'globalTagListing' and 'localHostname' in env.properties
        hostname, env_tag_listing = env_props.get_values('localHostname', 'globalTagListing')
        topo = QDTopo(env_tag_listing)
        thost = topo.get_host(hostname)
        # Use thost for getting host info
        thost.get_local_apps()
    '''
    def __init__(self, globalTagListing=None):
        self.hosts = []
        if globalTagListing:
            self.load_tag_listing_string(globalTagListing)

    def load_tag_listing_string(self, value):
        '''
        Load the encoded string in form of the globalTagListing property.

        :param value: string representing tags, separated by ;
        '''
        result = []
        host_values = qd.csv_split(value, ';')
        for host_value in host_values:
            thost = QDTopoHost(None)
            thost.load_tag_listing_string(host_value)
            result.append(thost)
        self.hosts = result

    def to_tag_listing_string(self):
        '''
        Get encoded string in form of the globalTagListing property for an UCD environment.

        :return: string representing global tag listing
        '''
        return ';'.join(h.to_tag_listing_string() for h in self.hosts)

    def add_or_update_host(self, new_thost):
        '''
        Add a new host or replace if the same host name exists.

        :param new_thost: new host name
        '''
        for i, thost in enumerate(self.hosts):
            if thost.hostname == new_thost.hostname:
                self.hosts[i] = new_thost
                break
        else:
            self.hosts.append(new_thost)

    def clean(self):
        '''
        Delete all info from this topology.
        '''
        self.hosts = []

    def get_hosts(self):
        '''
        Return copy of hosts list, i.e., topology.

        :return: copy of hosts list
        '''
        return self.hosts[:]

    def get_host(self, hostname):
        '''
        Get the host whose host name is equals to the given host name.

        :param hostname: host name
        :return: host with given host name
        '''
        return next((host for host in self.hosts if host.hostname == hostname), None)

    def get_hosts_by_tags(self, *search_tags):
        '''
        Find hosts where at least one or more search_tags are applied.

        :param search_tags: one or more search tags
        :return: host with at least one of the given search tags
        '''
        assert len(search_tags) > 0
        # return hosts where any of tag in the search_tags in the host's tags
        return [h for h in self.hosts if any(t in h.tags for t in search_tags)]

    def get_db_host(self):
        '''
        Get DB host in this topology (assuming only one database in the topology).

        :return: database host
        '''
        db_hosts = self.get_hosts_by_tags(*ALL_DB_TAGS)
        assert len(db_hosts) == 1
        return db_hosts[0]

    def get_ihs_host(self):
        '''
        Get IHS host in this topology (assuming only one IHS in the topology).

        :return: IHS host
        '''
        ihs_hosts = self.get_hosts_by_tags(TAG_IHS)
        assert len(ihs_hosts) <= 1
        return ihs_hosts[0] if ihs_hosts else None

    def get_jts_host(self):
        '''
        Get JTS host in this topology (assuming only one JTS in the topology).

        :return: JTS host
        '''
        jts_hosts = self.get_hosts_by_tags(TAG_JTS)
        assert len(jts_hosts) == 1
        return jts_hosts[0]

    def get_app_hosts(self):
        '''
        Get all hosts which have one or more applications.

        :return: all hosts with one or more applications
        '''
        app_hosts = self.get_hosts_by_tags(*ALL_APP_TAGS)
        return app_hosts

    def get_db_hostname(self):
        '''
        Get the sole DB host name (assuming only one database in the topology).

        :return: name of the sole database host
        '''
        return self.get_db_host().hostname

    def get_all_apps_list(self):
        '''
        Get list of all applications.

        :return: list of all applications
        '''
        # gather app tags where one or more hosts assigned
        return [app for app in ALL_APP_TAGS if self.get_hosts_by_tags(app)]

    def get_jts_url_host(self):
        '''
        Get JTS host name (assuming only one JTS in the topology).

        :return: name of the sole JTS host
        '''
        if self.get_ihs_host():
            return self.get_ihs_host().hostname
        else:
            return self.get_jts_host().hostname

    def get_hostname(self, host):
        '''
        Get host name.

        :param host: host to get name of
        :return: name of given host
        '''
        if self.get_ihs_host():
            return self.get_ihs_host().hostname
        elif isinstance(host, QDTopoHost):
            return host.hostname
        else:
            return self.get_host(host).hostname

    def get_user_registry_type_tmp(self):
        '''
        Get user registry type.

        :return: type of user registry, e.g., LDAP
        '''
        if not self.get_ihs_host():
            if self.get_jts_host().get_local_server_role() == TAG_LIBERTY:
                return 'DETECT'
        return 'LDAP'

    def get_table_list(self):
        '''
        Get list of database tables for applications.

        :return: list of database tables
        '''
        result = []
        for app_host in self.get_app_hosts():
            for app in app_host.get_local_apps():
                table_name = app_host.get_table_name_for_app(app)
                if table_name:
                    result.append(table_name)
        return result


class QDTopoHost(object):
    '''
    Class representing a QD topology host.
    '''
    def __init__(self, hostname=None, tags=None, tagListing=None):
        if tags is None:
            tags = []
        self.hostname = hostname
        self.tags = [tag.lower().strip() for tag in tags if tag.strip()]  # retain non-empty tags
        self.was_node_name = None
        if tagListing:
            self.load_tag_listing_string(tagListing)

    def load_tag_listing_string(self, value):
        '''
        Fill this instance from an encoded string.

        :param value: string tag listing
        '''
        hostname, value_without_host = value.split(':', 1)
        if ':' in value_without_host:
            tags_value, node_name = value_without_host.rsplit(':', 1)
        else:
            tags_value, node_name = value_without_host, None

        assert hostname
        self.hostname = hostname
        self.tags = qd.csv_split(tags_value.lower())
        self.was_node_name = node_name

    def to_tag_listing_string(self):
        '''
        Encode this instance to a string for UCD environment property.

        :return: string representing the tags for this host
        '''
        tags_value = (','.join(self.tags)).upper()

        result = '%s:%s' % (self.hostname, tags_value)
        if self.was_node_name:
            result += ':%s' % self.was_node_name
        return result

    def get_local_tags(self):
        '''
        Get tags for this host

        :return: this host's tags
        '''
        return self.tags

    def get_local_apps(self):
        '''
        Get applications for this host.

        :return: this host's applications
        '''
        return [t.lower() for t in self.tags if t.lower() in ALL_APP_TAGS]

    def get_table_name_for_app(self, app):
        '''
        Get database table name for an application on this host.

        :param app: application
        :return: database table name for the given application, or None if application has no table
        '''
        assert app in self.get_local_apps()
        if app in ALL_APP_TAGS_WITHOUT_TABLE:
            return None  # no DB
        return app  # return the app

    def get_local_server_role(self):
        '''
        Get local server role for this host.

        :return: this host's server role
        '''
        result = [tag.lower() for tag in self.tags if tag.lower() in ALL_APPSVR_TAGS]
        assert len(result) == 1
        return result[0]

    def get_local_db(self):
        '''
        Get local database for this host.

        :return: this host's database, or None if no database
        '''
        result = [tag.lower() for tag in self.tags if tag.lower() in ALL_DB_TAGS]
        if result:
            return result[0]
        return None

    def get_was_role(self):
        '''
        Get WAS role for this host.

        :return: 'was' or 'ihs' depending on the role; None if neither WAS nor IHS
        '''
        if TAG_WAS in self.tags:
            return 'was'
        elif TAG_IHS in self.tags:
            return 'ihs'
        else:
            return None
