#!/usr/bin/python
# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2016, 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Provides UCD client, which uses UCD REST APIs.
'''
import codecs
import httplib
import json
import os
import time
import urllib2

from qdlib import qd
from qdlib.url_utils import HttpClient


# pylint: disable=too-many-public-methods
class UDClient(HttpClient):
    '''This class implements UCD interfaces via UCD REST commands.

    - Basic authentication support
    - Token-based authentication support

    See the documentation for REST APIs:
        https://www-01.ibm.com/support/knowledgecenter/SS4GSP_6.1.3/com.ibm.udeploy.reference.doc/topics/rest_api_ref_overview.html
    '''
    @staticmethod
    def create(base_url):
        '''
        Create client using interactive user/pass prompt.

        :param base_url:
        :return: object representing a UCD connection client
        '''
        userid = raw_input('Login to %s:' % base_url)
        import getpass
        password = getpass.getpass('Password: ')
        return UDClient(base_url, userid, password)

    @staticmethod
    def create_from_env(prefix='DS'):
        '''
        Create client from default environment.
        Prefix seems to be DS (default) or AH

        :param prefix: prefix string to add to identify the environment variable containing
            the UCD web url and token
        :return: object representing a UCD connection client
        '''
        return UDClient.create_from_env2((prefix + '_' if prefix else '') + 'WEB_URL',
                                         (prefix + '_' if prefix else '') + 'AUTH_TOKEN')

    @staticmethod
    def create_from_env2(web_url_name, auth_token_name, allow_override=True):
        '''
        Create client from environment but get environment from url, token, etc.
        Prefix seems to be DS (default) or AH

        :param web_url_name: environment variable containing the UCD web url
        :param auth_token_name: environment variable containing the UCD authorisation token
        :param allow_override: True - override with url, token, user credentials from the
            OS environment ; False - use the url and token passed to the method
        :return: object representing a UCD connection client
        '''
        web_url = os.environ.get(web_url_name)
        auth_token = os.environ.get(auth_token_name)
        if allow_override:
            ds_web_url = os.environ.get('MY_WEB_URL')
            ds_auth_token = os.environ.get('MY_AUTH_TOKEN')

            ds_username = os.environ.get('MY_USERNAME')
            ds_password = os.environ.get('MY_PASSWORD')

            if ds_username or ds_password:
                qd.echo_info('Override ucd credentials with [%s/%s]'
                             % (ds_username, '*' * len(ds_password)))
                return UDClient(web_url, userid=ds_username, password=ds_password)
            else:
                if ds_web_url:
                    web_url = ds_web_url
                    qd.echo_info('Override web_url from $DS_WEB_URL [%s]' % ds_web_url)
                if ds_auth_token:
                    auth_token = ds_auth_token
                    qd.echo_info('Override web_url from $DS_WEB_URL [%s]' % ds_auth_token)

        assert auth_token and web_url, \
            "Both %s and %s environment variable need to be set." % (web_url_name, auth_token_name)
        return UDClient(web_url, auth_token=auth_token)

    def __init__(self, base_url=None, userid=None, password=None, auth_token=None):
        '''
        Initialize.
        '''
        super(UDClient, self).__init__(base_url, userid, password, check_ssl_cert=False)
        self.auth_header = None
        if auth_token:
            self.auth_header = 'Basic ' + qd.encode_base64('PasswordIsAuthToken:{"token":"%s"}'
                                                           % auth_token)

    # pylint: disable=too-many-arguments
    def execute_cli(self, path, params=None, contents=None, method=None, expected_errors=None):
        '''
        Execute a REST api request against the UCD web interface

        :param path: REST command suffix eg resource/setProperty
        :param params: dictionary containing the REST command parameters
        :param contents: used if the REST command takes input from a JSON request string or file
        :param method: HTTP method the command uses eg PUT (if None defaults to GET/PUSH)
        :param expected_errors: list containing any expected HTTP errors
        :return: the response from the REST command
        '''

        if params is None:
            params = {}
        if expected_errors is None:
            expected_errors = []

        sesson_id = self._find_cookie_for(self.base_url, 'UCD_SESSION_KEY')
        headers = {'UCD_SESSION_KEY': sesson_id} if sesson_id else {}
        if self.auth_header:
            headers['Authorization'] = self.auth_header

        try:
            qd.echo_info('Invoking UCD REST API: '
                         '[%s], params: [%s], contents: [%s], headers: [%s], method: [%s]'
                         % (path, params, contents, headers, method))
            response = self.execute('cli/' + path, params, contents, headers=headers, method=method)
            qd.echo_debug('UCD REST API response: [%s]' % response.info())
            content_type = response.headers.get('Content-Type')
            if content_type and '/json' in content_type:
                result = json.load(response, encoding='utf-8')
            else:
                result = codecs.decode(response.read(), 'utf-8')
            qd.echo_debug('UCD REST API result: [%s]' % result)
            return result
        except urllib2.HTTPError as error:
            content = error.read()
            error.read = lambda: content
            if error.code in expected_errors:
                qd.echo_debug('Expected error response from UCD REST: %s' % str(error))
                qd.echo_debug('Error message: %s' % content)
                return None
            else:
                qd.echo_error('Error response from UCD REST: %s' % str(error))
                qd.echo_debug('Info: %s' % error.info())
                qd.echo_error('Error message: %s' % content)
                raise

    # ---------------
    # Agent
    def test_agent(self, agent_name_or_id):
        '''
        Test the UCD agent connection to the server.

        :param agent_name_or_id: UCD agent name or reference id
        '''
        # from UCD
        self.execute_cli('agentCLI/test', {'agent': agent_name_or_id}, method='PUT')

    def get_agent_prop(self, agent_name_or_id, name):
        '''
        Get agent property value.

        :param agent_name_or_id: UCD agent name or reference id
        :param name: agent property name
        :return: property value
        '''
        # from UCD
        result = self.execute_cli('agentCLI/getProperty', {'agent': agent_name_or_id, 'name': name})
        return result

    # ---------------
    # Token
    def create_auth_token(self, user, available_duration_in_seconds=600):
        '''
        Create UCD authorisation token.

        :param user: registered UCD user who will own the token
        :param available_duration_in_seconds: how long the token will be valid for
        :return: new authorisation token
        '''
        exp = time.strftime('%m-%d-%Y-%H:%M',
                            time.gmtime(time.time() + available_duration_in_seconds))
        result = self.execute_cli('teamsecurity/tokens', {'user': user, 'expireDate': exp},
                                  method='PUT')
        return result['token']

    # ---------------
    # Application
    def get_app(self, app_name):
        '''
        Get a JSON representation of an application.

        :param app_name: UCD application name or reference id
        :return: JSON representation of the application
        '''
        app_json = self.execute_cli('application/info', {'application': app_name})
        return app_json

    def get_envs_in_app(self, app_name):
        '''
        Get information about all environments in application.

        :param app_name: UCD application name or reference id
        :return: JSONArray of environments, each entry contains a dictionary of properties
        '''
        env_list = self.execute_cli('application/environmentsInApplication',
                                    {'application': app_name})
        return env_list  # .name, .id

    # ---------------
    # Environment
    def get_env(self, app_name, env_name):
        '''
        Get a JSON representation of an environment.

        :param app_name: UCD application name or reference id
        :param env_name: UCD environment name or reference id
        :return: JSON representation of the environment
        '''
        try:
            res = self.execute_cli('environment/info', {'application': app_name,
                                                        'environment': env_name})
            return res
        except urllib2.HTTPError:
            return None

    def create_env(self, app_name, env_name):
        '''
        Create a new UCD environment.

        :param app_name: UCD application name or reference id
        :param env_name: UCD environment name or reference id
        :return: JSON representation of the environment
        '''
        result = self.execute_cli('environment/createEnvironment', {'application': app_name,
                                                                    'name': env_name},
                                  method='PUT')
        return result

    def delete_env(self, app_name, env_name):
        '''
        Delete a UCD environment.

        :param app_name: UCD application name or reference id
        :param env_name: UCD environment name or reference id
        '''
        self.execute_cli('environment/deleteEnvironment',
                         {'application': app_name, 'environment': env_name,
                          'deleteAttachedResources': 'false',
                          'deleteCloudInstances': 'false'},
                         method='DELETE')

    def get_env_props(self, app_name, env_name):
        '''
        Get all properties on a UCD environment.

        :param app_name: UCD application name or reference id
        :param env_name: UCD environment name or reference id
        :return: JSONArray of property names, each entry contains a dictionary of properties
        '''
        # from UCD
        result = self.execute_cli('environment/getProperties', {'application': app_name,
                                                                'environment': env_name})
        return result

    def get_env_prop(self, app_name, env_name, prop_name):
        '''
        Get a specific property on a UCD environment.

        :param app_name: UCD application name or reference id
        :param env_name: UCD environment name or reference id
        :param prop_name: environment property name
        :return: property value
        '''
        # from UCD
        result = self.execute_cli('environment/getProperty',
                                  {'application': app_name, 'environment': env_name,
                                   'name': prop_name},
                                  expected_errors=[httplib.NOT_FOUND])
        return result

    def set_env_prop(self, app_name, env_name, prop_name, value):
        '''
        Set a specific property on a UCD environment.

        :param app_name: UCD application name or reference id
        :param env_name: UCD environment name or reference id
        :param prop_name: environment property name
        :param value: property value
        :return: JSONArray containing the new property name's properties
        '''
        # from UCD
        result = self.execute_cli('environment/propValue',
                                  {'application': app_name, 'environment': env_name,
                                   'name': prop_name, 'value': value},
                                  method='PUT')
        return result

    def get_all_base_resources_in_env(self, app_name, env_name):
        '''
        Get all resources added as base resources to a UCD environment.

        :param app_name: UCD application name or reference id
        :param env_name: UCD environment name or reference id
        :return: JSONArray representation of base resources, each entry is a dictionary
            containing the resource's properties
        '''
        reses = self.execute_cli('environment/getBaseResources', {'application': app_name,
                                                                  'environment': env_name})
        return reses  # reses[0]['path']

    def add_base_resource_to_env(self, app_name, env_name, res_path_or_id):
        '''
        Add base resource to a UCD environment.

        :param app_name: UCD application name or reference id
        :param env_name: UCD environment name or reference id
        :param res_path_or_id: UCD application name or reference id of resource to add
        '''
        self.execute_cli('environment/addBaseResource', {'application': app_name,
                                                         'environment': env_name,
                                                         'resource': res_path_or_id},
                         method='PUT')

    # ---------------
    # Resource
    def get_resource(self, res_path):
        '''
        Get a JSON representation of a resource, including its security settings.

        :param res_path: path to the resource, such as /ResourceGroup/Agents/Agent1
        :return: JSON representation of the resource
        '''
        # from UCD
        result = self.execute_cli('resource/info', {'resource': res_path})
        return result

    def get_child_resources(self, res_path_or_id):
        '''
        Get JSONArray representation of all top-level resources,
        or all children of a specified parent resource.

        :param res_path_or_id: path to the resource, such as /ResourceGroup/Agents/Agent1
        :return: JSONArray representation of all resources
        '''
        result = self.execute_cli('resource', {'parent': res_path_or_id})
        return result  # list of resource jsons

    def get_resource_property(self, res_path_or_id, name):
        '''
        Get specific property value from a UCD resource.

        :param res_path_or_id: path to the resource, such as /ResourceGroup/Agents/Agent1
        :param name: resource property name
        :return: property value
        '''
        result = self.execute_cli('resource/getProperty',
                                  {'resource': res_path_or_id, 'name': name},
                                  expected_errors=[httplib.NOT_FOUND])
        return result

    def get_resource_properties(self, res_path_or_id):
        '''
        Get all properties on a UCD resource.

        :param res_path_or_id: path to the resource, such as /ResourceGroup/Agents/Agent1
        :return: JSONArray of property names, each entry contains dictionary of properties
        '''
        result = self.execute_cli('resource/getProperties',
                                  {'resource': res_path_or_id},
                                  expected_errors=[httplib.NOT_FOUND])
        return result

    def set_resource_property(self, res_path_or_id, name, value):
        '''
        Set a UCD resource property.

        :param res_path_or_id: path to the resource, such as /ResourceGroup/Agents/Agent1
        :param name: property name
        :param value: new property value
        :return: JSONArray containing the new property name's properties
        '''
        result = self.execute_cli('resource/setProperty',
                                  {'resource': res_path_or_id, 'name': name, 'value': value},
                                  method='PUT')
        return result

    def create_child_resource_component(self, parent_path_or_id, description='',
                                        component_name=None):
        '''
        Create UCD child resource (for a UCD component).

        :param parent_path_or_id: path to the resource, such as /ResourceGroup/Agents/Agent1
        :param description: description of the new resource
        :param comonent_name: name of existing UCD component
        :return: JSONArray of property names
        '''
        content = {'description': description, 'parent': parent_path_or_id, 'role': component_name}
        result = self.execute_cli('resource/create', {}, json.dumps(content), method='PUT')
        return result  # a resource json
