# ********************************************************************************
# Licensed Materials - Property of IBM
# IBM Quick Deployer
# (c) Copyright IBM Corporation 2017. All Rights Reserved.
#
# U.S. Government Users Restricted Rights:
# Use, duplication or disclosure restricted by GSA ADP Schedule
# Contract with IBM Corp.
#
# ********************************************************************************
'''
Provides functions and classes for doing things with urls including, accessing files from HTTP,
HTTPS, and FTP servers.  Functionality includes:

- Function to determine whether or not a string is a valid URL
- Basic HTTP client w/ Basic authentication and controlling SSL certificate check
- Basic FTP client that can list files in an FTP directory
'''
import urllib
import urllib2
import urlparse
import os

from HTMLParser import HTMLParser
import re
from ftplib import FTP


# Disable pylint dangerous-default-value. List is technically mutable but the function does not
# modify it; ignoring warning.
# pylint: disable=dangerous-default-value
def is_url(url, accept_schemas=['http', 'https', 'ftp']):
    '''
    Determine whether the url is acceptable and valid, or not.

    :param url: url to be checked
    :param accept_schemas: list of allowed schemas
    :return: True if the url is acceptable else False
    '''
    result = urlparse.urlparse(url)
    return result.netloc and result.scheme and (not accept_schemas or
                                                result.scheme in accept_schemas)


# pylint: disable=too-few-public-methods
class HttpClient(object):
    '''
    A basic HTTP/HTTPS client that can exec various commands as well as get a list of files in a
    directory specified by a URL.
    '''

    # pylint: disable=too-many-arguments
    def __init__(self, base_url, userid=None, password=None, check_ssl_cert=True, Cookie_Jar=None):
        '''
        Constructor for the class
        '''
        import cookielib
        self._cookie_jar = Cookie_Jar or cookielib.CookieJar()
        self.base_url = base_url + ('' if base_url.endswith('/') else '/')
        self.opener = _create_url_opener(base_url, userid, password,
                                         check_ssl_cert, self._cookie_jar)

    # pylint: disable=too-many-arguments
    def execute(self, url, params=None, contents=None, headers=None, method=None):
        '''
        Make http request to download single file

        :param url: location to be read
        :param params: parameters to be appended to URL
        :param contents: request contents
        :param headers: request contents
        :param method: request method (PUT, DELETE, etc.)
        :return: response to request
        '''

        if params is None:
            params = {}
        if headers is None:
            headers = {}

        requrl = _rel_url(self.base_url, url, params, append=False)
        request = urllib2.Request(requrl, contents, headers)
        if method:
            request.get_method = lambda: method
        response = self.opener.open(request)
        return response

    def get_files(self, url):
        '''
        Get the list of files from http web server

        :param url: url to be processed
        :return: list of files at specified url
        '''
        try:
            urllib2.install_opener(self.opener)
            fhandler = urllib2.urlopen(url)
            html = fhandler.read()
            fhandler.close()
        except urllib2.HTTPError as message:
            from qdlib.qd import echo_error
            from qdlib.qd import exit_with_err_msg
            echo_error('Error message: [%s] ' % message)
            exit_with_err_msg("Failed to get the list files from url [%s]" % url)
            return

        parser = MyHTMLParser()
        parser.reset_file_list()
        parser.feed(html)
        return parser.get_file_list()

    def _find_cookie_for(self, request, name):
        '''
        A utility method to get a cookie value. NOTE internal Python method is involved.

        :param request: request for which the cookie is desired
        :param name: cookie name
        :return: cookie value
        '''
        if not isinstance(request, urllib2.Request):
            request = urllib2.Request(str(request))
        # forgive me. want to get cookie for given request
        # Disable pylint protected-access so we can get the cookie.
        # pylint: disable=protected-access
        for cookie in self._cookie_jar._cookies_for_request(request):
            if cookie.name == name:
                return cookie.value
        return None


class MyHTMLParser(HTMLParser):
    '''
    Parse the html page, find all links and use urlparser to find all files on the web page.
    '''
    file_list = []

    def handle_starttag(self, tag, attrs):
        '''
        Parses start tag looking for files.

        :param tag: start tag to be parsed
        :param attrs: not used
        :return: file name
        '''
        # Find the tag with Anchor tag
        if tag != 'a':
            return
        # find attribute that with href
        for item in attrs:
            if item[0].lower() == "href":
                # get the url path
                path = urlparse.urlparse(item[1]).path
                # sometimes href path is empty, not display those
                if not path:
                    return
                else:
                    # Get rid of the parent directory path url link
                    file_name = os.path.basename(path)
                    if not file_name:
                        return
                    else:
                        if file_name not in self.file_list:
                            self.file_list.append(file_name)

    def reset_file_list(self):
        '''
        Empty file list.
        '''
        self.file_list = []

    def get_file_list(self):
        '''
        Get the file list.

        :return: file list retrieved from webserver
        '''
        return self.file_list


def _rel_url(base, relative, params=None, append=False):
    '''
    Calculate relative URL from the base URL.

    :param base: base url
    :param relative: relative url
    :param params: parameters for url
    :param append: whether or not to add a slash to the end of the URL
    :return: relative URL
    '''
    result = base
    if relative is not None:
        result = urlparse.urljoin(result +
                                  ('' if not append or result.endswith('/') else '/'), relative)
    if params:
        parts = list(urlparse.urlparse(result))
        query = dict(urlparse.parse_qsl(parts[4]))
        query.update(params)
        parts[4] = urllib.urlencode(query)
        result = urlparse.urlunparse(parts)
    return result


def _create_url_opener(base_url=None, userid=None, password=None, check_ssl_cert=True,
                       cookie_jar=None):
    '''
    Creates URL opener.

    :param base_url: base URL
    :param userid: id of user to access URL
    :param password: password of user to access URL
    :param check_ssl_cert: whether or not to check SSL
    :param cookie_jar: cookies
    :return: url opener
    '''
    # build http options
    handlers = []

    if userid or password:
        assert base_url and userid and password, \
            'base_url, userid, and password_encoded must not be empty.'
        # create password manager for BASIC authentication (curl -u user:password)
        password_manager = urllib2.HTTPPasswordMgrWithDefaultRealm()

        password_manager.add_password(None, base_url, userid, password)
        handlers.append(urllib2.HTTPBasicAuthHandler(password_manager))

    # ignore SSL (curl -k)
    # This whole block is based on 'opting out' from certificate validation.
    # From Python 2.7.9, server certificates are strictly validated.
    # @see https://www.python.org/dev/peps/pep-0476/
    # Disable all errors/warnings.
    # pylint: disable=invalid-name,protected-access,unexpected-keyword-arg
    try:
        import ssl
        _create_unverified_https_context = ssl._create_unverified_context  # @UndefinedVariable
        if not check_ssl_cert:
            handlers.append(urllib2.HTTPSHandler(context=_create_unverified_https_context()))
    except AttributeError:
        if check_ssl_cert:
            from qdlib.qd import echo_warning
            echo_warning('check_ssl_cert is set, but Python version you are using '
                         'does not verify HTTPS certificates.')

    # add cookie jar
    if cookie_jar:
        handlers.append(urllib2.HTTPCookieProcessor(cookie_jar))

    opener = urllib2.build_opener(*handlers)
    return opener


class FtpClient(object):
    '''
    Handle FTP file access.  Currently no support for user credentials.
    '''

    def get_file_list(self, url):
        '''
        Get the list of files from an ftp server

        :param url: url to be processed
        :return: list of files at url location
        '''
        ftp = FTP()
        # If there is a port specified, split the host and port
        if ':' in self.base_url(url):
            host, port = self.base_url(url).split(':')
            ftp.connect(host, port)
        else:
            ftp.connect(self.base_url(url))

        ftp.login()
        ftp.cwd(self.relative_path(url))
        file_list = ftp.nlst()

        ftp.quit()

        return file_list

    @staticmethod
    def base_url(url):
        '''
        Determines the base url (host name and port) for a url

        :param url: full url
        :return: base url
        '''
        temp = url.replace("ftp://", "")
        return re.sub(r'/.*$', '', temp)

    @staticmethod
    def relative_path(url):
        '''
        Determines the relative part of the url (relative to the host)

        :param url: full url
        :return: relative path
        '''
        return url.replace("ftp://", "")[len(FtpClient.base_url(url)):].rstrip('/').lstrip('/')
